/*
 * IOimg.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    I/O functions for frames.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "nrctypes.h"
#include "nccglob.h"
#include "debug.h"
#include "mainFlags.h"
#include "IOimg.h"

/*
 * Define local function prototypes
 */

static void ioiIntToAsc(int32 n, char *s);
void ioiAllocPic(u_char **picptr, int32 nr, int32 nc);

/*
 * Function implementations
 */

void ioiRotateLongTermMem(struct frame **longTermMem,
			  int *numFramesInLongTermMem,
			  int longTermMemSize)
{
  struct frame *freedFr;
  int          i;

  /* Rotate frame pointers in the buffer */
  freedFr = longTermMem[longTermMemSize-1];

  for(i = longTermMemSize-1;i > 0;i--)
    longTermMem[i] = longTermMem[i-1];

  longTermMem[0] = freedFr;

  if(*numFramesInLongTermMem < longTermMemSize)
    (*numFramesInLongTermMem)++;
}


/*
 *
 * ioiIntToAsc:
 *
 * Parameters:
 *      n   I:  Integer to be converted
 *      s   O:  Resulting string
 *
 * Function:
 *      Converts long integer to string.
 *
 * Returns:
 *      Nothing.
 */

static void ioiIntToAsc(int32 n, char *s)
{
  register int32
    i = 0, j, sign;
  char
    c;

  /*
   * Convert to string starting from the end of number.
   */
  if((sign = n) < 0)
    n = -n;
  while(n != 0){
    s[i++] = n % 10 + '0';
    n = n / 10;
  }

  /*
   * Add sign if needed.
   */
  if(sign < 0)
    s[i++] = '-';
  s[i] = '\0';

  /*
   * Invert the string
   */
  for(i = 0, j = strlen(s) - 1; i < j; i++, j--){
    c = s[i];
    s[i] = s[j];
    s[j] = c;
  }
}


/*
 *
 * ioiPutFrameYUV:
 *
 * Parameters:
 *      image     I:  Frame structure
 *      fileName  I:  Name of the output file
 *      n         I:  Frame number
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *
 * Function:
 *      Writes Y, u, and v components of the given frame structure 
 *      to files fileName.n.Y, fileName.n.U, and fileName.n.V. Assumes
 *      4:2:0 format
 *
 * Returns:
 *      Nothing.
 */

void ioiPutFrameYUV(struct frame *image, char *fileName, int32 n, 
                    int32 picSizeX, int32 picSizeY)
{
  char
    frameNo[20],
    frameName[80],
    frameNameY[100],
    frameNameU[100],
    frameNameV[100];

  /*
   * Construct complete file names.
   */
  ioiIntToAsc(n, frameNo);
  strcpy(frameName, fileName);

  strcpy(frameNameY, frameName);
  strcat(frameNameY, ".Y");

  strcpy(frameNameU, frameName);
  strcat(frameNameU, ".U");

  strcpy(frameNameV, frameName);
  strcat(frameNameV, ".V");

  /*
   * Write Y, u, and v components to file.
   */
  ioiPutImage(frameNameY, image->yptr, picSizeX, picSizeY);
  ioiPutImage(frameNameU, image->uptr, picSizeX >> 1, picSizeY >> 1);
  ioiPutImage(frameNameV, image->vptr, picSizeX >> 1, picSizeY >> 1);
}

/*
 *
 * ioiPutImage:
 *
 * Parameters:
 *      fileName  I:  Name of the output file
 *      image     I:  Byte image
 *      picSizeX  I:  Width of the frame
 *      picSizeY  I:  Height of the frame
 *
 * Function:
 *      Writes image to file.
 *
 * Returns:
 *      Nothing.
 */

void ioiPutImage(char *fileName, u_char *image, int32 picSizeX, 
                        int32 picSizeY) 
{
  FILE *fp;

  if((fp = fopen(fileName, "ab")) 
     == NULL){
    deb1p("ioiPutImage: ERROR. Can't open file %s for writing\n", fileName);
    nccExit(1);
  }
  fwrite((char *) image, picSizeX * picSizeY, sizeof(u_char), fp);
  fclose(fp);
}


/*
 *
 * ioiClearSequence:
 *
 * Parameters:
 *      fileName  I:  Name of the output file
 *
 * Function:
 *      Clears the Y, u, and v files of the given sequence name.
 *
 * Returns:
 *      Nothing.
 */

void ioiClearSequence(char *fileName)

{
  FILE *fp;
  char tmpName[100];
  
  sprintf(tmpName, "%s.Y", fileName);
  if ((fp = fopen(tmpName, "wb")) == NULL) {
    deb1p("ioiClearSequence: ERROR. Can't clear file %s", tmpName);
    nccExit(1);
  }
  fclose(fp);

  sprintf(tmpName, "%s.U", fileName);
  if ((fp = fopen(tmpName, "wb")) == NULL) {
    deb1p("ioiClearSequence: ERROR. Can't clear file %s", tmpName);
    nccExit(1);
  }
  fclose(fp);

  sprintf(tmpName, "%s.V", fileName);
  if ((fp = fopen(tmpName, "wb")) == NULL) {
    deb1p("ioiClearSequence: ERROR. Can't clear file %s", tmpName);
    nccExit(1);
  }
  fclose(fp);
}


/*
 *
 * ioiCalcErrYUV:
 *
 * Parameters:
 *      firstFrame    I:  First frame
 *      secondFrame   I:  Second frame
 *      diffFrame     O:  Difference frame
 *      scale         I:  Scaling parameter
 *      pelsY         I:  Number of pixels in luminance frame
 *      psnrY         O:  PSNR value of the Y frame
 *      psnrU         O:  PSNR value of the u frame
 *      psnrV         O:  PSNR value of the v frame
 *      sqErrY        O:  Squared error of the Y frames
 *      sqErrU        O:  Squared error of the u frames
 *      sqErrV        O:  Squared error of the v frames
 *
 * Function:
 *      Calculates differences between Y, u, and v components of the
 *      given frames. Computes also PSNR values (peak value assumed to 
 *      be 255) and sums of squared errors for the components. Assumes
 *      4:2:0 format. Pixel values in difference images are scaled and
 *      offset of 128 is added:
 *      
 *        diffFRame = 128 + scale*(firstFrame - secondFrame)
 *
 * Returns:
 *      Nothing.
 */

void ioiCalcErrYUV(struct frame *firstFrame, struct frame *secondFrame, 
                   struct frame *diffFrame, int32 scale, int32 pelsY, 
                   float64 *psnrY, float64 *psnrU, float64 *psnrV,
                   int32 *sqErrY, int32 *sqErrU, int32 *sqErrV)

{
  register int32
    i;
  int32
    tmpY, tmpU, tmpV,  
    sqDiffY = 0,
    sqDiffU = 0, 
    sqDiffV = 0;

  /*
   *  Calculate difference image for Y component.
   */
  for(i = 0; i < pelsY; i++){
    if(firstFrame->yptr[i] !=  secondFrame->yptr[i]){ 
      tmpY = (int32)firstFrame->yptr[i] - (int32)secondFrame->yptr[i];
      sqDiffY += tmpY * tmpY;
      
      tmpY = 128 + scale * tmpY;
      if(tmpY > 255)
        tmpY = 255;
      if(tmpY < 0)
        tmpY = 0;
      diffFrame->yptr[i] = (u_char)tmpY;
    }
    else  
      diffFrame->yptr[i] = (u_char)128;
  }

  /*
   * Calculate PSNR and sum of squared errors for Y component.
   */
  *psnrY = 10 * log10((float64)pelsY * 65025.0 / (float)sqDiffY);
  if(sqErrY != NULL)
    *sqErrY = sqDiffY;

  /*
   * Calculate difference images for u and v components. 
   */
  for(i = 0; i < pelsY / 4; i++){
    if(firstFrame->uptr[i] !=  secondFrame->uptr[i]){ 
      tmpU = (int32)firstFrame->uptr[i] - (int32)secondFrame->uptr[i];
      sqDiffU += tmpU * tmpU;
      
      tmpU = 128 + scale * tmpU;
      if(tmpU > 255)
        tmpU = 255;
      if(tmpU < 0)
        tmpU = 0;
      diffFrame->uptr[i] = (u_char)tmpU;
    }
    else  
      diffFrame->uptr[i] = 128;

    if(firstFrame->vptr[i] !=  secondFrame->vptr[i]){ 
      tmpV = (int32)firstFrame->vptr[i] - (int32)secondFrame->vptr[i];
      sqDiffV += tmpV * tmpV;
      
      tmpV = 128 + scale * tmpV;
      if(tmpV > 255)
        tmpV = 255;
      if(tmpV < 0)
        tmpV = 0;
      diffFrame->vptr[i] = (u_char)tmpV;
    }
    else  
      diffFrame->vptr[i] = (u_char)128;
  }

  /*
   * Calculate PSNRs and sums of squared errors for u and v components.
   */
  *psnrU = 10 * log10((float64)pelsY * 65025.0 / (4.0 * (float64)sqDiffU));
  *psnrV = 10 * log10((float64)pelsY * 65025.0 / (4.0 * (float64)sqDiffV));

  if(sqErrU != NULL)
    *sqErrU = sqDiffU;
  if(sqErrV != NULL)
    *sqErrV = sqDiffV;

}

/*
 *
 * ioiAllocFrame:
 *
 * Parameters:
 *      fr        I:  Frame structure
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *
 * Function:
 *      Allocates frame structure.
 *
 * Returns:
 *      Nothing.
 */

void ioiAllocFrame(struct frame *fr, int32 picSizeX, int32 picSizeY)
{
  ioiAllocPic(&(fr->yptr), picSizeY,   picSizeX);
  ioiAllocPic(&(fr->uptr), picSizeY >> 1, picSizeX >> 1);
  ioiAllocPic(&(fr->vptr), picSizeY >> 1, picSizeX >> 1);
}


/*
 *
 * ioiCopyFrame:
 *
 * Parameters:
 *      srcFr     I:  Source frame structure
 *      destFr    I:  Destination frame structure
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *
 * Function:
 *      Copies a frame structure into another.
 *
 * Returns:
 *      Nothing.
 */

void ioiCopyFrame(struct frame *srcFr, struct frame *destFr,
                  int32 picSizeX, int32 picSizeY)
{
  memcpy(destFr->yptr, srcFr->yptr, picSizeX*picSizeY*sizeof(Byte));
  memcpy(destFr->uptr, srcFr->uptr, (picSizeX*picSizeY*sizeof(Byte)) >> 2);
  memcpy(destFr->vptr, srcFr->vptr, (picSizeX*picSizeY*sizeof(Byte)) >> 2);
}




/*
 *
 * ioiInitFrame:
 *
 * Parameters:
 *      srcFr     I:  Source frame structure
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *
 * Function:
 *      Initializes a picture with some default constant value.
 *
 * Returns:
 *      Nothing.
 */

void ioiInitFrame(struct frame *srcFr, int32 picSizeX, int32 picSizeY)
{
  memset(srcFr->yptr, 128, picSizeX*picSizeY*sizeof(Byte));
  memset(srcFr->uptr, 128, (picSizeX*picSizeY*sizeof(Byte)) >> 2);
  memset(srcFr->vptr, 128, (picSizeX*picSizeY*sizeof(Byte)) >> 2);
}



/*
 *
 * ioiExchangeFrames:
 *
 * Parameters:
 *      srcFr     I:  Source frame structure
 *      destFr    I:  Destination frame structure
 *
 * Function:
 *      Exchanges the pointer addresses of two frame structures.
 *
 * Returns:
 *      Nothing.
 */

void ioiExchangeFrames(struct frame *srcFr, struct frame *destFr)
{
  Byte
    *tmpPointer;

  tmpPointer   = srcFr->yptr;
  srcFr->yptr  = destFr->yptr;
  destFr->yptr = tmpPointer;

  tmpPointer   = srcFr->uptr;
  srcFr->uptr  = destFr->uptr;
  destFr->uptr = tmpPointer;

  tmpPointer   = srcFr->vptr;
  srcFr->vptr  = destFr->vptr;
  destFr->vptr = tmpPointer;
}




/*
 *
 * ioiFreeFrame:
 *
 * Parameters:
 *      fr        I:  Frame structure
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *
 * Function:
 *      Frees frame structure.
 *
 * Returns:
 *      Nothing.
 */

void ioiFreeFrame(struct frame *fr)
{
  nccFree(fr->yptr);
  nccFree(fr->uptr);
  nccFree(fr->vptr);
}


/*
 *
 * ioiReadFrameYUV:
 *
 * Parameters:
 *      fr        I:  Frame structure
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *      fileY     I:  File pointer for Y component
 *      fileU     I:  File pointer for u component
 *      fileV     I:  File pointer for v component
 *
 * Function:
 *      Reads Y, u, and v components of frame from files. Assumes
 *      4:2:0 format.
 *
 * Returns:
 *      Nothing.
 */

void ioiReadFrameYUV(struct frame *fr, int32 picSizeX, int32 picSizeY, 
                     FILE *fileY, FILE *fileU, FILE *fileV)
{
  if(fread(fr->yptr, sizeof(u_char), picSizeX * picSizeY * sizeof(u_char),
     fileY) != picSizeX * picSizeY * sizeof(u_char)){
    debPrintf("ioiReadFrameYUV: ERROR. Can't read y input\n");
    nccExit(1);
  }

  if(fread(fr->uptr, sizeof(u_char), (picSizeX * picSizeY * sizeof(u_char)) >> 2,
     fileU) != (picSizeX * picSizeY * sizeof(u_char)) >> 2){
    debPrintf("ioiReadFrameYUV: ERROR. Can't read u input\n");
    nccExit(1);
  }

  if(fread(fr->vptr, sizeof(u_char), (picSizeX * picSizeY * sizeof(u_char)) >> 2,
     fileV) != (picSizeX * picSizeY * sizeof(u_char)) >> 2){
    debPrintf("ioiReadFrameYUV: ERROR. Can't read v input\n");
    nccExit(1);
  }
}


/*
 *
 * ioiSkipFrames:
 *
 * Parameters:
 *      n         I:  Number of frames to skip
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *      fileY     I:  File pointer for Y component
 *      fileU     I:  File pointer for u component
 *      fileV     I:  File pointer for v component
 *
 * Function:
 *      Skips given number of frames. Assumes
 *      4:2:0 format.
 *
 * Returns:
 *      Nothing.
 */

void ioiSkipFrames(int32 n, int32 picSizeX, int32 picSizeY, FILE *fileY, 
                   FILE *fileU, FILE *fileV)
{
  fseek(fileY, n * picSizeX * picSizeY, SEEK_CUR);
  fseek(fileU, (n * picSizeX * picSizeY) >> 2, SEEK_CUR);
  fseek(fileV, (n * picSizeX * picSizeY) >> 2, SEEK_CUR);
}


/*
 *
 * ioiOpenFile:
 *
 * Parameters:
 *      fileY     I:  File pointer for Y component
 *      fileU     I:  File pointer for u component
 *      fileV     I:  File pointer for v component
 *      seqName   I:  Sequence name
 *
 * Function:
 *      Opens sequence for reading. Y, u, and v components are
 *      assumed to be in files seqName.y, seqName.u, and seqName.v.
 *
 * Returns:
 *      Nothing.
 */

void ioiOpenFile(FILE **fileY, FILE **fileU, FILE **fileV, char *seqName)
{

  char
    seqY[200],
    seqU[200],
    seqV[200];

  /*
   * Construct complete file names.
   */

  strcpy(seqY, seqName);
  strcpy(seqU, seqName);
  strcpy(seqV, seqName);

  strcat(seqY, ".y");
  strcat(seqU, ".u");
  strcat(seqV, ".v");

  /*
   * Open files.
   */
  if((*fileY = fopen(seqY, "rb")) == NULL){
    deb1p("ioiOpenFile: ERROR. Can't open file %s for reading\n", seqY);
    nccExit(1);
  }
  
  if((*fileU = fopen(seqU, "rb")) == NULL){
    deb1p("ioiOpenFile: ERROR. Can't open file %s for reading\n", seqU);
    nccExit(1);
  }

  if((*fileV = fopen(seqV, "rb")) == NULL){
    deb1p("ioiOpenFile: ERROR. Can't open file %s for reading\n", seqV);
    nccExit(1);
  }
}


/*
 *
 * ioiAllocPic:
 *
 * Parameters:
 *      fileY     I:  File pointer for Y component
 *      fileU     I:  File pointer for u component
 *      fileV     I:  File pointer for v component
 *      seqName   I:  Sequence name
 *
 * Function:
 *      Opens sequence for reading. Y, u, and v components are
 *      assumed to be in files seqName.y, seqName.u, and seqName.v.
 *
 * Returns:
 *      Nothing.
 */

void ioiAllocPic(u_char **picptr, int32 picSizeX, int32 picSizeY)
{
  *picptr = (u_char *)nccCalloc(picSizeX * picSizeY, sizeof(u_char));
}


