/*
 * PredErrorDecoder.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Higher-level functions for decoding INTER frame. 
 *
 */
#include <stdlib.h>
#include <limits.h>

#include "mainFlags.h"
#include "toolMCC.h"
#include "decMCC.h"
#include "IntraPredDecoder.h"
#include "PredErrorDecoder.h"
#include "mcp.h"
#include "decDemux.h"
#include "motionPred.h"
#include "deblockFilt.h"

extern u_char 
  annexTlookup[32];

extern int decLongTermMemSize;       /* Defined in dec_bits.c and set by
                                      * picture header */

static void decDecodeMccRegion(struct frame *predFr, struct frame *recoFr,
                               object *region, int32 regionSize, 
                               int32 regionMode,
                               int32 sizex, int32 sizey, int32 qp,
                               int *bits_reg_mcc, int *bits_reg_mcc_uv,
                               int *bits_reg_ovh,
                               vdxInstance_t *instance, int *BitError);

static void 
  decDecodeDctRegion(struct frame *predFr, /* In:  Prediction frame      */
                     struct frame *recoFr, /* Out: Reconstruction frame  */
                     object *currRegion,   /* In:  Ptr to the region     */
                     int regionMode,       /* In:  Region mode           */
                     int regionSize,       /* In:  Region size (16 or 8) */
                     int picSizeX,         /* In:  Width of the image    */
                     int picSizeY,         /* In:  Height of the image   */
                     int qp,               /* In:  Quant. param. for luma*/
                     int *bitsOvr,         /* Out: Bits for CNC patterns */
                     int *bitsDctY,        /* Out: Bits for luma DCT     */
                     int *bitsDctC,        /* Out: Bits for chrom DCT    */
                     vdxInstance_t *instance,/* In: Decoding instance    */
                     int *BitError);       /* Out: Bit error flag        */
static void
decDecodeIntraDctRegion(struct frame *recoFr, /* Out: Reconstruction frame  */
                        object *currRegion,   /* In:  Ptr to the region     */
                        int regionMode,       /* In:  Region mode           */
                        int regionSize,       /* In:  Region size (16 or 8) */
                        int pecMode,          /* In:  Pred error coding mode*/
                        int picSizeX,         /* In:  Width of the image    */
                        int picSizeY,         /* In:  Height of the image   */
                        int qp,               /* In:  Quant. param. for luma*/
                        int *bitsOvr,         /* Out: Bits for CNC patterns */
                        int *bitsDctY,        /* Out: Bits for luma DCT     */
                        int *bitsDctC,        /* Out: Bits for chrom DCT    */
                        vdxInstance_t *instance,/* In: Decoding instance    */
                        int *BitError);       /* Out: Bit error flag        */

/*
 *
 * pedRecoInterBlk:
 *
 * Parameters:
 *      predFrame   I:  Predicted frame
 *      diffFRame   I:  Prediction error frame
 *      recoFrame   O:  Reconstructed frame
 *      y           I:  Y - coordinate of the block's upper left corner
 *      x           I:  X - coordinate of the block's upper left corner
 *      xRes        I:  Width of the frame
 *      delta       I:  Width of the block
 *      codedFlag   I:  Coded/Not Coded - flag
 *
 * Function:
 *      Reconstructs image block.
 *
 * Returns:
 *      Nothing. 
 */

void pedRecoInterBlk(int32 *predFrame, int32 *diffFrame, u_char *recoFrame, 
                     int32 y, int32 x, int32 xRes, int32 delta, 
                     int32 codedFlag)
{
  int32 
    i, j,                 /* Loop counters */
    tmp,                  /* Auxiliary variable */
    wrap = xRes - delta;  /* Image wraping parameter */

  /*
   * Find the upper left corner of the block in reconstructed
   * image.
   */
  recoFrame += y * xRes + x;
  if (codedFlag) {

    /*
     * Coded block, sum prediction and prediction error.
     */
    for(i = 0; i < delta; i++, recoFrame += wrap) {
      for(j = 0; j < delta; j++) {
        tmp = *predFrame++ + *diffFrame++;
        *recoFrame++ = tmp < 0 ? (u_char)0 : tmp > 255 ? (u_char)255 : 
                       (u_char)tmp;
      }
    }
  }
  else {
    
    /*
     * Not coded block, copy prediction.
     */
    for(i = 0; i < delta; i++, recoFrame += wrap) {
      for(j = 0; j < delta; j++) {
        tmp = *predFrame++;
        *recoFrame++ = tmp < 0 ? (u_char)0 : tmp > 255 ? (u_char)255 : 
                       (u_char)tmp;
      }
    }
  }
}



/*
 *
 * pedRecoCopyRegionYUV:
 *
 * Parameters:
 *      predFrame   I:  Predicted frame
 *      recoFrame   O:  Reconstructed frame
 *      region      I:  Region object
 *      picSizeX    I:  Width of the Y frame
 *      blockInd    I:  Index used for 8x8 blocks (0, 1, 2, 3)
 *
 * Function:
 *      Reconstructs Y, u, and v regions coded as COPY.
 *      Assumes 4:2:0 image format.
 *
 * Returns:
 *      Nothing. 
 */

void pedRecoCopyRegionYUV(struct frame *predFrame, struct frame *recoFrame, 
                          object *region, int32 picSizeX, int32 blockInd)
{
  int32 
    x, y, xPred, yPred,   /* Loop counters */ 
    coordUV, predCoordUV; /* Counters for pixel location */
  
  u_char 
    *predY, *recoY,   /* Predicted and reconstructed Y frame */ 
    *predU, *recoU,   /* Predicted and reconstructed u frame */ 
    *predV, *recoV;   /* Predicted and reconstructed v frame */

  predY = predFrame->yptr;
  predU = predFrame->uptr;
  predV = predFrame->vptr;
  recoY = recoFrame->yptr;
  recoU = recoFrame->uptr;
  recoV = recoFrame->vptr;

  /*
   * Go through luminance pixels in region's bounding box.
   */
  for(y = region->ymin, yPred = (blockInd & 2) << 2; y <= region->ymax;
    y++, yPred++) {
    for(x = region->xmin, xPred = (blockInd & 1) << 3; x <= region->xmax;
    x++, xPred++) {

      recoY[y * picSizeX + x] = predY[(yPred << 4) + xPred];

      /*
       * Check if chrominance pixels exist and copy.
       */
      if(!(x & 1) && !(y & 1)) {
        coordUV = y * (picSizeX >> 2) + (x >> 1);
        predCoordUV = (yPred << 2) + (xPred >> 1);
        recoU[coordUV] = predU[predCoordUV];
        recoV[coordUV] = predV[predCoordUV];
      }
    }
  }
}

static void createRegion(object *currRegion, int regionIndex, int posX, 
                         int posY, int picSizeX, int regionSize)
{
  currRegion->xmin = posX;
  currRegion->ymin = posY;
  currRegion->xmax = posX + regionSize - 1;
  currRegion->ymax = posY + regionSize - 1;

  currRegion->no_of_pels = (regionSize == 8) ? 64 : 256;
}

static void findMotionPred(int32 *affineDataPred, int regionIndex, 
                           int regionSize, int posX, int posY, 
                           objectBuffer_t *objectList, int picSizeX, 
                           int predDir, int predSel, int refFrame)
{

  int
    predInd,
    numOfNeighs,
    neighIndexes[2],
    mbPerLine = picSizeX/16,
    posNeighX, posNeighY,
    motionNeighs;

  int32
    *affineDataPtr;

  mcpFindNeighsDecoder(objectList, regionIndex, regionSize, predDir, 
                       mbPerLine, posX, posY, &numOfNeighs, neighIndexes,
                       refFrame, &motionNeighs);

  if(predSel == PREDSEL_NOTNEEDED || predSel == PREDSEL_1)
    predInd = neighIndexes[0];
  else
    predInd = neighIndexes[1];
    
  posNeighX = objectList[predInd].xMin;
  posNeighY = objectList[predInd].yMin;

  affineDataPtr = &(objectList[predInd].affineData[0]);

  affineDataPred[0] = (affineDataPtr[0] +
		       affineDataPtr[1]*(posY-posNeighY) +
		       affineDataPtr[2]*(posX-posNeighX));
  affineDataPred[1] = affineDataPtr[1];
  affineDataPred[2] = affineDataPtr[2];

  affineDataPred[3] = (affineDataPtr[3] +
		       affineDataPtr[4]*(posY-posNeighY) +
		       affineDataPtr[5]*(posX-posNeighX));
  affineDataPred[4] = affineDataPtr[4];
  affineDataPred[5] = affineDataPtr[5];

}

int32 DecodePFrame(struct frame **longTermMem, struct frame *predfr,
                   struct frame *recfr, int picSizeX, int picSizeY, int l,
                   objectBuffer_t *objectList, 
                   int16 *slicePredMap,
                   int sliceMode, int qp, int aQP, int numActiveRefFrames,
                   int *orderedRefFrames,
                   int32 *bitsTotalMotion,
                   int32 *bitsTotalDctY, int32 *bitsTotalDctU, 
                   int32 *bitsTotalDctV, int32 *bitsTotalOvrY,
                   int32 *bitsTotalOvrC, 
                   int32 *bitsTotalMode,
                   int32 *cumBitsSliceHeaders,
                   struct frame *outfr, 
                   vdxInstance_t *instance, 
                   int *BitError,
                   int useLoopFilter)
{
  int
    IPB = 0,
    i, j, iPred, jPred, coordUV, predCoordUV, blockInd,
    bitsDct, /*Coding bits of Reg, will be init. in the fnc it's sent */
    bitsDctC,
    bitsOvr, bitsOvrY, bitsOvrC, /*Ovrhead bits of Reg, will be init. in the fnc it's sent */
    bitsTotal = 0;/*The total bits will be returned to the calling function */

  int
    regionsToDecode = 1,
    splittedMacroBlock,
    frameIndex, regionIndex, regionLast,
    mbPosX, mbPosY, mbPerLine, numMBsInFrame,
    numRefFrame = 0; /* Number of the reference frame in longterm memory */

  int
    mbNum, intraSliceFlag, requestNewSlice, noMoreSlices = 0,
    tmp, lbl, sliceXmin, sliceXmax, sliceYmin, sliceYmax, currSliceID = 0,
    sliceCurrBits = *cumBitsSliceHeaders,
    sliceBits = DEFAULT_SLICE_BITS,
    size, size_UV, sizex_UV, sizey_UV,
    regionMode, regionSize,
    pecMode,
    bitsMode,
    bitsMotion,
    motionNeighs;

  u_char
    predPossible, predDirsAvail, predSelNeededUp, predSelNeededLeft, predDir;
    
  struct frame
    *prevfr,
    predBuffer;

  u_char
    predBufferY[256],
    predBufferU[64], 
    predBufferV[64];
  
  object
    currRegion;

  predBuffer.yptr = predBufferY;
  predBuffer.uptr = predBufferU;
  predBuffer.vptr = predBufferV;

  sizex_UV = (int32) (picSizeX >> 1);
  sizey_UV = (int32) (picSizeY >> 1);
  size     = (int32) picSizeX * picSizeY;
  size_UV  = (int32) sizex_UV * sizey_UV;

  numMBsInFrame = (picSizeX >> 4)*(picSizeY >> 4);

  sliceXmin = picSizeX;
  sliceYmin = picSizeY;
  sliceXmax = 0;
  sliceYmax = 0;

  for (tmp = (picSizeX * picSizeY >> 6) - 1; tmp >= 0; tmp--)
    slicePredMap[tmp] = SHRT_MAX;

  if (sliceMode)
  {
    tmp = vdxGetSliceHead(&currSliceID, &intraSliceFlag,
                          NULL, NULL, &picSizeX, &picSizeY, 0, 1,
                          &IPB, instance, (int32 *)BitError);
    if (*BitError == VDX_SAW_SYNC) 
      return (int32)0;

    *cumBitsSliceHeaders += tmp;

    sliceCurrBits = tmp;
  }

  /*
   * Decoder main loop
   */
  *bitsTotalMode = *bitsTotalMotion = *bitsTotalDctY = 0;
  *bitsTotalDctU = *bitsTotalDctV = *bitsTotalOvrY = *bitsTotalOvrC = 0;

  mbPosX = mbPosY = mbNum = 0;
  mbPerLine = picSizeX >> 4;

  while(regionsToDecode){

    if (sliceMode)
    {
      u_int32 currBitPos = bibNumberOfFlushedBits(buff_in);

      /*
       * Check if the bits left in the buffer are sufficient
       * to decode one more macroblock inside the current slice
       */

      if (buff_in->bitsLeft < 8)
      {
        requestNewSlice = 1;

        tmp = buff_in->bitsLeft + buff_in->numEmulPreventBits;

        if (buff_in->bitsLeft > 0)
        {
          /* Detect if all bits left are 0s */

          if (! vdxOnlyZeroBits(instance, (int32 *)BitError))
            requestNewSlice = 0;
          if(*BitError == VDX_SAW_SYNC) return (int32)0;
        }
        else if (buff_in->bitsLeft < 0) {
          *BitError = VDX_SAW_SYNC;
          return (int32)0;
        }

        if (requestNewSlice)
        {

          *cumBitsSliceHeaders += tmp;

          /* Request new slice buffer refill */

          dcmRequestSliceRefill(&buff_in, &noMoreSlices);

          if (noMoreSlices)
          {
            break;
          }
          else
          {
            int
              localFrNum    = l,
              localIPB      = IPB,
              localPicSizeX = picSizeX,
              localPicSizeY = picSizeY;

            tmp = vdxGetSliceHead(&currSliceID, &intraSliceFlag,
                                  &qp, &localFrNum,
                                  &localPicSizeX, &localPicSizeY,
                                  1, 1, &localIPB,
                                  instance, (int32 *)BitError);
            if(*BitError == VDX_SAW_SYNC) return (int32)0;

            if (localFrNum != l || localIPB != IPB ||
                localPicSizeX != picSizeX || localPicSizeY != picSizeY) {
              deb0p("Locals are different than frame globals in P frame");
              nccExit(-1);
            }

            *cumBitsSliceHeaders += tmp;

            sliceCurrBits = tmp;

            fprintf(stdout, "Slice start at P %3d %2d\n", l, currSliceID);
            fprintf(stderr, "Slice start at P %3d %2d\n", l, currSliceID);

            if (currSliceID < mbNum)
            {
              /* Slice of a future frame, perhaps */
              *BitError = VDX_SAW_SYNC;
              return (int32)0;
            }
          }

          sliceXmin = picSizeX;
          sliceYmin = picSizeY;
          sliceXmax = sliceYmax = 0;
        }
      }

      if (currSliceID > mbNum)
      {
        mbNum = currSliceID;
        mbPosY = (div(currSliceID,mbPerLine)).quot;

        mbPosX = currSliceID - mbPosY * mbPerLine;
        mbPosY <<= 4;
        mbPosX <<= 4;
      }
    } /* End of slice mode */

    regionIndex = ((mbPosX >> 4) + (mbPosY >> 4) * mbPerLine) << 2;

    objectList[regionIndex].sliceID = currSliceID;

    /* Is prediction possible */
    mcpFindPredAlternativesDecoder(objectList, regionIndex, 16, mbPosX, 
                                   mbPosY, mbPerLine, &predPossible, 
                                   &predDirsAvail, &predDir, 
                                   &predSelNeededUp, &predSelNeededLeft,
                                   0, &motionNeighs);


    /* This is needed for the long-term reference memory case. Motion
     * parameter prediction is possible with any neighbor with motion
     * information since we don't know yet which reference frame is
     * to be used.
     */
    if(motionNeighs)
      predPossible = 1;

    /* Region mode for 16x16 macroblock */
    bitsMode = vdxGetRegionMode(&regionMode, &pecMode, predPossible, 16,
                                instance, (int32*)BitError);

    if(*BitError == VDX_SAW_SYNC) return 0;

    /* Reference frame indicator */
    if(decLongTermMemSize > 1){
      if(numActiveRefFrames > 1 && regionMode != REGMODE_INTRA){
        bitsMode += vdxGetLongtermRVLC((int32 *)&frameIndex, instance, (int32 *)BitError);
        if(*BitError == VDX_SAW_SYNC) return 0;
        numRefFrame = orderedRefFrames[frameIndex];
      }
      else
        numRefFrame = orderedRefFrames[0];
    }
    else{
      numRefFrame = 0;
    }

    /* Update motion parameter prediction info if reference frame
     * number zero was not chosen
     */
    if(numRefFrame != 0)
      mcpFindPredAlternativesDecoder(objectList, regionIndex, 16, mbPosX, 
                                     mbPosY, mbPerLine, &predPossible, 
                                     &predDirsAvail, &predDir, 
                                     &predSelNeededUp, &predSelNeededLeft,
                                     numRefFrame, &motionNeighs);

    objectList[regionIndex].refFrame = numRefFrame;

    prevfr = longTermMem[numRefFrame];

    *bitsTotalMode += bitsMode;

    sliceCurrBits += bitsMode;

    currRegion.regionMode = objectList[regionIndex].regionMode = regionMode;
    objectList[regionIndex].pecMode = pecMode;

    if(regionMode == REGMODE_SPLITTED)
    {
      /*
       * Macroblock is splitted.
       */
      regionLast         = regionIndex + 4;
      blockInd           = 0;
      splittedMacroBlock = 1;
    }
    else
    {
      /* 
       * Create new 16x16 region.
       */

      createRegion(&currRegion, regionIndex, mbPosX, mbPosY, picSizeX, 16);

      currRegion.pecMode           = pecMode;
      currRegion.predPossible      = predPossible;

      currRegion.predDirsAvail     = predDirsAvail;
      currRegion.predDir           = predDir;
      currRegion.predSelNeededUp   = predSelNeededUp;
      currRegion.predSelNeededLeft = predSelNeededLeft;

      regionLast                   = regionIndex + 1;
      splittedMacroBlock           = 0;
    }

    /* 
     * Decode macroblock (one 16x16 region or four 8x8 regions).
     */
    while(regionIndex < regionLast){

      /* 
       * Read 8x8 header if root MB was splitted.
       */
      if(splittedMacroBlock){

        /* 
         * Create new 8x8 region.
         */
        createRegion(&currRegion, regionIndex,
                     mbPosX + ((blockInd & 1) << 3), 
                     mbPosY + ((blockInd & 2) << 2), picSizeX, 8);

        blockInd++;

        /* Is prediction possible */
        mcpFindPredAlternativesDecoder(objectList, regionIndex, 8, 
                                       currRegion.xmin, currRegion.ymin,
                                       mbPerLine, &predPossible,
                                       &predDirsAvail, &predDir, 
                                       &predSelNeededUp, &predSelNeededLeft,
                                       numRefFrame, &motionNeighs);

        /* Region mode for 8x8 block */
        bitsMode = vdxGetRegionMode(&regionMode, &pecMode, predPossible, 8,
                                    instance, (int32*)BitError);

        regionMode += REGMODE_INTER_NOPRED_8x8;
        if(*BitError == VDX_SAW_SYNC) return 0;

        *bitsTotalMode += bitsMode;

        sliceCurrBits += bitsMode;

        currRegion.regionMode = objectList[regionIndex].regionMode 
                              = regionMode;

        objectList[regionIndex].pecMode  = pecMode;
        objectList[regionIndex].refFrame = numRefFrame;

        currRegion.pecMode = pecMode;
        currRegion.predPossible = predPossible;

        currRegion.predDirsAvail     = predDirsAvail;
        currRegion.predDir           = predDir;
        currRegion.predSelNeededUp   = predSelNeededUp;
        currRegion.predSelNeededLeft = predSelNeededLeft;

      }

      regionSize = (currRegion.no_of_pels == 64) ? 8 : 16;

      objectList[regionIndex].xMin = currRegion.xmin;
      objectList[regionIndex].yMin = currRegion.ymin;

      tmp = (currRegion.ymin >> 3) * (picSizeX >> 3) + (currRegion.xmin >> 3);

      switch (regionMode) {

      case REGMODE_INTER_NOPRED:
      case REGMODE_INTER_NOPRED_8x8:
      case REGMODE_INTER_PRED:
      case REGMODE_INTER_PRED_8x8:
      case REGMODE_INTER_PRED_DELTA:
      case REGMODE_INTER_PRED_DELTA_8x8:
      case REGMODE_INTER_COPY:
      case REGMODE_INTER_COPY_8x8:

        if (regionSize == 16)
          slicePredMap[tmp] =
          slicePredMap[tmp + 1] =
          slicePredMap[tmp + (picSizeX >> 3)] =
          slicePredMap[tmp + (picSizeX >> 3) + 1] = currSliceID;
        else
          slicePredMap[tmp] = currSliceID;

        /************** BUILD MOTION COMPENSATED PREDICTION ************/

        bitsMotion = 0;

        /* Figure out motion prediction */
        if(regionMode == REGMODE_INTER_PRED       ||
           regionMode == REGMODE_INTER_PRED_8x8   ||
           regionMode == REGMODE_INTER_PRED_DELTA ||
           regionMode == REGMODE_INTER_PRED_DELTA_8x8) {

          int predSel = PREDSEL_NOTNEEDED;

          bitsMotion += vdxGetMotionPredInfo(currRegion.predDirsAvail,
                                             currRegion.predSelNeededUp,
                                             currRegion.predSelNeededLeft,
                                             &currRegion.predDir,
                                             &predSel, instance, (int32 *)BitError);

          findMotionPred(currRegion.affineDataPred, regionIndex, regionSize,
                         currRegion.xmin, currRegion.ymin, objectList, 
                         picSizeX, currRegion.predDir, predSel, numRefFrame);

          if(*BitError == VDX_SAW_SYNC) return 0;

        }

        if(regionMode == REGMODE_INTER_PRED_DELTA     ||
           regionMode == REGMODE_INTER_PRED_DELTA_8x8 ||
           regionMode == REGMODE_INTER_NOPRED         ||
           regionMode == REGMODE_INTER_NOPRED_8x8)
          bitsMotion += vdxGetMotionCoeffs(currRegion.motCoefOrth, aQP,
                                           instance, (int32 *)BitError);

        *bitsTotalMotion += bitsMotion;

        sliceCurrBits += bitsMotion;

        if(*BitError == VDX_SAW_SYNC) return 0;

        mcpCalcFinalParams(&currRegion);
        
        mcpBuildSegmPred(picSizeX, picSizeY, prevfr->yptr, 
                         prevfr->uptr, prevfr->vptr, &currRegion,
                         predBuffer.yptr, predBuffer.uptr, predBuffer.vptr);
        memcpy(objectList[regionIndex].affineData, 
               currRegion.affineData, 6 * sizeof(int32));

        if (predfr->yptr)
        {
          for (i = currRegion.ymin, iPred = 0; i <= currRegion.ymax; i++, 
               iPred++)
          {
            for (j = currRegion.xmin, jPred = 0; j <= currRegion.xmax; j++, 
                 jPred++)
            {
              predfr->yptr[i * picSizeX + j] = predBuffer.yptr[(iPred << 4) + 
                                               jPred];
              if(!(i & 1) && !(j & 1)) 
              {
                coordUV = i * (picSizeX >> 2) + (j >> 1);
                predCoordUV = (iPred << 2) + (jPred >> 1);
                predfr->uptr[coordUV] = predBuffer.uptr[predCoordUV];
                predfr->vptr[coordUV] = predBuffer.vptr[predCoordUV];
              }
            }
          }

        }
        /****************   DECODING OF PREDICTION ERROR ***************/

        if (pecMode == PECMODE_CODED) {
          decDecodeMccRegion(&predBuffer, recfr, &currRegion, regionSize,
                             regionMode, picSizeX, picSizeY, qp,
                             &bitsDct, &bitsDctC, &bitsOvr,
                             instance, BitError);

          if(*BitError == VDX_SAW_SYNC) return 0;

          *bitsTotalDctY += bitsDct;
          *bitsTotalDctU += bitsDctC;
          *bitsTotalOvrY += bitsOvr;

          sliceCurrBits += bitsDct + bitsDctC + bitsOvr;
        }
        else {
          /* in decoder an 8x8 block in the macroblock is always at (0,0) offset */
          pedRecoCopyRegionYUV(&predBuffer, recfr, &currRegion, picSizeX, 0);
        }

        break;

      case REGMODE_INTRA:
      case REGMODE_INTRA_8x8:

        if (predfr->yptr)
        {
          for (i = currRegion.ymin, iPred = 0; i <= currRegion.ymax; i++, 
               iPred++)
          {
            for (j = currRegion.xmin, jPred = 0; j <= currRegion.xmax; j++, 
                 jPred++)
            {
              predfr->yptr[i * picSizeX + j] = 0;
              if(!(i & 1) && !(j & 1)) 
              {
                coordUV = i * (picSizeX >> 2) + (j >> 1);
                predCoordUV = (iPred << 2) + (jPred >> 1);
                predfr->uptr[coordUV] = 0;
                predfr->vptr[coordUV] = 0;
              }
            }
          }
        }
        lbl = -(currSliceID+1);

        if (regionSize == 16)
          slicePredMap[tmp] =
          slicePredMap[tmp + 1] =
          slicePredMap[tmp + (picSizeX >> 3)] =
          slicePredMap[tmp + (picSizeX >> 3) + 1] = lbl;
        else
          slicePredMap[tmp] = lbl;

        /* Update the geometry of the intra part of
           slice which developped until now */

        if (currRegion.ymin < sliceYmin)
          sliceYmin = currRegion.ymin;
        if (currRegion.xmin < sliceXmin)
          sliceXmin = currRegion.xmin;
        if (currRegion.ymin + regionSize - 1 > sliceYmax)
          sliceYmax = currRegion.ymin + regionSize - 1;
        if (currRegion.xmin + regionSize - 1 > sliceXmax)
          sliceXmax = currRegion.xmin + regionSize - 1;

        decDecodeIntraMccRegion(recfr, slicePredMap, lbl,
                                sliceXmin, sliceXmax, sliceYmin, sliceYmax,
                                &currRegion, regionSize,
                                picSizeX, picSizeY, INTRA_BLOCK, qp,
                                &bitsDct, &bitsDctC, &bitsOvrY, &bitsOvrC,
                                instance, BitError);
        bitsOvr = bitsOvrY + bitsOvrC;

        if(*BitError == VDX_SAW_SYNC) return 0;

        *bitsTotalDctY += bitsDct;
        *bitsTotalDctU += bitsDctC;
        *bitsTotalOvrY += bitsOvr;

        sliceCurrBits += bitsDct + bitsDctC + bitsOvr;

        break;

      default :

        deb1p("Unknown regionMode %d in DecodePFrame()\n", regionMode);
        nccExit(1);

        break;
      }

      regionIndex++;
    }

    /* Take the next Macro Block */
    if(mbPosX + 16 < picSizeX)
      mbPosX += 16;
    else if(mbPosY + 16 < picSizeY){
      mbPosX  =  0;
      mbPosY += 16;
    }
    else
      regionsToDecode = 0;

    mbNum++;
  }

 if (! noMoreSlices)		   
  {
    *cumBitsSliceHeaders += buff_in->bitsLeft + buff_in->numEmulPreventBits;
    dcmRequestSliceRefill(&buff_in, &noMoreSlices);
  }

  /*
   * Apply deblocking filter.
   */

  filCompFilt(recfr->yptr, picSizeX, picSizeY, qp, outfr->yptr, 0, 0, NULL, objectList,
              useLoopFilter);

  if(useLoopFilter){
    for (i = 0; i < picSizeY*picSizeX; i++) {
      (recfr->yptr)[i] = (outfr->yptr)[i];
    }
  }

  for (i = 0; i < picSizeY * (picSizeX >> 2); i++) {
    (outfr->uptr)[i] = (recfr->uptr)[i];
    (outfr->vptr)[i] = (recfr->vptr)[i];
  }

  bitsTotal = *bitsTotalMotion +
              *bitsTotalDctY   + *bitsTotalDctU  + *bitsTotalDctV  +
              *bitsTotalOvrY   + *bitsTotalOvrC  + *bitsTotalMode;

  return bitsTotal;
}

/*
 *
 * pedLoadMethod:
 *
 * Parameters:
 *      blockClass      I:  Class of the block
 *      codingMethod    O:  Coding method for block
 *      blkSize         I:  Size of block (SIZE_8x8 or SIZE_4x4)
 *      chroma          I:  Flag for luminance and chrominance blocks
 *      mode            I:  Mode of block (intra/inter)
 *      qp              I:  Quantization parameter
 *
 * Function:
 *      Loads the coding method for 8x8 INTER Y, u, or v blocks.
 *
 * Returns:
 *      Number of bits needed
 */

int32 pedLoadMethod(int32 class, int32 *codingMethod, 
                    int32 chroma, int32 blkSize,
                    vdxInstance_t *instance, int* BitError)
{
  int32
    ovhNo, ind,
    bits;                /* Number of used bits */

  if (blkSize == SIZE_8x8) {

    /*
     * Select the VLC codebook according to class of the block and
     * Decode the index of the 8x8 coding method.
     */
    ovhNo = overheadNumbers8x8[class];
    bits = vdxGetCodingMethod(instance, ovhNo, &ind, (int32*)BitError);
    if (*BitError == VDX_SAW_SYNC) 
      return (int32)0;

    /*
     * Find the actual 8x8 coding method from the coding method subset.
     */
    *codingMethod = overheadOrder8x8[class * (MAX_OF_8x8_METH_OVH + 1) + ind];
  }
  else {

    /*
     * Select the VLC codebook according to class of the block and
     * Decode the index of the 8x8 coding method.
     */
    ovhNo = overheadNumbers4x4[class];
    bits = vdxGetCodingMethod(instance, ovhNo, &ind, (int32*)BitError);
    if (*BitError == VDX_SAW_SYNC) 
      return (int32)0;

    /*
     * Find the actual 8x8 coding method from the coding method subset.
     */
    *codingMethod = overheadOrder4x4[class * (MAX_OF_4x4_METH_OVH + 1) + ind];
  }

  return bits;
}



/*
 *
 * pedLoad4x4Info:
 *
 * Parameters:
 *      methCoeffs      O:  Coding information for the block
 *      chroma          I:  Flag for luminance and chrominance blocks
 *      mode            I:  Flag for intra and inter blocks
 *      blkSize         I:  Size of the coded block
 *      codingMethod    I:  Coding method of the block
 *
 * Function:
 *      Loads the information for those 4x4 INTER Y, u, or v blocks
 *      which are identified to be coded by the class of the 8x8 block 
 *      they belong.
 *
 * Returns:
 *      Number of bits needed
 */

int32 pedLoad4x4Info(int32 *methCoeffs, int32 chroma, int32 mode,
                     int32 qp, int32 blkSize_dummy, int32 codingMethod,
                     vdxInstance_t *instance, int* BitError)
{
  int32 
    i,                        /* Loop counter */
    class4x4,                 /* Orientation of the block */
    method4x4,                /* Codebook parameters */
    codedBlocks,              /* Number of coded 4x4 blocks */
    start = 0,                /* Starting point of block parameters */
    bits = 0;                 /* Number of bits */

  codedBlocks = codingMethod == METH_4x4_15 ? 4 : 1;

  for (i = 0; i < codedBlocks; i++) {

    /*
     * Decode 4x4 coding method according to class of the 4x4 block.
     */
    class4x4 = methCoeffs[start + 13];
    bits += pedLoadMethod(class4x4, &method4x4, chroma, SIZE_4x4,
                          instance, BitError);
    if (*BitError == VDX_SAW_SYNC) 
      return (int32)0;

    /*
     * Store the 4x4 coding method.
     */
    methCoeffs[start + 10] = method4x4;
    start += SIZE_4x4;
  }

  return bits;
}


/*
 *
 * pedLoadDctCoeffs:
 *
 * Parameters:
 *      methCoeffs      O:  Coding information for the block
 *      codingMethod    I:  Coding method of the block
 *      chroma          I:  Flag for luminance and chrominance blocks
 *      mode            I:  Mode of block (intra/inter)
 *
 * Function:
 *      Loads the DCT information for those 4x4 INTER Y, u, or v blocks
 *      which are identified to be coded by the class of the 8x8 block 
 *      they belong.
 *
 * Returns:
 *      Number of bits needed
 */

int32 pedLoadDctCoeffs(int32 *methCoeffs, int32 codingMethod, int32 chroma,
                       int32 qp, int32 mode, int *save4, 
                       vdxInstance_t *instance, int* BitError)
{
  int32 
    num,              /* Auxiliary variable */
    add,              /* Auxiliary variable */ 
    size,             /* Size of the block */
    codedBlocks,      /* Number of coded blocks */
    start,            /* Starting point of block parameters */
    last, run, level, /* Triplet parameters */
    bits = 0,         /* Number of bits */
    j = 0, method4x4 = 0,
    levelBits, runBits, tableNo;

  /*
   * Find the transform size for the coding method.
   */
  switch(codingMethod) {
    case METH_8x8_DCT:

    case METH_DCT_8x8_N_0_8: case METH_DCT_8x8_N_2_8:
    case METH_DCT_8x8_N_4_8:

    case METH_KLT_8x8_N_1_8: case METH_KLT_8x8_N_2_8: 
    case METH_KLT_8x8_N_3_8: case METH_KLT_8x8_N_5_8:
    case METH_KLT_8x8_N_6_8: case METH_KLT_8x8_N_7_8:
      size = 64;
      break;

    case METH_4x8_DCT_1: case METH_4x8_DCT_2:
    case METH_4x8_DCT_3: case METH_4x8_DCT_4:
    case METH_8x4_DCT_1: case METH_8x4_DCT_2:
    case METH_8x4_DCT_3: case METH_8x4_DCT_4:
      size = 32;
      break;

    default:
      size = 16;
      break;
  }

  /*
   * Find the amount of transform coded blocks and start of the
   * coefficient data.
   */
  codedBlocks = findNumMethods(codingMethod, size);
  start = find8x8Start(codingMethod, size);

  /*
   * If 4x4 blocks are coded separately, find
   * 4x4 coding method.
   */
  if (codingMethod >= METH_4x4_DCT_1 && codingMethod <= METH_4x4_DCT_15)
    method4x4 = save4[j++];
  
  /*
   * Find the appropriate VLC codebook for coefficients.
   */
  findTripletInfos(codingMethod, method4x4, qp, chroma, mode, (int *)&tableNo,
                   (int *)&runBits, (int *)&levelBits);
  add = num = 0;
  for (;;) {

    /*
     * Decode one LAST-RUN-LEVEL event.
     */
    bits += vdxGetTriplet(instance, tableNo, levelBits,
                          runBits, &last, &run, &level,
                          (int32*)BitError);
    if (*BitError == VDX_SAW_SYNC) 
      return (int32)0;

    /*
     * Perform inverse Run-Length coding.
     */
    methCoeffs[start + add + run] = level;
    add += run + 1;

    if (last) {
      num++;

      /*
       * Break the loop if all transform coded blocks are
       * processed.
       */
      if (num == codedBlocks)
        break;


      /*
       * If 4x4 blocks are coded separately, skip 4x4 blocks that
       * are not coded using transform coding.
       */
      if (codingMethod == METH_4x4_DCT_9)
        start += 3 * size;

      else if (codingMethod==METH_4x4_DCT_5 || codingMethod==METH_4x4_DCT_10 || 
               (num == 2 && codingMethod == METH_4x4_DCT_11) || 
               (num == 1 && codingMethod == METH_4x4_DCT_13)) {
        start += size << 1;
      }
      else
        start += size;

      add = 0;

        /*
         * If 4x4 blocks are coded separately, find 4x4 coding method
         * and appropriate VLC codebook for coefficients.
         */
      if (codingMethod >= METH_4x4_DCT_1 && codingMethod <= METH_4x4_DCT_15) {
        method4x4 = save4[j++];

        findTripletInfos(codingMethod, method4x4, qp, chroma, mode, 
                         (int *)&tableNo, (int *)&runBits, (int *)&levelBits);
      }
    }
  }

  return bits;
}

/* qpBlockLuma used to pass qpY to vdxGetDctTable() and vdxGetKltTable() */
extern int
  qpBlockLuma;

static void decDecodeMccRegion(struct frame *predFr, struct frame *recoFr,
                               object *region, int32 regionSize, 
                               int32 regionMode,
                               int32 sizex, int32 sizey, int32 qp,
                               int *bits_reg_mcc, int *bits_reg_mcc_uv,
                               int *bits_reg_ovh,
                               vdxInstance_t *instance, int *BitError)
{
  u_char
    *predY = predFr->yptr,
    *predU = predFr->uptr,
    *predV = predFr->vptr,
    *recoY = recoFr->yptr,
    *recoU = recoFr->uptr,
    *recoV = recoFr->vptr;

  u_char
    cnc_y[4], cnc_u, cnc_v;

  int32
    x,y,k,l,p,q, tmp = 0, qp_UV, xPred, yPred,
    sxmin, sxmax, symax, symin,
    sizex_ch = sizex >> 1, sizey_ch = sizey >> 1,
    regSx,
    class,
    methCoeffs[64], dctCoeffs[64],
    PredBlk[64], DiffBlk[64],
    rotateFlag = UNIT_OPER,
    bits_blk_mcc,
    bit_for_dct,  /* bits spent by DCTs */
    bit_for_mes,  /* bits spent for method selection */ 
    ExtBlk[16], save4[4], bit_for_4x4info,
    codingMethod, dct4x4Methods,
    pels; /* Amount of pels in extrapSource */
  
  *bits_reg_mcc = *bits_reg_mcc_uv = *bits_reg_ovh = 0;

  sxmin = region->xmin;
  sxmax = region->xmax;
  symax = region->ymax;
  symin = region->ymin;
  regSx = regionSize >> 3;

  /* qpBlockLuma used to pass qpY to vdxGetDctTable() and vdxGetKltTable() */
  qpBlockLuma = qp;

  qp_UV = annexTlookup[qp];

  /*
   * Decode CNC information.
   */
  *bits_reg_ovh = vdxDctGetCncInfo(cnc_y, &cnc_u, &cnc_v,
                                   regionMode, regionSize,
                                   instance, (int32*)BitError);
  if (*BitError == VDX_SAW_SYNC)
    return;

  /* 
   * Scan the 8x8 luminance block(s). 
   */
  for (y = symin, yPred = 0; y < symax; y += 8, yPred += 8) {
    for (x = sxmin, xPred = 0; x < sxmax; x += 8, xPred += 8) {

      /*
       * Copy prediction block from the prediction frame.
       */
      for (k = 0; k < 8; k++) {
        p = k << 3;
        q = ((yPred + k) << 4) + xPred;
        for (l = 0; l < 8; l++, p++, q++) 
          PredBlk[p] = predY[q];
      }

      /*
       * Check if block is coded.
       */
      if (cnc_y[tmp]) {

        bits_blk_mcc = bit_for_mes = bit_for_4x4info = bit_for_dct = 0;
        for (k = 0; k < 64; k++) 
          methCoeffs[k] = dctCoeffs[k]  = 0;

        /*
         * Classify the block according to QP for the
         * 8x8 coding method subset selection and
         * decode the 8x8 coding method information.
         */
        class = changeClassNew(0, 0, qp, LUMA, INTER_BLOCK, 
                               (int *)&rotateFlag);

        bit_for_mes = pedLoadMethod(class, &codingMethod, LUMA,
                                    SIZE_8x8, instance, BitError);
        if (*BitError == VDX_SAW_SYNC) 
          return;

        /*
         * If 8x8 method signals that 4x4 Quadrants are coded separately,
         * classify the Quadrants for 4x4 coding method subset selection.
         */
        if (codingMethod == METH_4x4_15) {
          findPredNums8x8((int *)PredBlk, rotateFlag, qp, (int *)methCoeffs,
                          LUMA, codingMethod, INTER_BLOCK);

          /*
           * Decode 4x4 coding method information.
           */
          bit_for_4x4info = pedLoad4x4Info(methCoeffs, LUMA, INTER_BLOCK,
                                           qp, SIZE_8x8, codingMethod,
                                           instance, BitError);
          if (*BitError == VDX_SAW_SYNC) 
            return;

          /*
           * Find Quadrants that are coded with DCT or KLT.
           * If any of the Quadrants is coded with 4x4 DCT or 4x4 KLT, 
           * decode transform coefficient information.
           */
          dct4x4Methods = findDctMeth((int *)methCoeffs, codingMethod, 
                                      (int *)save4);

          if (dct4x4Methods >= METH_4x4_DCT_1) {
            bit_for_dct = pedLoadDctCoeffs(dctCoeffs, dct4x4Methods, LUMA,
                                           qp, INTER_BLOCK, (int *)save4,
                                           instance, BitError);
            if (*BitError == VDX_SAW_SYNC) 
              return;
          }
        }

        /*
         * If 8x8 block is coded with 8x8 DCT, 8x4 DCT or KLT, decode 
         * transform coefficient information.
         */
        else if ((codingMethod >= METH_8x8_DCT &&
                  codingMethod <= METH_8x4_DCT_4) ||
                 (codingMethod >= METH_KLT_8x8_N_0_8 &&
                  codingMethod <= METH_KLT_8x8_N_7_8) ||
                 (codingMethod >= METH_DCT_8x8_N_0_8 &&
                  codingMethod <= METH_DCT_8x8_N_4_8)) {
          bit_for_dct = pedLoadDctCoeffs(methCoeffs, codingMethod, LUMA,
                                         qp, INTER_BLOCK, NULL,
                                         instance, BitError);
          if (*BitError == VDX_SAW_SYNC) 
            return;
        }

        /*
         * Reconstruct the 8x8 prediction error block.
         */
        bits_blk_mcc = dec8x8((int *)PredBlk, (int *)DiffBlk, 
                              (int *)methCoeffs, (int *)dctCoeffs, qp,
                              codingMethod, class, LUMA, INTER_BLOCK, 0);

        *bits_reg_mcc += bit_for_mes + bit_for_4x4info + bit_for_dct;
      }

      /*
       * Create final reconstruction block.
       */
      pedRecoInterBlk((int32*)PredBlk, (int32*)DiffBlk, recoY, y, x, sizex,
                      8, cnc_y[tmp]);
      tmp++;
    }
  }
  

  /* 
   * Decode chrominance blocks.
   */
  x = sxmin >> 1; 
  y = symin >> 1;

  switch (regionSize) {  
    
    /*
     * 8x8 chrominance block of 16x16 macroblock.
     */
    case 16:

      /*
       * Copy prediction block for U component from the prediction frame.
       */
      for (k = 0; k < 8; k++) {
        p = k << 3;
        for (l = 0; l < 8; l++, p++)  
          PredBlk[p] = (int)(predU[p]);
      }

      /*
       * Check if U block is coded.
       */
      if (cnc_u) {
        bits_blk_mcc = bit_for_mes = bit_for_4x4info = bit_for_dct = 0;

        for (k = 0; k < 64; k++) 
          methCoeffs[k] = dctCoeffs[k]  = 0;

        /*
         * Classify the block according to QP for the
         * 8x8 coding method subset selection and
         * decode the 8x8 coding method information.
         */
        class = changeClassNew(0, 0, qp_UV, CHROMA, INTER_BLOCK, 
                               (int *)&rotateFlag);
        
        bit_for_mes = pedLoadMethod(class, &codingMethod, CHROMA,
                                    SIZE_8x8, instance, BitError);
        if (*BitError == VDX_SAW_SYNC) 
          return;
        
        /*
         * If 8x8 method signals that 4x4 Quadrants are coded separately,
         * classify the Quadrants for 4x4 coding method subset selection.
         */
        if (codingMethod == METH_4x4_15) {
          findPredNums8x8((int *)PredBlk, rotateFlag, qp_UV, 
                          (int *)methCoeffs, CHROMA,
                          codingMethod, INTER_BLOCK);
          
          /*
           * Decode 4x4 coding method information.
           */
          bit_for_4x4info = pedLoad4x4Info(methCoeffs, CHROMA, INTER_BLOCK,
                                           qp_UV, SIZE_8x8, codingMethod,
                                           instance, BitError);
          if (*BitError == VDX_SAW_SYNC) 
            return;
          
          /*
           * Find Quadrants that are coded with DCT or KLT.
           * If any of the Quadrants is coded with 4x4 DCT or 4x4 KLT, 
           * decode transform coefficient information.
           */
          dct4x4Methods = findDctMeth((int *)methCoeffs, codingMethod, 
                                      (int *)save4);
          
          if (dct4x4Methods >= METH_4x4_DCT_1) {
            bit_for_dct = pedLoadDctCoeffs(dctCoeffs, dct4x4Methods, CHROMA,
                                           qp_UV, INTER_BLOCK, (int *)save4,
                                           instance, BitError);
            if (*BitError == VDX_SAW_SYNC) 
              return;
          }
        }

        /*
         * If 8x8 block is coded with 8x8 DCT, 8x4 DCT or KLT, decode 
         * transform coefficient information.
         */
        else if ((codingMethod >= METH_8x8_DCT &&
                  codingMethod <= METH_8x4_DCT_4) ||
                 (codingMethod >= METH_KLT_8x8_N_0_8 &&
                  codingMethod <= METH_KLT_8x8_N_7_8) ||
                 (codingMethod >= METH_DCT_8x8_N_0_8 &&
                  codingMethod <= METH_DCT_8x8_N_4_8)) {
          bit_for_dct = pedLoadDctCoeffs(methCoeffs, codingMethod, CHROMA,
                                         qp_UV, INTER_BLOCK, NULL,
                                         instance, BitError);
          if (*BitError == VDX_SAW_SYNC) 
            return;
        }
        
        /*
         * Reconstruct the 8x8 prediction error block.
         */
        bits_blk_mcc = dec8x8((int *)PredBlk, (int *)DiffBlk, 
                              (int *)methCoeffs, (int *)dctCoeffs,qp_UV,
                              codingMethod, class, CHROMA, INTER_BLOCK, 0);
      
        *bits_reg_mcc_uv += bit_for_mes + bit_for_4x4info + bit_for_dct;      
      }

      /*
       * Create final reconstruction block.
       */
      pedRecoInterBlk((int32*)PredBlk, (int32*)DiffBlk, recoU, y, x,
                      sizex_ch, 8, cnc_u);
    
      /*
       * Copy prediction block for V component from the prediction frame.
       */
      for (k = 0; k < 8; k++) {
        p = k << 3;
        for (l = 0; l < 8; l++, p++)  
          PredBlk[p] = (int)(predV[p]);
      }
    
      /*
       * Check if V block is coded.
       */
      if (cnc_v) {
        bits_blk_mcc = bit_for_mes = bit_for_4x4info = bit_for_dct = 0;

        for (k = 0; k < 64; k++) 
          methCoeffs[k] = dctCoeffs[k] = 0;

        /*
         * Classify the block according to QP for the
         * 8x8 coding method subset selection and
         * decode the 8x8 coding method information.
         */
        class = changeClassNew(0, 0, qp_UV, CHROMA, INTER_BLOCK, 
                                  (int *)&rotateFlag);

        bit_for_mes = pedLoadMethod(class, &codingMethod, CHROMA,
                                    SIZE_8x8, instance, BitError);
        if (*BitError == VDX_SAW_SYNC) 
          return;
        
        /*
         * If 8x8 method signals that 4x4 Quadrants are coded separately,
         * classify the Quadrants for 4x4 coding method subset selection.
         */
        if (codingMethod == METH_4x4_15) {
          findPredNums8x8((int *)PredBlk, rotateFlag, qp_UV, 
                          (int *)methCoeffs, CHROMA,
                          codingMethod, INTER_BLOCK);
        
          /*
           * Decode 4x4 coding method information.
           */
          bit_for_4x4info = pedLoad4x4Info(methCoeffs, CHROMA, INTER_BLOCK,
                                           qp_UV, SIZE_8x8, codingMethod,
                                           instance, BitError);
          if (*BitError == VDX_SAW_SYNC) 
            return;
        
          /*
           * Find Quadrants that are coded with DCT or KLT.
           * If any of the Quadrants is coded with 4x4 DCT or 4x4 KLT, 
           * decode transform coefficient information.
           */
          dct4x4Methods = findDctMeth((int *)methCoeffs, codingMethod, 
                                      (int *)save4);
        
          if (dct4x4Methods >= METH_4x4_DCT_1) {
            bit_for_dct = pedLoadDctCoeffs(dctCoeffs, dct4x4Methods, CHROMA,
                                           qp_UV, INTER_BLOCK, (int *)save4,
                                           instance, BitError);
            if (*BitError == VDX_SAW_SYNC) 
              return;
          }
        }

        /*
         * If 8x8 block is coded with 8x8 DCT, 8x4 DCT or KLT, decode 
         * transform coefficient information.
         */
        else if ((codingMethod >= METH_8x8_DCT &&
                  codingMethod <= METH_8x4_DCT_4) ||
                 (codingMethod >= METH_KLT_8x8_N_0_8 &&
                  codingMethod <= METH_KLT_8x8_N_7_8) ||
                 (codingMethod >= METH_DCT_8x8_N_0_8 &&
                  codingMethod <= METH_DCT_8x8_N_4_8)) {
          bit_for_dct = pedLoadDctCoeffs(methCoeffs, codingMethod, CHROMA,
                                         qp_UV, INTER_BLOCK, NULL,
                                         instance, BitError);
          if (*BitError == VDX_SAW_SYNC) 
            return;
        }
      
        /*
         * Reconstruct the 8x8 prediction error block.
         */
        bits_blk_mcc = dec8x8((int *)PredBlk, (int *)DiffBlk, 
                              (int *)methCoeffs, (int *)dctCoeffs,qp_UV,
                              codingMethod, class, CHROMA, INTER_BLOCK, 0);

        *bits_reg_mcc_uv += bit_for_mes + bit_for_4x4info + bit_for_dct;     
      }
    
      /*
       * Create final reconstruction block.
       */
      pedRecoInterBlk((int32*)PredBlk, (int32*)DiffBlk, recoV, y, x,
                      sizex_ch, 8, cnc_v);
      break;
    
    
    /*
     * 4x4 chrominance block of splited 16x16 macroblock.
     */
    case 8:
    
      /*
       * Copy prediction block for U component from the prediction frame.
       */
      for (k = 0; k < 4; k++) {
        p = k << 2;
        q = k << 3;        
        for (l = 0; l < 4; l++, p++, q++)
          PredBlk[p] = (int)(predU[q]);
      }

      /*
       * Check if U block is coded.
       */
      if (cnc_u) { 
        bits_blk_mcc = bit_for_4x4info = bit_for_dct = 0;
        
        /*
         * Classify the block for 4x4 coding method subset selection.
         */
        class = findPredNum4x4((int *)PredBlk, qp_UV, CHROMA, INTER_BLOCK,
                               SIZE_4x4, 0, (int *)&rotateFlag);
        
        for (k = 0; k < 64; k++) 
          methCoeffs[k] = dctCoeffs[k] = 0;

        /*
         * Store class information. METH_4x4_1 signals that there is only
         * one 4x4 block.
         */
        methCoeffs[13] = class;
        methCoeffs[14] = rotateFlag;
        codingMethod = METH_4x4_1;
        
        /*
         * Decode 4x4 coding method information.
         */
        bit_for_4x4info = pedLoad4x4Info(methCoeffs, CHROMA, INTER_BLOCK,
                                         qp_UV, SIZE_4x4, codingMethod,
                                         instance, BitError);
        if (*BitError == VDX_SAW_SYNC) 
          return;
        
        /*
         * If block is coded with 4x4 DCT or 4x4 KLT, 
         * decode transform coefficient information.
         */
        dct4x4Methods = findDctMeth((int *)methCoeffs, codingMethod, 
                                    (int *)save4);
        
        if (dct4x4Methods >= METH_4x4_DCT_1) {
          bit_for_dct = pedLoadDctCoeffs(dctCoeffs, dct4x4Methods, CHROMA,
                                         qp_UV, INTER_BLOCK, (int *)save4,
                                         instance, BitError);
          if (*BitError == VDX_SAW_SYNC) 
            return;
        }
        
        /*
         * If block is coded with extrapolation, extract the
         * reference pixels.
         */
        if (methCoeffs[10] == EXTRAPOL_METH) 
          pels = getExtSouJunct((int *)PredBlk, (int *)ExtBlk, 
                                COLS_4x4, ROWS_4x4);
        
        /*
         * Reconstruct the 4x4 prediction error block.
         */
        bits_blk_mcc = dec4x4((int *)PredBlk, (int *)DiffBlk, 
                              (int *)methCoeffs, (int *)dctCoeffs,
                              qp_UV, (int *)ExtBlk, pels, CHROMA, INTER_BLOCK,
                              SIZE_4x4);
                
        *bits_reg_mcc_uv += bit_for_4x4info + bit_for_dct;       
      }
      
      /*
       * Create final reconstruction block.
       */
      pedRecoInterBlk((int32*)PredBlk, (int32*)DiffBlk, recoU, y, x,
                      sizex_ch, 4, cnc_u);
      
      /*
       * Copy prediction block for V component from the prediction frame.
       */
      for (k = 0; k < 4; k++) {
        p = k << 2;
        q = k << 3;
        for (l = 0; l < 4; l++, p++, q++) 
          PredBlk[p] = (int)(predV[q]);
      }

      /*
       * Check if V block is coded.
       */
      if (cnc_v) {
        bits_blk_mcc = bit_for_4x4info = bit_for_dct = 0;
        
        /*
         * Classify the block for 4x4 coding method subset selection.
         */
        class = findPredNum4x4((int *)PredBlk, qp_UV, CHROMA, INTER_BLOCK,
                               SIZE_4x4, 0, (int *)&rotateFlag);
        
        for (k = 0; k < 64; k++) 
          methCoeffs[k] = dctCoeffs[k]  = 0;
        
        /*
         * Store class information. METH_4x4_1 signals that there is only
         * one 4x4 block.
         */
        methCoeffs[13] = class;
        methCoeffs[14] = rotateFlag;
        codingMethod = METH_4x4_1;
        
        /*
         * Decode 4x4 coding method information.
         */
        bit_for_4x4info = pedLoad4x4Info(methCoeffs, CHROMA, INTER_BLOCK,
                                         qp_UV, SIZE_4x4, codingMethod,
                                         instance, BitError);
        if (*BitError == VDX_SAW_SYNC) 
          return;
        
        /*
         * If block is coded with 4x4 DCT or 4x4 KLT, 
         * decode transform coefficient information.
         */
        dct4x4Methods = findDctMeth((int *)methCoeffs, codingMethod, 
                                    (int *)save4);
        
        if (dct4x4Methods >= METH_4x4_DCT_1) {
          bit_for_dct = pedLoadDctCoeffs(dctCoeffs, dct4x4Methods, CHROMA,
                                         qp_UV, INTER_BLOCK, (int *)save4,
                                         instance, BitError);
          if(*BitError == VDX_SAW_SYNC) return;
        }
        
        /*
         * If block is coded with extrapolation, extract the
         * reference pixels.
         */
        if (methCoeffs[10] == EXTRAPOL_METH) 
          pels = getExtSouJunct((int *)PredBlk, (int *)ExtBlk, COLS_4x4, 
                                ROWS_4x4);
        
        /*
         * Reconstruct the 4x4 prediction error block.
         */
        bits_blk_mcc = dec4x4((int *)PredBlk, (int *)DiffBlk, 
                              (int *)methCoeffs, (int *)dctCoeffs,
                              qp_UV, (int *)ExtBlk, pels, CHROMA, INTER_BLOCK,
                              SIZE_4x4);
                
        *bits_reg_mcc_uv += bit_for_4x4info + bit_for_dct;        
      }
      
      /*
       * Create final reconstruction block.
       */
      pedRecoInterBlk((int32*)PredBlk, (int32*)DiffBlk, recoV, y, x,
                      sizex_ch, 4, cnc_v);      
      break;
  }
}









