/*
 * decCore.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center.
 *
 * Contains:
 *    This module, namely the Video Decoder Core (VDC), is the external 
 *    interface for the core MVC decoder. It hides the internals of 
 *    the decoder. 
 *
 */

#include "nccglob.h"
#include "decCore.h"
#include "debug.h"
#include "mainFlags.h"
#include "mvc.h"
#include "mcp.h"
#include "deblockFilt.h"
#include "decDemux.h"
#include "IntraPredDecoder.h"
#include "PredErrorDecoder.h"
#include "IOimg.h"

/*
 * vdcCloseCore
 *
 * Parameters:
 *    instance                   a pointer to the instance data
 *
 * Function:
 *    Closes the instance. Frees all memory allocated for the instance.
 *
 * Returns:
 *    >= 0  if successful
 *    < 0   if an error occured
 */

int vdcCloseCore(vdcInstance_t *instance)
{
  int
    retValue = VDC_OK; /* value returned from this function */

  if (!instance)
    return VDC_OK;

  if (vdxClose(instance->vdxInstance) < 0)
    retValue = VDC_ERROR;

  nccFree((void *) instance->slicePredMap);
  nccFree((void *) instance->object_list);

  nccFree(instance);

  return retValue;
}

  
/*
 * vdcDecodeFrame
 *    Copyright (c) 1997 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    instance                   a pointer to the instance data
 *    prevfr                     the previous (reference) frame
 *    predfr                     the prediction frame
 *    recfr                      the reconstructed new reference frame
 *    outfr                      the postfiltered recfr
 *
 * Function:
 *    Decodes the next frame.
 *
 * Returns:
 *    >= 0  if successful
 *    < 0   if an error occured
 */

int vdcDecodeFrame(vdcInstance_t *instance, struct frame **longTermMem, 
                   struct frame *predfr, struct frame *recfr, 
                   struct frame *outfr, int *BitError)
{
  static int   qp, OK, pic_sizex, pic_sizey, sliceMode, useLoopFilter;
  int32 numHeaderBits;
  int aQP;

  nccAssert(instance != NULL);
  nccAssert(predfr != NULL);
  nccAssert(recfr != NULL);
  nccAssert(outfr != NULL);

  /* Reading the vital params from bitstream file */

  pic_sizex = instance->pic_sizex;
  pic_sizey = instance->pic_sizey;

  /*
   * Detect if a slice SSC follows in the bitstream.
   * If yes read its header. If no then a PSC is
   * read together with frame header parameters.
   */
  if (vdxCheckSSC(&instance->IPB, /* 0- P, 2- I*/
                  &pic_sizex, &pic_sizey,
                  &instance->frameNum, &instance->skip,
                  &qp, &numHeaderBits,
                  instance->vdxInstance, (int32 *)BitError))
  {
    sliceMode = 1;
    OK = 1;
  }
  else
    OK = vdxGetPicHead(&instance->IPB /* 0- P, 2- I*/,
                       &instance->frameNum,
                       &instance->skip,
                       &sliceMode,
                       &useLoopFilter,
                       &qp,
                       &pic_sizex, &pic_sizey,
                       &instance->numActiveRefFrames,
                       (int32 *)instance->orderedRefFrames,
                       &numHeaderBits,
                       1, instance->vdxInstance, (int32 *)BitError);
  if(*BitError == VDX_SAW_SYNC) return 0;

  ioiInitFrame(recfr, pic_sizex, pic_sizey);

  /*
   * If we expect an I picture but header points
   * otherwise then exit with error.
   */
  if (instance->fStartOfSeq || instance->IPB == 2)
  {
    instance->fStartOfSeq = 0;

    if((instance->IPB!=2) || (!OK))
      return VDC_ERROR;
    if (pic_sizex != instance->pic_sizex || pic_sizey != instance->pic_sizey)
      return VDC_ERROR;

    instance->qp = qp;

    deb4p("I frame: %d size= %dx%d qp= %d\n",
          instance->frameNum, pic_sizex, pic_sizey, qp);

    /*
     * Decode the I picture layer.
     */
    instance->total_bits = Decode_I_Frame(recfr, instance->pic_sizex,
                                          instance->pic_sizey,
                                          instance->frameNum, qp,
                                          &instance->bits_dct_y,
                                          &instance->bits_dct_u,
                                          &instance->bits_dct_v,
                                          &instance->bits_ovr_y,
                                          &instance->bits_ovr_uv,
                                          &numHeaderBits,
                                          instance->slicePredMap,
                                          sliceMode,
                                          instance->vdxInstance,
                                          BitError);

    if(*BitError == VDX_SAW_SYNC) 
      return 0;

  }

  else
  {
    if (!OK || pic_sizex != instance->pic_sizex || pic_sizey != 
        instance->pic_sizey)
      return VDC_ERROR;

    instance->qp = qp;

    /*
     * Affine QP is a function of the frame QP
     */
    aQP = affineQPs[qp];
  
    deb4p("P frame: %d size= %dx%d qp= %d\n",
          instance->frameNum, pic_sizex, pic_sizey, qp);
    
    ioiInitFrame(outfr, pic_sizex, pic_sizey);

    /*
     * Decode the P picture layer.
     */
    instance->total_bits = DecodePFrame(longTermMem, predfr, recfr, 
                                        pic_sizex, pic_sizey, 
                                        instance->frameNum, 
                                        instance->object_list, 
                                        instance->slicePredMap,
                                        sliceMode,
                                        qp, aQP,
                                        instance->numActiveRefFrames,
                                        instance->orderedRefFrames,
                                        &instance->frame_bits, 
                                        &instance->bits_dct_y,
                                        &instance->bits_dct_u, 
                                        &instance->bits_dct_v, 
                                        &instance->bits_ovr_y, 
                                        &instance->bits_ovr_uv, 
                                        &instance->bits_mode,
                                        &numHeaderBits,
                                        outfr,
                                        instance->vdxInstance, 
                                        BitError,
                                        useLoopFilter);
    if(*BitError == VDX_SAW_SYNC) return 0;

  }

  #ifdef BIBLIN
  {
    int numStuffBits;

    if (vdxCheckSync(&numStuffBits, instance->vdxInstance) == VDX_SNC_NO_SYNC)
    {
      deb0f(stderr, "Sync lost.\n");
      exit(1);
    }

    instance->header_bits = numHeaderBits + numStuffBits;

    /* Note that sync code emulation prevention bits are not counted here. */
  }
  #endif

  return VDC_OK;
}


/*
 * vdcExit
 *    Copyright (c) 1997 Nokia Research Center, Tampere, FINLAND
 *
 * Function:
 *    Can be used for debugging. Helps to track the caller of exit.
 */

void vdcExit(void)
{
  while (0) {} /* debug */
}


/*
 * vdcFree
 *    Copyright (c) 1997 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    none
 *
 * Function:
 *    Frees the module. Releases the resources allocated by vdcLoad.
 *    After a call to this function, functions of this module
 *    must no longer be called.
 *
 * Returns:
 *    Nothing
 */

void vdcFree(void)
{
}


/*
 * vdcLoad
 *    Copyright (c) 1997 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    none
 *
 * Function:
 *    Loads the module. All instance-independent intializations are placed
 *    in this function. This function must be called before any other function
 *    of this module.
 *
 * Returns:
 *    >= 0  if successful
 *    < 0   if an error occured
 */

int vdcLoad(void)
{
  return VDC_OK;
}


/*
 * vdcOpenCore
 *    Copyright (c) 1997 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    pic_sizex, pic_sizey       the width and height of the luminance image
 *
 * Function:
 *    Opens a new instance of this module. Allocates and initializes 
 *    the instance data.
 *
 * Returns:
 *    a pointer to the instance data
 */

vdcInstance_t *vdcOpenCore(int pic_sizex, int pic_sizey)
{
  vdcInstance_t *instance;

  /* Allocate instance data */
  instance = nccMalloc(sizeof(vdcInstance_t));
  if (!instance)
    return NULL;
  memset(instance, 0, sizeof(vdcInstance_t));

  /* Beginning of sequence */
  instance->fStartOfSeq = 1;

  instance->pic_sizex = pic_sizex;
  instance->pic_sizey = pic_sizey;

  if((instance->slicePredMap = (short *)nccCalloc(pic_sizex * pic_sizey >> 6,
     sizeof(int))) == NULL)
  {
    nccFree(instance);
    return NULL;
  }

  if ((instance->object_list = (objectBuffer_t *)nccCalloc((pic_sizex * 
      pic_sizey) >> 6, sizeof(objectBuffer_t)))  == NULL)
  {
    nccFree(instance);
    return NULL;
  }

  /*** INITIALIZATIONS ***/

  mvcInitKernel();
  filInitOverallActTable();
  filInitDiffActTable();

  instance->vdxInstance = vdxOpen();
  if (!instance->vdxInstance)
    goto errVdxOpen;

  return instance;

  /* Error situations */

  vdxClose(instance->vdxInstance);
  errVdxOpen:

  nccFree(instance);
  return NULL;
}
