/*
 * decDemux.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Functions for video demultiplexer.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bib.h"
#include "nrctypes.h"
#include "nccglob.h"
#include "decDemux.h"
#include "debug.h"
#include "mainFlags.h"
#include "toolMCC.h"

/*
 * Include bitstream decoding tables used in video
 * demultiplexer. All these are static fucntions.
 */
#include "decDemux.dat"

/*
 * Global variables
 */

int 
  decLongTermMemSize = 1; /* Set by picture header */

/*
 * NULLing is needed because first time dcmRequestSliceRefill() is
 * called it will try to free the structures pointed to by buff_in
 */
bibBuffer_t 
  *buff_in = NULL;

/* CNC coding information tables */
int 
  tableInter[65] = {15, 15, 15, 15, 0, 15, 15, 15,
                    1, 15, 15, 15, 2, 15, 15, 15,
                    3, 15, 15, 15, 4, 15, 15, 15,
                    5, 15, 15, 15, 6, 15, 15, 15,
                    7, 15, 15, 15, 8, 15, 15, 15,
                    9, 15, 15, 15, 10, 15, 15, 15,
                    11, 15, 15, 15, 12, 15, 15, 15,
                    13, 15, 15, 15, 14, 15, 15, 15, 15}; 

/* 15 is the escape index*/

int 
  tableIntra[65] = {0, 18, 18, 18, 1, 18, 18, 18, 
                    2, 18, 18, 18, 3, 18, 18, 18, 
                    4, 18, 18, 18, 5, 18, 18, 18, 
                    18, 18, 18, 18, 6, 18, 18, 18, 
                    7, 18, 18, 18, 8, 18, 18, 18, 
                    9, 18, 18, 18, 10, 18, 18, 18, 
                    11, 18, 18, 18, 12, 18, 18, 18, 
                    13, 18, 18, 18, 14, 15, 16, 17, 18};

/* 18 is the escape index*/

int
  invtableIntra[18] = {0, 4, 8, 12, 16, 20, 28, 32, 36, 
						   40, 44, 48, 52, 56, 60, 61, 62, 63};
int
  invtableInter[15] = {4, 8, 12, 16, 20, 24, 28, 32,  
							  36, 40, 44, 48, 52, 56, 60};
/*
 * Local variables
 */

/*
 * Empty structs used as arguments for vdxGetVlcItemIn3Stages
 * when decoding is not using the second or third stage.
 */ 
static vdxVlcToPairIndTbl_t
  vdxVlcToPairInd_NULL = {
    NULL, 0
  };

static vdxVlcPairToIndTbl_t
  vdxVlcPairToInd_NULL = {
    NULL, 0
  };


/*
 * Local function prototypes
 */

static void vdxShowBits(int32 numberOfBits, bibBuffer_t *buffer, u_int32* code,
                        vdxInstance_t *instance);

static void vdxFlushBits(int32 numberOfBits, bibBuffer_t *buffer, 
                         vdxInstance_t *instance, int32 *bitError);

static int32 vdxGetVlcItem(int32 maxLength, inv_vlc_item *vlcToEventIndex, 
                           int32 *actualLength, u_int32 *actualCode, 
                           vdxInstance_t *instance, int32 *bitError);

static int32 vdxGetVlcItemIn2Stages(int32 headingZeros, int32 maxLength1, 
                                    inv_vlc_item *vlc1ToEventIndex, 
                                    int32 maxLength2, 
                                    inv_vlc_item *vlc2ToEventIndex, 
                                    int32 *actualLength, u_int32 *actualCode, 
                                    vdxInstance_t *instance, int32 *bitError);

static int32 vdxGetVlcItemWithMatching(vdxInstance_t *instance,
                                       int32 maxLength,
                                       vlc_item *eventIndexToVlc, 
                                       int32 *index, int32 *bitError);

static int32 vdxGetVlcItemIn3Stages(vdxInstance_t *instance, 
                                    vdxVlcToIndTbl_t vlc1ToEventIndex, 
                                    vdxVlcToPairIndTbl_t vlc2ToPair1Index, 
                                    vdxVlcToPairIndTbl_t vlc3ToPair2Index, 
                                    vdxVlcPairToIndTbl_t vlcPair1ToEventIndex, 
                                    vdxVlcPairToIndTbl_t vlcPair2ToEventIndex, 
                                    int32 *index, int32 *bitError);

static int32 vdxGetPicStart(int32 *startOrEnd, vdxInstance_t *instance, 
                            int32 *bitError);


/* This function returns one if picture header OK, zero
 * if end of session detected, exits to operating system
 * if neither PSC nor EOS detected.
 * Exceptionally bit is not returned as a function value
 * but instead the bit pointer, being one of the paramters, 
 * is incremented accordingly.
 */


static void vdxShowBits(int32 numberOfBits, bibBuffer_t *buffer, 
                        u_int32 *code_u32, vdxInstance_t *instance)
{
  int32 bitsGot, syncStartIndex;
  int16 ownError = 0;

  *code_u32 = bibShowBits(numberOfBits, buffer, &bitsGot, 
                          &syncStartIndex, &ownError); 

  if (ownError == ERR_BIB_NOT_ENOUGH_DATA)
    *code_u32 <<= numberOfBits - bitsGot;
  else if (ownError)
    nccExit(1);
 
  return;
}


static void vdxFlushBits(int32 numberOfBits, bibBuffer_t *buffer, 
                        vdxInstance_t *instance, int32 *bitError)
{
  int32 bitsGot, syncStartIndex;
  int16 ownError = 0;

  *bitError =  NO_BIT_ERROR;    
  
  bibFlushBits(numberOfBits, buffer, &bitsGot, &syncStartIndex, 
               &ownError);

  if (syncStartIndex != NO_BIT_ERROR)
  {
    *bitError = VDX_SAW_SYNC; 
     return;
  }

  if (ownError)
    nccExit(1);
  
  return;
}


/*
 *
 * vdxGetVlcItem:
 *
 * Parameters:
 *      maxLength             I: Maximum length of the VLC
 *      vlcToEventIndex       I: Look-up table from code to 
 *                               event index for decoding the VLC
 *      actualLength          O: Actual length of the decoded VLC
 *      actualCode            O: Actual code of the decoded VLC
 *      instance,             I: Decoding instance
 *      bitError              O: Bit error flag
 *
 * Function:
 *      Decodes VLC using look-up table.
 *      
 * Returns:
 *      Index of the decoded VLC
 */

static int32 vdxGetVlcItem(int32 maxLength, inv_vlc_item *vlcToEventIndex, 
                           int32 *actualLength, u_int32 *actualCode, 
                           vdxInstance_t *instance, int32 *bitError)
{
  int32 
    length, index;
  u_int32 
    code;               

  /*
   * Do the decoding using the look-up table, flush the decoded 
   * bits from the bitstream, and return the actual length of the VLC. 
   */
  vdxShowBits(maxLength, buff_in, &code, instance);
  length = vlcToEventIndex[code].len;
  index = vlcToEventIndex[code].index;

  vdxFlushBits(length, buff_in, instance, bitError);
  if (*bitError == VDX_SAW_SYNC) 
    return 0;

  code >>= maxLength - length;
  *actualLength = length;
  *actualCode = code;

  return index;
}


/*
 *
 * vdxGetVlcItemIn2Stages:
 *
 * Parameters:
 *      headingZeros          I: Number of heading zeros
 *      maxLength1            I: Maximum length of the first VLC
 *      vlc1ToEventIndex      I: Look-up table from code to event 
 *                               index for decoding the first VLC
 *      maxLength2            I: Maximum length of the second VLC
 *      vlc2ToEventIndex      I: Look-up table from code to event 
 *                               index for decoding the second VLC
 *      actualLength          O: Actual length of the decoded VLC
 *      actualCode            O: Actual code of the decoded VLC
 *      instance,             I: Decoding instance
 *      bitError              O: Bit error flag
 *
 * Function:
 *      Decodes VLC using two look-up tables and splitting of the long
 *      VLC. The first part consist of short codewords, while the latter
 *      part consist of longer codewords. 
 *      
 * Returns:
 *      Index of the decoded VLC
 */

static int32 vdxGetVlcItemIn2Stages(int32 headingZeros, int32 maxLength1,
                                    inv_vlc_item *vlc1ToEventIndex, 
                                    int32 maxLength2,
                                    inv_vlc_item *vlc2ToEventIndex, 
                                    int32 *actualLength, u_int32 *actualCode, 
                                    vdxInstance_t *instance, int32 *bitError)
{
  int32 
    length, index, maxLength;
  inv_vlc_item 
    *vlcToEventIndex;
  u_int32 
    code;               

  *actualLength = 0;

  /*
   * First, decode the beginning of the possible long VLC.
   */
  vdxShowBits(headingZeros, buff_in, &code, instance);

  if (code != 0) 
  {
    /*
     * If some of the bits were not zero, select the first decoding table.
     */
    vlcToEventIndex = vlc1ToEventIndex;
    maxLength = maxLength1;
  }
  else 
  { 
    /*
     * Only zero bits received, flush the decoded bits from the buffer
     * and select the second table to decode the rest of the long VLC.
     */
    vlcToEventIndex = vlc2ToEventIndex;
    maxLength = maxLength2;

    vdxFlushBits(headingZeros, buff_in, instance, bitError);
    if (*bitError == VDX_SAW_SYNC) 
      return 0;
    *actualLength += headingZeros;
  }
  
  /*
   * Do the decoding using the selected look-up table, flush the 
   * decoded bits from the bitstream, and return the actual length 
   * of the VLC.
   */
  vdxShowBits(maxLength, buff_in, &code, instance);

  length = vlcToEventIndex[code].len;
  index = vlcToEventIndex[code].index;

  vdxFlushBits(length, buff_in, instance, bitError);
  if (*bitError == VDX_SAW_SYNC) 
    return 0;

  code >>= maxLength - length;
  *actualLength += length;
  *actualCode = code;

  return index;
}


/*
 *
 * vdxGetVlcItemIn3Stages:
 *
 * Parameters:
 *      instance,             I: Decoding instance
 *      vlc1ToEventIndex      I: Struct containing look-up table from code to 
 *                               event index for decoding the beginning of the
 *                               long VLC
 *      vlc2ToPair1Index      I: Struct containing look-up table from code to 
 *                               pair table index for decoding the middle part 
 *                               of the long VLC
 *      vlc3ToPair2Index      I: Struct containing look-up table from code to  
 *                               pair table index for decoding the end of the
 *                               long VLC
 *      vlcPair1ToEventIndex  I: Struct containing look-up table from first   
 *                               code pair to event index
 *      vlcPair2ToEventIndex  I: Struct containing look-up table from seond  
 *                               code pair to event index
 *      index                 O: Index of the event
 *      bitError              O: Bit error flag
 *
 * Function:
 *      Decodes VLC Field1{Field2}{Field3} by dividing the reading 
 *      up to three stages. Function can be used to decode VLCs also in 
 *      one (Field2 and Field3 are optional) or two stages (Field3 is 
 *      optional) by passing appropriate decoding table structs 
 *      and using dummy structs for unused arguments.In each stages, 
 *      subset of the assumed long VLC is read and the final VLC
 *      is parsed using the decoding tables.
 *      
 * Returns:
 *      Actual length of the decoded VLC
 *
 */

static int32 vdxGetVlcItemIn3Stages(vdxInstance_t *instance, 
                                    vdxVlcToIndTbl_t vlc1ToEventIndex, 
                                    vdxVlcToPairIndTbl_t vlc2ToPair1Index, 
                                    vdxVlcToPairIndTbl_t vlc3ToPair2Index, 
                                    vdxVlcPairToIndTbl_t vlcPair1ToEventIndex, 
                                    vdxVlcPairToIndTbl_t vlcPair2ToEventIndex, 
                                    int32 *index, int32 *bitError)
{
  int32 
    length, 
    index1, index2, index3, tableIndex;

  u_int32 
    code;               

  /*
   * Decode the Field1 of the assumed long VLC using decoding table
   * vlcToInd1.
   */
  vdxShowBits(vlc1ToEventIndex.maxLength, buff_in, &code, instance);
  length = vlc1ToEventIndex.table[code].len;
  index1 = vlc1ToEventIndex.table[code].index;

  /*
   * If legal VLC was decoded, flush the decoded bits from the buffer, 
   * and return the actual VLC length.
   */ 
  if (length != LEN_SIGNAL) 
  {
    *index = index1;

    vdxFlushBits(length, buff_in, instance, bitError);
    if (*bitError == VDX_SAW_SYNC) 
      return 0;

    return length;
  }

  /* 
   * Legal VLC was not found yet. Flush the decoded bits and decode 
   * Field2 of the assumed long VLC.
   */
  else 
  {
    vdxFlushBits(vlc1ToEventIndex.maxLength, buff_in, instance, bitError);
    if (*bitError == VDX_SAW_SYNC) 
      return 0;

    vdxShowBits(vlc2ToPair1Index.maxLength, buff_in, &code, instance);
    index2 = vlc2ToPair1Index.table[code];

    /*
     * If legal VLC was now found, get the actual VLC using look-up table  
     * pairToInd1 for the pair of Field1 and Field2, flush the decoded
     * bits, and return the actual VLC length.
     */
    tableIndex = index1 * vlcPair1ToEventIndex.tableWidth + index2; 
    length = vlcPair1ToEventIndex.table[tableIndex].len;

    if (length != LEN_SIGNAL) 
    {
      *index = vlcPair1ToEventIndex.table[tableIndex].index;

      vdxFlushBits(length - vlc1ToEventIndex.maxLength, buff_in, 
                   instance, bitError);
      if (*bitError == VDX_SAW_SYNC)
        return 0;

      return length;
    }
    
    /* 
     * Legal VLC was not found yet. Flush the decoded bits and decode 
     * Field3 of the long VLC.
     */
    else 
    {
      index2 = vlcPair1ToEventIndex.table[tableIndex].index;

      vdxFlushBits(vlc2ToPair1Index.maxLength, buff_in, 
                   instance, bitError);
      if (*bitError == VDX_SAW_SYNC)
        return 0;

      vdxShowBits(vlc3ToPair2Index.maxLength, buff_in, &code, instance);
      index3 = vlc3ToPair2Index.table[code];

      /*
       * Get the actual VLC using look-up table pairToInd2 for the pair 
       * of Field1Field2 and Field3, Flush the decoded bits, and return
       * the actual VLC length.
       */
      tableIndex = index2 * vlcPair2ToEventIndex.tableWidth + index3;
      *index = vlcPair2ToEventIndex.table[tableIndex].index;
      length = vlcPair2ToEventIndex.table[tableIndex].len;

      vdxFlushBits(length - vlc1ToEventIndex.maxLength - 
                   vlc2ToPair1Index.maxLength, buff_in, instance, 
                   bitError);
      if (*bitError == VDX_SAW_SYNC)
        return 0;

      return length;
    }
  }
}


/*
 *
 * vdxGetVlcItemWithMatching:
 *
 * Parameters:
 *      instance,            I: Decoding instance
 *      maxLength            I: Number of bits to be read for the longest VLC 
 *      eventIndexToVlc      I: Look-up table from event index to VLC  
 *      index                O: Index of the event
 *      bitError             O: Bit error flag
 *
 * Function:
 *      Decodes VLC from the bitstream by matching known VLC to
 *      bitstream. Function is slow, and it will be obsolete
 *      later. At the moment, function is used only in vdxGetCodingMethod.
 *
 * Returns:
 *      Actual length of the decoded VLC
 */

static int32 vdxGetVlcItemWithMatching(vdxInstance_t *instance, 
                                       int32 maxLength,
                                       vlc_item *eventIndexToVlc, int32 *index, 
                                       int32 *bitError)
{
  int32
    i = 0, length;

  u_int32
    code, tmpCode;

  /*
   * Read the maximum amount of bits from the buffer.
   */
  vdxShowBits(maxLength, buff_in, &code, instance);
  
  /*
   * Starting from the beginning, scan through the possible
   * VLCs and match bit patterns against the read bits untill
   * correct VLC is found.
   */ 
  length = eventIndexToVlc[i].len;
  tmpCode = code >> (maxLength - eventIndexToVlc[i].len);

  while (tmpCode != eventIndexToVlc[i].code) 
  {
    i++;
    length = eventIndexToVlc[i].len;
    tmpCode = code >> (maxLength - length);
  }
  
  *index = i;

  /*
   * Flush the decoded bits from the buffer and
   * return actual VLC length.
   */
  vdxFlushBits(length, buff_in, instance, bitError);
  if (*bitError == VDX_SAW_SYNC)
    return 0;

  return length;
} 


/*
 *
 * vdxGetTriplet:
 *
 * Parameters:
 *      instance,       I: Decoding instance
 *      vlcTableInd     I: Index of the used triplet VLC table
 *      levelBits       I: Width of the LEVEL bit field in case of escape
 *      runBits         I: Width of the RUN bit field in case of escape
 *      last            O: LAST flag for the coefficient
 *      run             O: RUN for the coefficient
 *      level           O: Signed LEVEL of the coefficient
 *      bitError        O: Bit error flag
 *
 * Function:
 *      Decodes index of the LAST-RUN-LEVEL triplet from the
 *      bitstream.
 *
 * Returns:
 *      Amount of the decoded bits.
 */

int32 vdxGetTriplet(vdxInstance_t *instance, int32 vlcTableInd, 
                    int32 levelBits, int32 runBits, int32 *last, 
                    int32 *run, int32 *level, int32 *bitError)
{ 
  int32
    index, bits; 
 
  u_int32 
    sign, tmp;
 
  /*
   * Decode the VLC of the triplet index from the bitstream. 
   */

  bits = vdxGetVlcItemIn3Stages(instance, 
                                vlc1ToTripletIndex[vlcTableInd], 
                                vlc2ToTripletPair1Index[vlcTableInd], 
                                vlc3ToTripletPair2Index[vlcTableInd], 
                                vlcPair1ToTripletIndex[vlcTableInd], 
                                vlcPair2ToTripletIndex[vlcTableInd],
                                &index, bitError);

  if (*bitError == VDX_SAW_SYNC)
    return 0;

  /*
   * If decoded index was not ESCAPE, get the corresponding
   * LAST-RUN-LEVEL triplet from the table.
   */
  if (index != tripletEscapeIndex[vlcTableInd]) 
  {
    *last = (indexToTriplet[vlcTableInd])[index].last;
    *run = (indexToTriplet[vlcTableInd])[index].run;
    *level = (indexToTriplet[vlcTableInd])[index].level;
    /*
     * *triplet = tripletLut[index];
     */

    bits += vdxGetBits(1, buff_in, &sign, instance, bitError); 
    if (*bitError == VDX_SAW_SYNC)
      return 0;
  }

  /*
   * In case of ESCAPE, decode fixed-length LAST, RUN and signed LEVEL
   * separately.
   */
  else 
  {
    bits += vdxGetBits(1, buff_in, &tmp, instance, bitError); 
    if (*bitError == VDX_SAW_SYNC)
      return 0;
    *last = tmp;

    bits += vdxGetBits(runBits, buff_in, &tmp, instance, bitError);
    if (*bitError == VDX_SAW_SYNC)
      return 0;
    *run = tmp;

    bits += vdxGetBits(1, buff_in, &sign, instance, bitError);
    if (*bitError == VDX_SAW_SYNC)
      return 0;

    bits += vdxGetBits(levelBits, buff_in, &tmp, instance, bitError);
    if (*bitError == VDX_SAW_SYNC)
      return 0;
    *level = tmp;
  }
  if (sign)
    *level = -(*level);

 if (!(*level)) 
  {
    deb0f(stderr, "Zero level of AC\n");
    nccExit(1);
  }

  return bits;
}


/*
 *
 * vdxGetCodingMethod:
 *
 * Parameters:
 *      instance        I: Decoding instance
 *      vlcTableInd     I: Index of the used triplet VLC table
 *      index           O: Index of the coding method
 *      bitError        O: Bit error flag
 *
 * Function:
 *      Decodes coding method of the block from the bitstream.
 *
 * Returns:
 *      Amount of the decoded bits.
 *
 */

int32 vdxGetCodingMethod(vdxInstance_t *instance, int32 vlcTableInd, 
                         int32 *index, int32 *bitError)
{
  int32
    bits;

  /*
   * Decode the VLC of the coding method selection
   * from the bitstream
   */
  bits = vdxGetVlcItemWithMatching(instance, 11, 
                                   &vlcTableMethSel[vlcTableInd * 14 + 1],
                                   index, bitError);
  *index = *index + 1;  /* because it starts from 1 to 12 */
  return bits;
}


/*
 *
 * vdxGetIntraBMvector:
 *
 * Parameters:
 *      *bmVectY            O: vertical intra BM component
 *      *bmVectX            O: horizontal intra BM component
 *      reducedSearchRange  I: Flag to signal 136 search range instead of 320 
 *      instance,           I: Decoding instance
 *      bitError            O: Bit error flag
 *
 * Function:
 *      Decodes index of the INTRA block matching vector from the
 *      bitstream.
 *
 * Returns:
 *      Amount of the decoded bits.
 *
 */

int32 vdxGetIntraBMvector(int32 *bmVectY, int32 *bmVectX,  int32 reducedSearchRange,
                          vdxInstance_t *instance, int32 *bitError)
{
  int32
    bits;

  u_int32
    code;

  int
    BitsXLeft,
    BitsXRight,
    offsetX;

  if (reducedSearchRange)
  {
    BitsXLeft  = 0;
    BitsXRight = 4;
    offsetX    = 8;
  }
  else
  {
    BitsXLeft  =  3;
    BitsXRight =  5;
    offsetX    = 15;
  }

  bits = vdxGetBits(4, buff_in, &code, instance, bitError);
  if (*bitError == VDX_SAW_SYNC) 
    return 0;

  *bmVectY = - ((int32)code);

  if (*bmVectY > -8)
  {
    if (BitsXLeft)
    {
      bits += vdxGetBits(BitsXLeft, buff_in, &code, instance, bitError);
      if (*bitError == VDX_SAW_SYNC) 
        return 0;
    }
    else
      code = 0;
  }
  else
  {
    bits += vdxGetBits(BitsXRight, buff_in, &code, instance, bitError);
    if (*bitError == VDX_SAW_SYNC) 
      return 0;
  }

  *bmVectX = (int32)code - offsetX;

  return bits;
}


/*
 *
 * vdxGetIntraPredDcdCnc:
 *
 * Parameters:
 *      instance,       I: Decoding instance
 *      qp              I: Quantization parameter
 *      predIndex       O: Index of the INTRA predition method
 *      dcdLevel        O: DC difference LEVEL
 *      cnc             O: CNC information
 *      bitError        O: Bit error flag
 *
 * Function:
 *      Decodes INTRA prediction, and CNC information and DC difference
 *      from the bitstream.
 *
 * Returns:
 *      Amount of the decoded bits.
 */

int32 vdxGetIntraPredDcdCnc(vdxInstance_t *instance,
                            int32 qp, int32 ipContextNo, int32 *predIndex,
                            int32 *dcdLevel, int32 *cnc, int32 *bitError)
{
  static unsigned char
    auxLog[5] = {0, 0, 1, 2, 2},
    selLenList;

  int32
    offsetDcdSubindex,
    bitsDcdSubindex;

  int32
    bits,
    index,
    length;

  u_int32
    code;

  selLenList = ipContextLists[ipContextNo*(MAX_IP_LIST_SIZE+1)];

  /* Get the INTRA prediction information */

  if (selLenList > 6 || selLenList < 1) {
    deb1f(stderr, "Invalid length of list %d\n", (int)selLenList);
    nccExit(1);
  }
  else if (selLenList < 2)
    bits = 0;
  else
  {
    vdxShowBits(2, buff_in, &code, instance);

    if (code > 1)
    {
      *predIndex = 0;
      vdxFlushBits(1, buff_in, instance, bitError);
      if (*bitError == VDX_SAW_SYNC)
        return 0;

      bits = 1;
    }
    else if (code == 1)
    {
      *predIndex = 1;
      vdxFlushBits(2, buff_in, instance, bitError);
      if (*bitError == VDX_SAW_SYNC)
        return 0;

      bits = 2;
    }
    else
    {
      vdxFlushBits(2, buff_in, instance, bitError);
      if (*bitError == VDX_SAW_SYNC)
        return 0;

      bits = 2;

      /* Additional Fixed Length */

      length = auxLog[selLenList - 2];

      if (length)
      {
        bits += vdxGetBits(length, buff_in, &code,
                           instance, bitError);
        if (*bitError == VDX_SAW_SYNC)
          return 0;

        *predIndex = 2 + code;
      }
      else
        *predIndex = 2;
    }
  }


  /* Reading the CNC bit */

  bits += vdxGetBits(1, buff_in, &code, instance, bitError);

  if (*bitError == VDX_SAW_SYNC)
    return 0;

  *cnc = code;


  /* Reading the DCD information */

  /*
   * Decode VLC for range index of DC difference LEVEL from the bitstream
   */  
  length = vdxGetVlcItemIn3Stages(instance, 
                                  vlc1ToIntraDcdRangeIndex,
                                  vlc2ToPair1IntraDcdRangeIndex,
                                  vlc3ToPair2IntraDcdRangeIndex,
                                  vlcPair1ToIntraDcdRangeIndex,
                                  vlcPair2ToIntraDcdRangeIndex,
                                  &index, bitError);
  if (*bitError == VDX_SAW_SYNC)
    return 0;
  
  bits += length;

  /*
   * Check the offset of the DC difference LEVEL and the length of 
   * possible fixed length code.
   */
  dcdRangeIndexToBits(index, qp,
                      &offsetDcdSubindex, &bitsDcdSubindex);

  /*
   * Read the possible additional bits, flush the bits from the buffer,
   * and add refine the DC difference LEVEL to it's final value.
   */
  if (bitsDcdSubindex) {
    bits += vdxGetBits(bitsDcdSubindex, buff_in, &code, instance, bitError);
    if (*bitError == VDX_SAW_SYNC)
      return 0;

    *dcdLevel = offsetDcdSubindex + (int32)code;
  }
  else
    *dcdLevel = offsetDcdSubindex;

  /*
   * Read the SIGN bit if DC difference LEVEL differs from zero and
   * flush the bit from the buffer.
   */

  if (*dcdLevel)
  {
    bits += vdxGetBits(1, buff_in, &code, instance, bitError);
    if (*bitError == VDX_SAW_SYNC)
      return 0;

    if (code) {
      *dcdLevel = -(*dcdLevel);
    }
  }

  return bits;

}

/*
 * Functions needed in MCC, video multiplexer, and video demultiplexer,
 * should go to some other module later in order to avoid the need of
 * video demultiplexer in encoder.
 */

/*
 *
 * vdxGetDctTable:
 *
 * Parameters:
 *      cols, rows,     I: Size of block
 *      qp,             I: QP quantization parameter
 *      chroma,         I: Flag to signal luma or chroma
 *      mode,           I: Flag to signal INTER or INTRA
 *      tableNo,        O: Number of the triplet VLC table to use
 *      runBits,        O: Number of bits to encode the RUN
 *      levelBits,      O: Number of bits to encode the LEVEL
 *
 * Function:
 *      This function translates the block parameters into index
 *      of triplet VLC table and it also return the amount of bits
 *      by which RUN and LEVEL should be encoded with.
 *
 * Returns:
 *      Maximum level for coefficient
 */

/* Definition of tables for different modes, component type and QP values */
#define METHOD_8x8		0
#define METHOD_4x4		1
#define METHOD_4x8		2
#define METHOD_3x8		3
#define METHOD_NB_ELEM	4

#define NB_MODES		3
#define NB_CHROMA		2
#define QP_LAST_LEVEL	28

static int RunBits[METHOD_NB_ELEM]={6,4,5,5};

static int LevelBits[NB_CHROMA][QP_LAST_LEVEL+2]=
{
	/* LUMA */
	{ 
		9,9,8,7,7,7,6,6,6,6,6,6,5,5,5,5,5,5,5,5,5,5,5,5,4,4,4,4,4,4
	},
	/* CHROMA */
	{ 
		8,8,7,6,6,6,5,5,5,5,5,5,4,4,4,4,4,4,4,4,4,4,4,4,3,3,3,3,3,3
	}
};

static int MaxLevel[NB_CHROMA][QP_LAST_LEVEL+2] =
{
	/* LUMA */
	{ 
		LEVEL_09,LEVEL_09,LEVEL_08,LEVEL_07,LEVEL_07,LEVEL_07,LEVEL_06,
		LEVEL_06,LEVEL_06,LEVEL_06,LEVEL_06,LEVEL_06,LEVEL_05,LEVEL_05,
		LEVEL_05,LEVEL_05,LEVEL_05,LEVEL_05,LEVEL_05,LEVEL_05,LEVEL_05,
		LEVEL_05,LEVEL_05,LEVEL_05,LEVEL_04,LEVEL_04,LEVEL_04,LEVEL_04,
		LEVEL_04,LEVEL_04
	},
	/* CHROMA */
	{ 
		LEVEL_08,LEVEL_08,LEVEL_07,LEVEL_06,LEVEL_06,LEVEL_06,LEVEL_05,
		LEVEL_05,LEVEL_05,LEVEL_05,LEVEL_05,LEVEL_05,LEVEL_04,LEVEL_04,
		LEVEL_04,LEVEL_04,LEVEL_04,LEVEL_04,LEVEL_04,LEVEL_04,LEVEL_04,
		LEVEL_04,LEVEL_04,LEVEL_04,LEVEL_03,LEVEL_03,LEVEL_03,LEVEL_03,
		LEVEL_03,LEVEL_03
	}
};



/* qpBlockLuma used to pass qpY to vdxGetDctTable() and vdxGetKltTable() */
int
  qpBlockLuma;

static signed char
  tcoeffSelTableY[320] = 
  {
    -1, 1, 1, 1, 1, 1, 1, 1, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    -1, 0, 0, 0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    -1, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
    -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    -1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    -1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2
  };


/*
 *
 * vdxGetDctTable:
 *
 * Parameters:
 *      cols, rows,     I: Size of block
 *      qp,             I: QP quantization parameter
 *      chroma,         I: Flag to signal luma or chroma
 *      mode,           I: Flag to signal INTER or INTRA
 *      tableNo,        O: Number of the triplet VLC table to use
 *      runBits,        O: Number of bits to encode the RUN
 *      levelBits,      O: Number of bits to encode the LEVEL
 *
 * Function:
 *      This function translates the block parameters into index
 *      of triplet VLC table and it also return the amount of bits
 *      by which RUN and LEVEL should be encoded with.
 *
 * Returns:
 *      Maximum level for coefficient
 */

int32 vdxGetDctTable(int32 cols, int32 rows, int32 qp, int32 chroma, 
                     int32 mode, int32 *tableNo, int32 *runBits, 
                     int32 *levelBits)
{
  int
    blkType, method=0;

  /* Put qp into valid range */
	if (qp < 0)
		qp=0;
	else if (qp > QP_LAST_LEVEL+1)
		qp = QP_LAST_LEVEL+1;

	/* Put mode in valid range (in case of error) */
	if (mode<0)
		mode=0;
	else if (mode > 2)
		mode=2;

	/* Put chroma in valid range (in case of error) */
	if (chroma<0)
		chroma=0;
	else if (chroma>1)
		chroma=1;

  if (mode == INTER_BLOCK)
    blkType = 0;
  else
    blkType = 1;

	if (cols == COLS_8x8 && rows == ROWS_8x8)
  {
		blkType += 2 * 0;
    method = METHOD_8x8;
	}
	else if (cols == COLS_4x4 && rows == ROWS_4x4)
  {
		blkType += 2 * 1;
    method = METHOD_4x4;
	}
	else if (cols == COLS_4x8 && rows == ROWS_4x8)
  {
		blkType += 2 * 2;
    method = METHOD_4x8;
	}
	else 
	{
		deb2f(stderr, "Wrong cols %ld or rows %ld in vdxGetDctTable()\n", 
			  cols, rows);
		nccExit(1);
	}

  if (chroma == CHROMA)
  {
    *tableNo = tcoeffSelTableY[blkType*32 + qpBlockLuma];
  }
  else
  {
    *tableNo = tcoeffSelTableY[blkType*32 + qp];
  }

  /* Access data through table lookup */
	*runBits	=	RunBits[method];
	*levelBits	=	LevelBits[chroma][qp];

	return MaxLevel[chroma][qp];
}


/*
 *
 * vdxGetKltTable:
 *
 * Parameters:
 *      cols, rows,     I: Size of block
 *      qp,             I: QP quantization parameter
 *      chroma,         I: Flag to signal luma or chroma
 *      mode,           I: Flag to signal INTER or INTRA
 *      kltNum,         I: Flag to signal KLT direction
 *      tableNo,        O: Number of the triplet VLC table to use
 *      runBits,        O: Number of bits to encode the RUN
 *      levelBits,      O: Number of bits to encode the LEVEL
 *
 * Function:
 *      This function translates the block parameters into index
 *      of triplet VLC table and it also return the amount of bits
 *      by which RUN and LEVEL should be encoded with.
 *
 * Returns:
 *      Maximum level for coefficient
 */

int32 vdxGetKltTable(int32 cols, int32 rows, int32 qp, int32 chroma, 
                     int32 mode, int32 kltNum, int32 *tableNo, int32 *runBits, 
                     int32 *levelBits)
{
  int
    blkType;

  /* Put qp into valid range */
	if (qp < 0)
		qp=0;
	else if (qp > QP_LAST_LEVEL+1)
		qp = QP_LAST_LEVEL+1;

	/* Put mode in valid range (in case of error) */
	if (mode<0)
		mode=0;
	else if (mode > 2)
		mode=2;

	/* Put chroma in valid range (in case of error) */
	if (chroma<0)
		chroma=0;
	else if (chroma>1)
		chroma=1;

  if (mode == INTER_BLOCK)
    blkType = 0;
  else
    blkType = 1;

	if (cols == COLS_8x8 && rows == ROWS_8x8)
  {
		blkType += 2 * 0;
    *runBits = 4;
	}
	else if (cols == COLS_4x4 && rows == ROWS_4x4)
  {
		blkType += 2 * 1;
    *runBits = 3;
	}
	else 
	{
		deb2f(stderr, "Wrong cols %ld or rows %ld in vdxGetKltTable()\n",
   			  cols, rows);
		nccExit(1);
	}

	blkType += 6 * 1;

  /* Access data through table lookup */
  if (chroma == CHROMA)
  {
    /* void */
    *tableNo  = tcoeffSelTableY[blkType*32 + qpBlockLuma];
  }
  else
  {
    *tableNo  = tcoeffSelTableY[blkType*32 + qp];
  }

  /* Access data through table lookup */
	*levelBits	=	LevelBits[chroma][qp];

	return MaxLevel[chroma][qp];
}


/*
 *
 * vdxGetInterCncTable:
 *
 * Parameters:
 *      regionSize,     I: Size of region (16/8)
 *      regionMode,     I: Coding mode (REGMODE_INTRA/others)
 *      qp,             I: Quantization parameter
 *      tableNo,        O: Number of the Inter Cnc table to use
 *
 * Function:
 *      This function detects which Inter Cnc VLC table must be used.
 *
 * Returns:
 *      Void.
 */

void vdxGetInterCncTable(int32 regionSize, int32 regionMode, int32 qp,
                         int32 *tableNo)
{
  if(regionSize == 16) {
    if(regionMode == REGMODE_INTER_NOPRED) {            /* 60 */
      if(qp <= 8) {
        *tableNo = 0;
      }
      else {
        *tableNo = 1;
      }
    }
    else if(regionMode == REGMODE_INTER_PRED) {         /* 61 */
      if(qp <= 6) {
        *tableNo = 0;
      }
      else if(qp <= 17) {
        *tableNo = 1;
      }
      else {
        *tableNo = 2;
      }
    }
    else if(regionMode == REGMODE_INTER_PRED_DELTA) {   /* 62 */
      if(qp <= 10) {
        *tableNo = 0;
      }
      else {
        *tableNo = 1;
      }
    }
    else {      /* regionMode == REGMODE_INTER_COPY */  /* 63 */
      if(qp <= 3) {
        *tableNo = 0;
      }
      else if(qp <= 11) {
        *tableNo = 1;
      }
      else {
        *tableNo = 2;
      }
    }
    *tableNo += 3;
  }
  else {        /* regionSize == 8 */
    if(regionMode == REGMODE_INTER_NOPRED ||
       regionMode == REGMODE_INTER_NOPRED_8x8) {                /* 65 */
      if(qp <= 1) {
        *tableNo = 0;
      }
      else if(qp <= 23) {
        *tableNo = 1;
      }
      else {
        *tableNo = 2;
      }
    }
    else if(regionMode == REGMODE_INTER_PRED ||
            regionMode == REGMODE_INTER_PRED_8x8) {             /* 66 */
      if(qp <= 1) {
        *tableNo = 0;
      }
      else {
        *tableNo = 1;
      }
    }
    else if(regionMode == REGMODE_INTER_PRED_DELTA ||
            regionMode == REGMODE_INTER_PRED_DELTA_8x8) {       /* 67 */
      if(qp <= 1) {
        *tableNo = 0;
      }
      else if(qp <= 24) {
        *tableNo = 1;
      }
      else {
        *tableNo = 2;
      }
    }
    else {      /* regionMode == REGMODE_INTER_COPY_8x8 */      /* 68 */
      if(qp <= 1) {
        *tableNo = 0;
      }
      else {
        *tableNo = 1;
      }
    }
    *tableNo += 0;
  }
}


int
  /* For each QP this table indicates the
   * maximum amount of quantised DC levels.
   * maxDcLevels[qp] = floor(2048 / (2*QP))
   * Exceptions are QP 1-3 for which instead
   * of 1024, 512, 341 we have 256.
   */

  maxDcLevels[32] =
  {
      0, 256, 256, 256, 256, 204, 170, 146,
    128, 113, 102,  93,  85,  78,  73,  68,
     64,  60,  56,  53,  51,  48,  46,  44,
     42,  40,  39,  37,  36,  35,  34,  33
  };


/* For each QP this table indicates the
   fixed length bits needed to encode
   the highest levels of DC difference.
   The highest levels range from
   30 ... (maxDcLevels[qp]-1)
 */

unsigned char
  highestDcLevBits[32] = {
    8, 8, 8, 8, 8, 8, 8, 7,
    7, 7, 7, 6, 6, 6, 6, 6,
    6, 5, 5, 5, 5, 5, 4, 4,
    4, 4, 4, 3, 3, 3, 2, 2
  };


/*
 *
 * dcdRangeIndexToBits:
 *
 * Parameters:
 *      rangeIndex,        I: Index of vlc symbol in error coding dcd table
 *      qp,                I: QP quantization parameter
 *      offsetDcdSubindex, O: Offset value corresponding to the index range
 *      bitsDcdSubindex    O: Amount of bits to code the subindex
 *
 * Function:
 *      Given an index in range intra prediction dcd vlc table,
 *      this function returns the offset value and bits to encode
 *      the subindex within the corresponding range.
 *
 * Returns:
 *      Void.
 *
 */

void dcdRangeIndexToBits(int32 rangeIndex, int32 qp,
                         int32 *offsetDcdSubindex, int32 *bitsDcdSubindex)
{
  static unsigned char
    dcdOffs[12]      = {0, 1, 2, 3, 4, 6, 8, 10, 14, 18, 22, 30},
    dcdExtraBits[12] = {0, 0, 0, 0, 1, 1, 1,  2,  2,  2,  3, -1};

  *offsetDcdSubindex = dcdOffs[rangeIndex];

  if (rangeIndex < 11)
    *bitsDcdSubindex   = dcdExtraBits[rangeIndex];
  else
    *bitsDcdSubindex   = highestDcLevBits[qp];
}



/*
 *
 * indexToBitsIntraBMV:
 *
 * Parameters:
 *      index,             I: Index of vlc symbol in error coding dcd table
 *      chroma,            I: Flag to signal LUMA or CHROMA coding
 *      qp,                I: QP quantization parameter
 *      offsetDcdSubindex, O: Offset value corresponding to the index range
 *      bitsDcdSubindex    O: Amount of bits to code the subindex
 *
 * Function:
 *      Given an index in intra BMV vlc table,
 *      this function returns the offset value and bits to encode
 *      the subindex within the corresponding range.
 *
 * Returns:
 *      Void.
 */


void indexToBitsIntraBMV(int32 index, int32 *offsetBMVSubindex, int32 *bitsBMVSubindex)
{
  static int32
    rangeOffset[16] = {0, 2, 4, 6, 8,12,16,24,32,48,64,96,128,192,256,384};

  static signed char
    extraBits[16]   = {1, 1, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5,  6,  6,  7,  7};

  *offsetBMVSubindex = rangeOffset[index];
  *bitsBMVSubindex   =   extraBits[index];
}



int32 vdxGetLongtermRVLC(int32 *index, vdxInstance_t *instance, 
                         int32 *bitError)
{
  u_int32             
    code_u32;
  int32                 
    len;

  *index = vdxGetVlcItem(7, longtermRVLC, &len, &code_u32, instance, bitError);

  if(*bitError == VDX_SAW_SYNC)
    return 0;

  return len;
}


int vdxGetBits(int numberOfBits, bibBuffer_t *buffer, u_int32 *code_u32,
               vdxInstance_t *instance, int32 *bitError)
{
  int bitsGot, syncStartIndex;
  int16 ownError = 0;

  *bitError = NO_BIT_ERROR;     

  *code_u32 = bibGetBits(numberOfBits, buffer, &bitsGot, &syncStartIndex, 
                         &ownError);
  if (syncStartIndex != NO_BIT_ERROR)
  {
    *bitError = VDX_SAW_SYNC; 
    return 0;
  }
  
  if (ownError)
    nccExit(1);
  
  return bitsGot;
}


int
  PrevTR=0, PrevFrNo=0;

int vdxGetPicHead(int *IPB,          /* 0- P, 2- I */ 
                  int *FrNo,
                  int *skip,         /* Distance from the previous frame    */
                  int *sliceMode,    /* Flag to signal that slice mode is on*/
                  int *useLoopFilter,/* Signals loopfilter usage            */
                  int *QP, 
                  int *pic_sizex, int *pic_sizey,
                  int *numActiveRefFrames,/* Number of reference frames used*/
                  int32 *orderedRefFrames,/* Indexes of reference frames in
                                           * longterm memory                */
                  
                  /* memory location pointed by bits will be incremented
                   * by the number of bits read by this function
                   */
                  int32 *bits,
                  
                  /* 1- yes, one may not want to read PSC if sychronization
                   * was lost, and it has been recovered by detecting PCS
                   * by another routine
                   */
                  int ReadPSC,
                  vdxInstance_t *instance, int32 *bitError)
{
  int32
    PSC,TR, FrSz, index;
  u_int32 
    TR_u32, bit_u32, QP_u32, code_u32;

  *bits=0;

  if(ReadPSC){
    /* Picture Start Code */
    (*bits) += (int32) vdxGetPicStart(&PSC, instance, bitError);
    if(!PSC) return 0;
  }
  
  /* sliceMode */
  (*bits) +=  (int32) vdxGetBits(1, buff_in, &code_u32, instance, bitError);
  *sliceMode = (int) code_u32;

  /* temporal reference */
  (*bits) += (int32) vdxGetBits(8, buff_in, &TR_u32, instance, bitError);
  TR = (int) TR_u32;

  if(TR<PrevTR) *FrNo=PrevFrNo+TR+256-PrevTR;
  else *FrNo=PrevFrNo+TR-PrevTR;

  *skip=*FrNo-PrevFrNo-1;
  PrevFrNo=*FrNo;
  PrevTR=TR;
  
  /* PTYPE start */
  (*bits) +=  (int32) vdxGetBits(1, buff_in, &bit_u32, instance, bitError);
  if(!bit_u32) *IPB=0; /* P frame */
  else{
    (*bits) +=  (int32) vdxGetBits(1, buff_in, &bit_u32, instance, bitError);
    if(!bit_u32) *IPB=2; /* I frame */
    else{
        /* Reserved for future extensions */
        deb0f(stderr, "Unable to decode an extended syntax\n");
        nccExit(1);
		}
	  }
  
  (*bits) +=  (int32) vdxGetBits(2, buff_in, &bit_u32, instance, bitError);
  FrSz=(int)bit_u32;

  switch(FrSz)
  {
  case 0:
    *pic_sizex=176;
    *pic_sizey=144;
    break;
  case 3:
    *pic_sizex=352;
    *pic_sizey=288;
    break;
  case 2:
    *pic_sizex=128;
    *pic_sizey=96;
    break;
  case 1:
    (*bits) += (int32) vdxGetBits(7, buff_in, &bit_u32, instance, bitError);
    *pic_sizex = (int) (bit_u32+1)*16;
    (*bits) += (int32) vdxGetBits(7, buff_in, &bit_u32, instance, bitError);
    *pic_sizey = (int) (bit_u32+1)*16;
    break;
  default:
    deb0f(stderr, "Unrecognized picture format\n");
    nccExit(1);
    break;
  }
  /* PTYPE end */

  (*bits) +=  (int32) vdxGetBits(5, buff_in, &QP_u32, instance, bitError); 
  /* frame QP */
  *QP = (int) QP_u32;

  /* Loopfilter information for a P-frame */

  if(*IPB != 2){
    (*bits) += (int32) vdxGetBits(1, buff_in, &code_u32, instance, bitError);
    *useLoopFilter = (int) code_u32;
  }
 
  /* In long term memory mode read the reference frame info */
  if(decLongTermMemSize > 1 && *IPB == 0){

    /* Assume long-term memory with temporal order */

    (*bits) += (int32) vdxGetBits(1, buff_in, &code_u32, instance, bitError);

    if(code_u32)
      (*numActiveRefFrames) = decLongTermMemSize;
    else
      (*numActiveRefFrames) = 1;

    /* Reference frame indexes in the long term memory */
    for(index = 0;index < *numActiveRefFrames;index++)
      orderedRefFrames[index] = index;

  }

  /* Longterm memory size is found in intra picture header */
  if(*IPB == 2){

    /* Maximum number of reference frames */
    (*bits)           += vdxGetLongtermRVLC(&index, instance, bitError);
    decLongTermMemSize = index + 1;
  }

  return 1 /*Everything OK */;
}


int vdxGetSliceHead(int *mbNum, int *intraSliceFlag, int *sliceQuant,
                    int *frNo, int *picSizeX, int *picSizeY, int readSSC,
                    int readISIandMBA,
                    int *IPB,          /* 0- P, 2- I */
                    vdxInstance_t *instance, int32 *bitError)
{
  int
    sentTR,
    bits = 0, mbNumLen, criticalParamIndicator, extendedHeaderIndicator, FrSz;

  u_int32
    code_u32;

  if (readSSC)
  {
    bits += vdxGetBits(18, buff_in, &code_u32, instance, bitError);
    if (code_u32 != 3) *bitError = VDX_SAW_SYNC;
    if(*bitError == VDX_SAW_SYNC) return 0;

    bits += vdxGetBits(1, buff_in, &code_u32, instance, bitError);
    if(*bitError == VDX_SAW_SYNC) return 0;
    sentTR = (int) code_u32;

    if (sentTR)
    {
      bits += vdxGetBits(8, buff_in, &code_u32, instance, bitError);
      if(*bitError == VDX_SAW_SYNC) return 0;

      if(code_u32 < (u_int32)PrevTR)
        *frNo = PrevFrNo + code_u32 + 256 - PrevTR;
      else
        *frNo = PrevFrNo + code_u32 - PrevTR;

      PrevTR = (int) code_u32;

      /*****
      PrevFrNo = *frNo; done outside in vdxCheckSSC()
      *******/
    }
    else {
      deb0f(stderr, "Found TRI set to 0 in this implementation");
      nccExit(-1);
    }

    /* PTYPE start */
    bits += (int32) vdxGetBits(1, buff_in, &code_u32, instance, bitError);
    if(!code_u32) *IPB=0; /* P frame */
    else{
      bits += (int32) vdxGetBits(1, buff_in, &code_u32, instance, bitError);
      if(!code_u32) *IPB=2; /* I frame */
      else{
          /* Reserved for future extensions */
          deb0f(stderr, "Unable to decode an extended PTYPE in slices\n");
          nccExit(1);
        }
      }
 

    if (*IPB == 2)  /* I frame */
    {
      bits += vdxGetBits(1, buff_in, &code_u32, instance, bitError);
      if(*bitError == VDX_SAW_SYNC) return 0;
      criticalParamIndicator = (int) code_u32;

      if (criticalParamIndicator)
      {
        bits += (int32) vdxGetBits(2, buff_in, &code_u32, instance, bitError);
        FrSz = (int) code_u32;

        switch(FrSz)
        {
        case 0:
          *picSizeX=176;
          *picSizeY=144;
          break;
        case 3:
          *picSizeX=352;
          *picSizeY=288;
          break;
        case 2:
          *picSizeX=128;
          *picSizeY=96;
          break;
        case 1:
          bits += (int32) vdxGetBits(7, buff_in, &code_u32, instance, bitError);
          *picSizeX = (int) (code_u32+1)*16;
          bits += (int32) vdxGetBits(7, buff_in, &code_u32, instance, bitError);
          *picSizeY = (int) (code_u32+1)*16;
          break;
        default:
          deb0f(stderr, 
                "Unrecognized picture format in extended slice header\n");
          nccExit(1);
          break;
        }

        bits += vdxGetBits(1, buff_in, &code_u32, instance, bitError);
        if(*bitError == VDX_SAW_SYNC) return 0;
        extendedHeaderIndicator = (int) code_u32;

        if (extendedHeaderIndicator) {
          deb0f(stderr, "Unsupported extended header in slices\n");
          nccExit(1);
        }
      }
    }

    bits += vdxGetBits(5, buff_in, &code_u32, instance, bitError);
    if(*bitError == VDX_SAW_SYNC) return 0;
    *sliceQuant = (int) code_u32;
  }

  if (readISIandMBA)
  {
    if (*picSizeX == 176 && *picSizeY == 144)
      mbNumLen = 7;
    else if (*picSizeX == 352 && *picSizeY == 288)
      mbNumLen = 9;
    else if (*picSizeX == 128 && *picSizeY == 96)
      mbNumLen = 6;
    else
      mbNumLen = 14;

    if (*IPB != 2)  /* P frame only */
    {
      bits += vdxGetBits(1, buff_in, &code_u32, instance, bitError);
      if(*bitError == VDX_SAW_SYNC) return 0;
      *intraSliceFlag = (int) code_u32;
    }
    else
      *intraSliceFlag = 1;

    bits += vdxGetBits(mbNumLen, buff_in, &code_u32, instance, bitError);
    if(*bitError == VDX_SAW_SYNC) return 0;
    *mbNum = (int) code_u32;
  }

  return bits;
}


int vdxOnlyZeroBits(vdxInstance_t *instance, int32 *bitError)
{
  u_int32
    code_u32;

  vdxShowBits(buff_in->bitsLeft, buff_in, &code_u32, instance);

  if (code_u32 == 0)
  {
    /*
     * Detected End-Of-Slice, followed by stuffing.
     * Just skip the stuffing bits
    */

    vdxFlushBits(buff_in->bitsLeft, buff_in, instance, bitError);

    return 1;
  }
  else
    return 0;
}


int vdxCheckSync(int *numStuffBits, vdxInstance_t *instance)
{
  u_int32 result;
  int numBitsGot, syncStartIndex;
  int16 newError = 0;
  
  *numStuffBits = 0;
  
  if (buff_in == NULL)
    return VDX_SNC_PSC;

  /* If in the beginning of the buffer */
  if (bibNumberOfFlushedBits(buff_in) == 0) {
    
    /* Check if the buffer begins with a PSC */
    
    result = bibShowBits(22, buff_in, &numBitsGot, &syncStartIndex, 
                         &newError);
    
    if (newError) {
      deb1f(stderr, "vdxCheckSync: bibShowBits indicated error %d.\n",
            newError);
      nccExit(1);
    }
    
    if (result == 16)
      /* PSC */
      return VDX_SNC_PSC;
    
    else
      return VDX_SNC_NO_SYNC;
  }
  
  /* Else expecting to be in the end of the buffer */
  else {
    int
      i,
      expNumStuffBits,
      numBitsLeftInFrame = bibNumberOfBitsLeft(buff_in),
      numEmulPreventBits = bibNumberOfEmulationPreventionBits(buff_in);
    
    /* If enough bits to carry an EOS */
    if (numBitsLeftInFrame >= 22) {
      
      /* If possible stuffing before EOS */
      if (numBitsLeftInFrame > 22) {
        expNumStuffBits = numBitsLeftInFrame - 22;
        if (expNumStuffBits > 7)
          expNumStuffBits = 7;
        
        result = bibShowBits(expNumStuffBits, buff_in, &numBitsGot, 
                             &syncStartIndex, &newError);
        
        if (newError) {
          deb1f(stderr, "vdxCheckSync: bibShowBits indicated error %d.\n",
                newError);
          nccExit(1);
        }
        
        /* If non-zero stuffing */
        if (result != 0)
          return VDX_SNC_NO_SYNC;
      }
      
      /* Else no stuffing bits before possible EOS */
      else
        expNumStuffBits = 0;
      
      result = bibShowBits(22 + expNumStuffBits, buff_in, &numBitsGot, 
                           &syncStartIndex, &newError);
      
      /* The EOS can be in any postion of the read code
         --> investigate all possibilites */
      for (i = expNumStuffBits; i >= 0; i--) {
        if (result == 31) {
          *numStuffBits = i;
          return VDX_SNC_EOS;
        }
        result >>= 1;
      }
      
      return VDX_SNC_NO_SYNC;
    }
    
    /* If too many bits left in the buffer */

    if (numBitsLeftInFrame > 7)
      return VDX_SNC_NO_SYNC;
    
    expNumStuffBits = numBitsLeftInFrame;
    
    /* If stuffing bits expected */
    if (expNumStuffBits) {
      result = bibShowBits(expNumStuffBits + 1, buff_in, &numBitsGot, 
                           &syncStartIndex, &newError);
      
      if (newError) {
        deb1f(stderr, "vdxCheckSync: bibShowBits indicated error %d.\n",
              newError);
        nccExit(1);
      }
      
      /* If non-zero stuffing */
      if (result != 0)
        return VDX_SNC_NO_SYNC;
      
      /* If a PSC starts as expected */
      if (syncStartIndex == expNumStuffBits + 1) {
        *numStuffBits = syncStartIndex - 1;
        return VDX_SNC_PSC;
      }
      else {
        deb0f(stderr, "vdxCheckSync: internal error in bit buffer.\n");
        nccExit(1);
      }
    }
    
    /* Else return an indication of a PSC */
    return VDX_SNC_PSC;
  }
}

int vdxClose(vdxInstance_t *instance)
{
  if (instance)
    nccFree(instance);

  return NCC_OK;
}

vdxInstance_t *vdxOpen(void)
{
  vdxInstance_t *instance;

  instance = nccMalloc(sizeof(vdxInstance_t));
  memset(instance, 0, sizeof(vdxInstance_t));

  return instance;
}

int vdxSetInputBuffer(vdxInstance_t *instance, bibBuffer_t* buffer)
{
  nccAssert(instance != NULL);
  nccAssert(buffer != NULL);

  instance->buffer = buffer;

  return NCC_OK;
}


int vdxCheckSSC(int *IPB,   /* 0- P, 2- I*/
                int *pic_sizex, int *pic_sizey,
                int *FrNo, int *skip, /* Distance from the previous frame    */
                int *QP, int32 *bits,
                vdxInstance_t *instance, int32 *bitError)
{
  u_int32
    code_u32;

  *bits = 0;

  vdxShowBits(18, buff_in, &code_u32, instance);

  if (code_u32 == 3)
  {
    /* SSC detected */

    *bits += vdxGetSliceHead(NULL, NULL, QP, FrNo, pic_sizex, pic_sizey,
                             1, 0, IPB, instance, bitError);
    *skip = *FrNo-PrevFrNo-1;
    PrevFrNo = *FrNo;

    if (*skip < 0) {
      deb0f(stderr, "Negative skip detected at beginning of new frame");
      nccExit(-2);
    }

    return 1;
  }
  else
    return 0;
}


int vdxGetMotionPredInfo(int predDirsAvail, int predSelNeededUp,
                         int predSelNeededLeft, u_char *predDir, int *predSel,
                         vdxInstance_t *instance, int32 *bitError)
{
  u_int32 bit_u32;
  int bits=0, predSelNeeded;

  if(predDirsAvail == PREDDIR_BOTH){
    bits += vdxGetBits(1, buff_in, &bit_u32, instance, bitError);
    if(bit_u32 == 0)
      *predDir = PREDDIR_UP;
    else
      *predDir = PREDDIR_LEFT;
  }

  if(*predDir == PREDDIR_UP)
    predSelNeeded = predSelNeededUp;
  else
    predSelNeeded = predSelNeededLeft;

  if(predSelNeeded == PREDSEL_NEEDED){
    bits += vdxGetBits(1, buff_in, &bit_u32, instance, bitError);
    if(bit_u32 == 0)
      *predSel = PREDSEL_1;
    else
      *predSel = PREDSEL_2;
  }

  return bits;
}

int vdxGetMotionCoeffs(int32 *motCoefOrth, int aQP,
                       vdxInstance_t *instance, int32 *bitError)
{
  int
    ind, i, bits = 0, QCoeff[6];

  const int th1   = 21;             /* Threshold for the 1st Escape code   */
  const int th2   = 85;             /* Threshold for the 2nd Escape code   */
  const int LEN_1 = 6;              /* Number of bits following 1st Escape */
  const int LEN_2 = 10;             /* Number of bits following 2nd Escape */
  const int LEN_TAB = 22;           /* Length of the Huff table for amplit */
  
  u_int32 
    sign_u32, code_u32, offset_u32;

  int32
    len;

  /* Get zero pattern */
  ind = vdxGetVlcItem(8, affZeroPatDec, &len, &code_u32, instance, 
                      bitError) + 1;

  if(*bitError == VDX_SAW_SYNC) return 0;

  bits +=len;

  for(i = 0; i < 6; i++)
    QCoeff[i] = (ind & (1 << i)) ? 1 : 0;

  /* Get nonzero affine motion parameters */
  for(i=0; i < 6; i++)
    if(QCoeff[i] != 0){

      ind = vdxGetVlcItem(8, affAmplDec, &len, &code_u32, 
                          instance, bitError);

      if(*bitError == VDX_SAW_SYNC) return 0;

      bits += len;
      ind++;

      /* First Escape + 6 bits */
      if(ind == th1){

        bits += vdxGetBits(LEN_1, buff_in, &offset_u32, instance, bitError);

        if(*bitError == VDX_SAW_SYNC) return 0;

        ind = th1 + (int)offset_u32;
      }

      /* Second Escape + 10 bits */
      else if (ind == th1+1){
        
        bits += vdxGetBits(LEN_2, buff_in, &offset_u32, instance, bitError);
        
        if(*bitError == VDX_SAW_SYNC) return 0;
        
        ind = th2 + (int)offset_u32;
      }
      
      else if(ind > th1){
        deb0f(stderr, "Unknown index in MCoeffs\n");
        deb1f(stderr, "ind=%d\n", ind);
        nccExit(1);
      }
      
      /* Sign bit */
      bits += vdxGetBits(1, buff_in, &sign_u32, instance, bitError);
      
      if(*bitError == VDX_SAW_SYNC) return 0;

      QCoeff[i] = sign_u32 ? -ind : ind;
    }

  /* Dequantization */
  for(i=0; i < 6; i++)
    motCoefOrth[i] = QCoeff[i]*aQP;

  return bits;
}

static int32 vdxGetPicStart(int32 *startOrEnd, vdxInstance_t *instance, 
                            int32 *bitError)
{
  int32
    bits, code;
  u_int32
    code_u32;

  bits=vdxGetBits(22, buff_in, &code_u32, instance, bitError);
  code = (int) code_u32;
  switch(code)
  {
  case 16:
    *startOrEnd=1;
    break;
  case 31:
    *startOrEnd=0;
    break;
  default:
    deb0f(stderr, "Neither PSC nor EOS\n");
    deb1f(stderr, "code %d\n", code);
    nccExit(1);
    /* This function is not suitable for search scan for PSC.
     * In such case bitstream is to be scanned byte by byte,
     * and not 22 bits at once.
     */
    break;
  }
  return bits;
}


int vdxGetRegionMode(int *regionMode, int *pecMode, int predPossible,
                     int regSize, vdxInstance_t *instance, int32 *bitError)
{
  u_int32 
    tmp, pecInd;

  int32 
    bits = 0, ind, maxLen;

  inv_vlc_item 
    *vlcTable;

  if(predPossible)
  {
    if(regSize == 16)
    {
      vlcTable = vlcDecRegionModePred16;
      maxLen   = 5;
    }
    else
    {
      vlcTable = vlcDecRegionModePred8;
      maxLen   = 4;
    }
  }
  else
  {
    if(regSize == 16)
    {
      vlcTable = vlcDecRegionModeNoPred16;
      maxLen   = 4;
    }
    else
    {
      vlcTable = vlcDecRegionModeNoPred8;
      maxLen   = 3;
    }
  }

  ind = vdxGetVlcItem(maxLen, vlcTable, &bits, &tmp, instance, bitError);

  if (*bitError == VDX_SAW_SYNC) 
    return 0;

  if(predPossible){
    switch(ind){
    case 0: *regionMode = REGMODE_INTER_NOPRED;     pecInd = 0; break;
    case 1: *regionMode = REGMODE_INTER_NOPRED;     pecInd = 1; break;
    case 2: *regionMode = REGMODE_INTER_PRED;       pecInd = 0; break;
    case 3: *regionMode = REGMODE_INTER_PRED;       pecInd = 1; break;
    case 4: *regionMode = REGMODE_INTER_PRED_DELTA; pecInd = 0; break;
    case 5: *regionMode = REGMODE_INTER_PRED_DELTA; pecInd = 1; break;
    case 6: *regionMode = REGMODE_INTER_COPY;       pecInd = 0; break;
    case 7: *regionMode = REGMODE_INTER_COPY;       pecInd = 1; break;
    case 8: *regionMode = REGMODE_INTRA;            pecInd = 0; break;
    case 9: *regionMode = REGMODE_SPLITTED;         pecInd = 0; break;
    }
  }
  else{
    switch(ind){
    case 0: *regionMode = REGMODE_INTER_NOPRED;     pecInd = 0; break;
    case 1: *regionMode = REGMODE_INTER_NOPRED;     pecInd = 1; break;
    case 2: *regionMode = REGMODE_INTER_COPY;       pecInd = 0; break;
    case 3: *regionMode = REGMODE_INTER_COPY;       pecInd = 1; break;
    case 4: *regionMode = REGMODE_INTRA;            pecInd = 0; break;
    case 5: *regionMode = REGMODE_SPLITTED;         pecInd = 0; break;
    }
  }

  *pecMode = pecInd ? PECMODE_CODED : PECMODE_SKIPPED;

  return bits;
}


int vdxDctGetCncInfo(u_char *cnc_y, u_char *cnc_u, u_char *cnc_v,
                     int regionMode, int regionSize,vdxInstance_t *instance, 
                     int32 *bitError)
{
  u_int32 
    tmp, ddump;

  int32 
    bits = 0, dump1, len;

  if (regionSize == 8)
  {
    if (regionMode == REGMODE_INTRA || regionMode == REGMODE_INTRA_8x8) 
    {
      ddump = vdxGetVlcItem(7, Intra8, &bits, &tmp, instance, bitError);
      if (*bitError == VDX_SAW_SYNC) 
        return 0;
    }
    else 
    {
      ddump = vdxGetVlcItem(5, vlcDecCncInter8, &bits, &tmp, instance, 
                            bitError);
      if (*bitError == VDX_SAW_SYNC) 
        return 0;
      ddump += 1;                        /* 0-0-0 is not in the codebook */
    }

    cnc_y[0] = (ddump & 4) >> 2;
    *cnc_u = (ddump & 2) >> 1;
    *cnc_v = ddump & 1;
  }
  else
  {
    if (regionMode == REGMODE_INTRA) 
    {
	  dump1 = vdxGetVlcItem(6, Intra64, &len, &tmp, instance, bitError);
	  if (*bitError == VDX_SAW_SYNC) 
        return 0;
      bits += len;
      if (dump1 == 18 /*escape index*/) 
      {
        bits += vdxGetBits(6, buff_in, &ddump/*tmp*/, instance, bitError);
        if (*bitError == VDX_SAW_SYNC) 
          return 0;
      }
      else
		ddump = invtableIntra[dump1];
    }
    else{
      dump1 = vdxGetVlcItem(5, vlcDecCncInter64, &len, &tmp, instance, 
									   bitError);
      if (*bitError == VDX_SAW_SYNC) 
        return 0;
      bits += len;
      if (dump1 == 15 /*escape index*/) 
      {
        bits += vdxGetBits(6, buff_in, &ddump, instance, bitError);
        if (*bitError == VDX_SAW_SYNC) 
          return 0;
	  }
      else
		ddump = invtableInter[dump1];
    }

    cnc_y[0] = (ddump & 32) >> 5;
    cnc_y[1] = (ddump & 16) >> 4;
    cnc_y[2] = (ddump & 8) >> 3;
    cnc_y[3] = (ddump & 4) >> 2;
    *cnc_u = (ddump & 2) >> 1;
    *cnc_v = ddump & 1;
  }

  return bits;
}
