/*
 * decMCC.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Decoding functions for reconstructing the prediction error from 
 *    the coding information.
 *
 */

#include "decMCC.h"

extern void copyBlock8x8(int *source, int *destination, int rotateFlag);
extern void copyBlock4x4(int *source, int *destination, int rotateFlag);
extern void varSizeInvDctFixedp( int *source, int *destination, int cols, int rows );

/*
 * Define extern tables.
 */
extern u_char
  zig4x8[SIZE_8x4],
  zig4x4[SIZE_4x4],
  zig4x4Hori[SIZE_4x4],
  zig4x4Vert[SIZE_4x4],
  zig4x4Intra[SIZE_4x4],
  zig4x4IntraHori[SIZE_4x4],
  zig4x4IntraVert[SIZE_4x4],
  zig8x8[SIZE_8x8],
  zig8x8Hori[SIZE_8x8],
  zig8x8Vert[SIZE_8x8],
  zig8x8Intra[SIZE_8x8],
  zig8x8IntraHori[SIZE_8x8],
  zig8x8IntraVert[SIZE_8x8];

extern u_char
  mirrorVert4x4[SIZE_4x4], 
  mirrorDiag4x4[SIZE_4x4],
  rotateLeft4x4[SIZE_4x4], 
  rotateRight4x4[SIZE_4x4];

extern u_char
  mirrorVert8x8[SIZE_8x8],
  mirrorDiag8x8[SIZE_8x8],
  rotateLeft8x8[SIZE_8x8], 
  rotateRight8x8[SIZE_8x8];

extern u_char
  kltLookUpTable[10];

/*
 * Define local function prototypes.
 */

static int mcdDec8x8Dct(int *dctCoeffs, int *recoBlock, int qp, int chroma,
                        int mode, int methNumber);

static int mcdDec8x8Klt(int *KltCoeffs, int *recoBlock, int qp, int chroma,
                        int mode, int methNumber);

static void mcdDecDct(int *qcoeffs, int *recoBlock, int qp, int cols, int rows,
                      int chroma, int mode);

static int mcdDec8x4Dct(int *qCoeffs, int *recoBlock, int dctMeth,
                        int qp, int chroma, int mode);

static void mcdDec4x4Klt(int *kltCoeffs, int *recoBlock, int methNumber,
                         int qp, int chroma, int mode);

static int mcdDec4x4Dct(int *dctCoeffs, int *recoBlock, int qp, int chroma,
                        int mode, int dctMet);

static void mcdDecExtrapol(int *predBlock, int *recoBlock,
                           int methNumber);

static void mcdExtrapolate(int *predBlock, int *recoBlock,
                           int *extrapSource, int pels, int sizeOfBlk);


/*
 * dec8x8:
 *
 * Parameters:
 *      predBlock        I: Prediction block pointer
 *      recoBlock        O: Reconstruction block pointer
 *      methCoeffs       I: Transform coefficients for 8x8 block
 *      dctCoeffs        I: Transform coefficients of 4x4 blocks
 *      qp               I: Quantization parameter value
 *      methNumber       I: Used 8x8 method 
 *      class            I: Real 8x8 blocks prediction class
 *      chroma           I: Block type (luma or chroma)
 *      mode             I: Mode of block (intra/inter)
 *      predDirIn        I: Prediction direction number
 *
 * Function:
 *      Calculates reconstruction block for 8x8 method depending the
 *      information given in parameters.
 *
 * Returns:
 *      bits            Amount of bits used in 8x8 method.
 */

int dec8x8(int *predBlock, int *recoBlock, int *methCoeffs, int *dctCoeffs,
           int qp, int methNumber, int class, int chroma, int mode,
           u_char predDirIn)
{
  int
    i,		                    /* Loop counter */
    bits = 0,                 /* Amount of bits */
    predNumber,               /* 8x8 methods class after rotations */
    rotation,                 /* Rotation Flag value */
    pels,                     /* Amount of pels in extrapSource */
    tmpPred[SIZE_8x8],        /* Temporal prediction block */
    tmpReco[SIZE_8x8],        /* Temporal reconstruction block */
    tmpBlock[SIZE_8x8],       /* Temporal block */
    subReco[SIZE_4x8],        /* Reconstruction blocks subblock */
    subPred[SIZE_4x4],        /* Prediction blocks subblock */
    extrapSource[9];          /* The prediction block values, where
                                 the extrapolated value is calculated */

  /*
   * Initialize reconstruction block with 0.
   */
  for (i = 0; i < SIZE_8x8; i++)
    tmpReco[i] = 0;

  /*
   * Classify the block according to coding mode (INTER,INTRA), QP 
   * and component (Y,u/v). INTRA blocks are also classified according
   * to directionality of the prediction block using gradient based
   * classification. Check also if rotation is used during encoding.
   */
  if (predDirIn == UCHAR_MAX || mode != INTER_BLOCK)
    blockEdge(predBlock, 8, 8, &predDirIn);

  predNumber = changeClassNew(class, predDirIn, qp, chroma, mode, &rotation);

  /*
   * Perform decoding.
   */
  switch (methNumber) {

    /*
     * 8x8 DCT.
     */
    case METH_8x8_DCT: case METH_DCT_8x8_N_0_8: case METH_DCT_8x8_N_4_8:

      bits += mcdDec8x8Dct(methCoeffs, tmpReco, qp, chroma, mode,
                           methNumber);
      break;

    /*
     * 8x8 KLT.
     */
    case METH_KLT_8x8_N_1_8: case METH_KLT_8x8_N_2_8:
    case METH_KLT_8x8_N_3_8: case METH_KLT_8x8_N_5_8:
    case METH_KLT_8x8_N_6_8: case METH_KLT_8x8_N_7_8:

      bits += mcdDec8x8Klt(methCoeffs, tmpReco, qp, chroma, mode, 
                           methNumber);
      break;

    /* 
     * Vertical and horizontal 8x4 DCT methods. Reconstruction of 
     * horizontal DCT methods is rotated later since only vertical. 
     */
    case METH_4x8_DCT_1: case METH_4x8_DCT_2: case METH_4x8_DCT_3:
    case METH_8x4_DCT_1: case METH_8x4_DCT_2: case METH_8x4_DCT_3:
    case METH_4x8_DCT_4: case METH_8x4_DCT_4:

		  bits += mcdDec8x4Dct(methCoeffs, tmpReco, methNumber, qp, chroma, mode);
      break;

    /*
     * Extralpolation.
     */
    case METH_EXTRAPOL_12: case METH_EXTRAPOL_14:
    case METH_EXTRAPOL_28: case METH_EXTRAPOL_48:

      /*
       * Rotate/mirror 8x8 prediction block if needed.
       */
      copyBlock8x8(predBlock, tmpPred, rotation);

	    /* 
       * Calculate extrapolated reconstruction block for given prediction 
	     * block.
	     */
      mcdDecExtrapol(tmpPred, tmpReco, methNumber);
      break;

    /* 
     * 4x4 quadrants coded separately.
     */
    case METH_4x4_15:

      /*
       * Rotate/mirror the prediction block if needed.
       */
      copyBlock8x8(predBlock, tmpPred, rotation);

      /*
       * Get upper left quadrant of the 8x8 block.
       */
		  getIntBlock(tmpPred, subPred, COLS_4x4, ROWS_4x4,
                  START_COL_4x4_1, START_ROW_4x4_1, COLS_8x8);

      /*
       * Find the pixels for extrapolation if needed.
       */
		  if (methCoeffs[10] == EXTRAPOL_METH)
			  pels = getExtSou8x8(tmpPred, extrapSource, QUAD_1);

      bits += dec4x4(subPred, subReco, methCoeffs, dctCoeffs, qp,
                     extrapSource, pels, chroma, mode, SIZE_8x8);

      /*
       * Put the reconstructed quadrant to 8x8 block.
       */
      putIntBlock(subReco, tmpReco, COLS_4x4, ROWS_4x4,
                  START_COL_4x4_1, START_ROW_4x4_1, COLS_8x8);

      /*
       * Get upper right quadrant of the 8x8 block.
       */
		  getIntBlock(tmpPred, subPred, COLS_4x4, ROWS_4x4,
                  START_COL_4x4_2, START_ROW_4x4_2, COLS_8x8);

      /*
       * Find the pixels for extrapolation if needed.
       */
		  if (methCoeffs[START_VAL_4x4_2 + 10] == EXTRAPOL_METH)
			  pels = getExtSou8x8(tmpPred, extrapSource, QUAD_2);

      bits += dec4x4(subPred, subReco, methCoeffs + START_VAL_4x4_2, 
                     dctCoeffs + START_VAL_4x4_2, qp,
                     extrapSource, pels, chroma, mode, SIZE_8x8);
                     
      /*
       * Put the reconstructed quadrant to 8x8 block.
       */
		  putIntBlock(subReco, tmpReco, COLS_4x4, ROWS_4x4,
                  START_COL_4x4_2, START_ROW_4x4_2, COLS_8x8);

      /*
       * Get lower left quadrant of the 8x8 block.
       */
		  getIntBlock(tmpPred, subPred, COLS_4x4, ROWS_4x4,
                  START_COL_4x4_3, START_ROW_4x4_3, COLS_8x8);

      /*
       * Find the pixels for extrapolation if needed.
       */
		  if (methCoeffs[START_VAL_4x4_3 + 10] == EXTRAPOL_METH)
			  pels = getExtSou8x8(tmpPred, extrapSource, QUAD_3);

      bits += dec4x4(subPred, subReco, methCoeffs + START_VAL_4x4_3, 
                     dctCoeffs + START_VAL_4x4_3, qp,
                     extrapSource, pels, chroma, mode, SIZE_8x8);

      /*
       * Put the reconstructed quadrant to 8x8 block.
       */
      putIntBlock(subReco, tmpReco, COLS_4x4, ROWS_4x4,
                  START_COL_4x4_3, START_ROW_4x4_3, COLS_8x8);

      /*
       * Get lower right quadrant of the 8x8 block.
       */
		  getIntBlock(tmpPred, subPred, COLS_4x4, ROWS_4x4,
                  START_COL_4x4_4, START_ROW_4x4_4, COLS_8x8);

      /*
       * Find the pixels for extrapolation if needed.
       */
		  if (methCoeffs[START_VAL_4x4_4 + 10] == EXTRAPOL_METH)
			  pels = getExtSou8x8(tmpPred, extrapSource, QUAD_4);

      bits += dec4x4(subPred, subReco, methCoeffs + START_VAL_4x4_4, 
                     dctCoeffs + START_VAL_4x4_4, qp,
                     extrapSource, pels, chroma, mode, SIZE_8x8);

      /*
       * Put the reconstructed quadrant to 8x8 block.
       */
      putIntBlock(subReco, tmpReco, COLS_4x4, ROWS_4x4,
                  START_COL_4x4_4, START_ROW_4x4_4, COLS_8x8);
  }

  /*
   * The rotation of reconstruction block is needed if 
   * the selected method is horizontal 8x4 DCT.
   */
  if (methNumber >= METH_8x4_DCT_1 && methNumber <= METH_8x4_DCT_4) 
    for (i = 0; i < SIZE_8x8; i++)
      tmpBlock[i] = tmpReco[rotateRight8x8[i]];

  else
    memcpy(tmpBlock, tmpReco, sizeof(int) << 6);

  /*
   * Rotate/mirror the reconstruction block if needed.
   * ROTATE_LEFT and ROTATE_RIGHT are inverse operations,
   * all other operations work also for inverse. 
   */
  switch (rotation) {
    
    case ROTATE_LEFT:
      copyBlock8x8(tmpBlock, recoBlock, ROTATE_RIGHT);
      break;

    case ROTATE_RIGHT:
      copyBlock8x8(tmpBlock, recoBlock, ROTATE_LEFT);
      break;

    default:
      copyBlock8x8(tmpBlock, recoBlock, rotation);
  }

  return bits;
}


/*
 * dec4x4:
 *
 * Parameters:
 *      predBlock        I: Prediction block pointer
 *      recoBlock        O: Reconstruction block pointer
 *      methInfo         I: Used 4x4 method Information pointer
 *      dctCoeffs        I: Dct coefficients of 4x4 block
 *      qp               I: Quantization parameter value
 *      extrapSource     I: Pels used in extrapolation
 *      pels             I: Amount of pels in extrapSource
 *      chroma           I: Block type (luma or chroma)
 *      mode             I: Mode of block (intra/inter)
 *      blkSize          I: Size of coded block
 *
 * Function:
 *      Calculates reconstruction block for 4x4 method depending the
 *      information given in parameters.
 *
 * Returns:
 *      bits            Amount of bits used in 4x4 method.
 */

int dec4x4(int *predBlock, int *recoBlock, int *methInfo, int *dctCoeffs,
           int qp, int *extrapSource, int pels, int chroma, int mode,
           int blkSize)
{
  int
    i,		                      /* Loop counter */
    bits = 0,                   /* Amount of bits used in 4x4 method */
    methNumber,                 /* Used 4x4 method number */
    rotation,                   /* Rotation flag */
    tmpReco[SIZE_4x4],          /* Temporal reconstruction block */
    tmpPred[SIZE_4x4];          /* Temporal prediction block */

  /*
   * Find the information about the coding method and
   * possible rotation/mirroring of the final reconstruction block.
   */
  methNumber = methInfo[10];
  rotation = methInfo[14];

  /*
   * Perform decoding
   */
  switch (methNumber) {

    /*
     * NOT_CODED method.
     */
    case NOTCODED_METH:
      for (i = 0; i < SIZE_4x4; i++)
        recoBlock[i] = 0;
      break;

    /*
     * 4x4 DCT.
     */
    case METH_DCT_4x4: case METH_DCT_4x4_N_0_8:
    case METH_DCT_4x4_N_4_8:
      mcdDec4x4Dct(dctCoeffs, tmpReco, qp, chroma, mode, methNumber);
      break;

    /*
     * 4x4 extrapolation.
     */
    case EXTRAPOL_METH:
      copyBlock4x4(predBlock, tmpPred, rotation);
      mcdExtrapolate(tmpPred, tmpReco, extrapSource, pels, SIZE_4x4);
      break;

    /*
     * 4x4 KLT.
     */
    case KLT_4x4_N_1_8: case KLT_4x4_N_2_8:
    case KLT_4x4_N_3_8: case KLT_4x4_N_5_8:
    case KLT_4x4_N_6_8: case KLT_4x4_N_7_8:
      mcdDec4x4Klt(dctCoeffs, tmpReco, methNumber, qp, chroma, mode);
      break;
  }

  /*
   * Rotate/mirror the reconstruction block if needed.
   */
  if (methNumber != NOTCODED_METH)
  /*
   * Rotate/mirror the reconstruction block if needed.
   * ROTATE_LEFT and ROTATE_RIGHT are inverse operations,
   * all other operations work also for inverse. 
   */
  switch (rotation) {
    
    case ROTATE_LEFT:
      copyBlock4x4(tmpReco, recoBlock, ROTATE_RIGHT);
      break;

    case ROTATE_RIGHT:
      copyBlock4x4(tmpReco, recoBlock, ROTATE_LEFT);
      break;

    default:
      copyBlock4x4(tmpReco, recoBlock, rotation);
  }

  return bits;
}


/*
 * mcdDec8x8Dct:
 *
 * Parameters:
 *      dctCoeffs       I: Quantized dct coefficients in scanning order
 *      recoBlock       O: Reconstruction block pointer
 *      qp              I: Quantization Parameter value
 *      chroma          I: Mode of block (CHROMA or LUMA) 
 *      mode            I: Mode of block (INTRA_BLOCK or INTER_BLOCK) 
 *      methNumber      I: Number of dct method
 *
 * Function:
 *      Finds 8x8 DCT reconstruction block for given DCT 
 *      coefficients. 
 *
 * Returns:
 *      void
 */

static int mcdDec8x8Dct(int *dctCoeffs, int *recoBlock, int qp, int chroma,
                        int mode, int methNumber)
{
  int
    i,							            /* Loop counter */
    bits = 0,                   /* Amount of bits */
    invCoeffs[SIZE_8x8],        /* Inverse quantized coefficients */
    tmpCoeffs[SIZE_8x8];        /* Inverse zigzag-scanned coefficients */

  u_char
    *zigZagTable;               /* Scaning order table */

  /*
   * Scan the DCT coefficients.
   */
  switch (methNumber) {      

    case METH_8x8_DCT:
      zigZagTable = mode == INTER_BLOCK ? zig8x8 : zig8x8Intra;
    	break;

    case METH_DCT_8x8_N_0_8:
      zigZagTable = mode == INTER_BLOCK ? zig8x8Hori : zig8x8IntraHori;
    	break;

    default:
      zigZagTable = mode == INTER_BLOCK ? zig8x8Vert : zig8x8IntraVert;
  }
  for (i = 0; i < SIZE_8x8; i++)
    tmpCoeffs[i] = dctCoeffs[zigZagTable[i]];

  /*
   * Perform inverse quantization.
   */
  MCCInvQuantCoef(tmpCoeffs, invCoeffs, SIZE_8x8, qp, chroma, mode);

  /*
   * Calculate IDCT.
   */
  varSizeInvDctFixedp(invCoeffs, recoBlock, 8, 8);  

  return bits;
}


/*
 * mcdDec8x8Klt:
 *
 * Parameters:
 *      kltCoeffs       I: Quantized klt coefficients
 *      recoBlock       O: Reconstruction block pointer
 *      qp              I: Quantization Parameter value
 *      chroma          I: Mode of block (CHROMA or LUMA) 
 *      mode            I: Mode of block (INTRA_BLOCK or INTER_BLOCK) 
 *      methNumber      I: Number of klt method
 *      dcDiffInd       I: DC difference index
 *
 * Function:
 *      Finds 8x8 KLT reconstruction block for given KLT method
 *      and coefficients. 
 *
 * Returns:
 *      void
 */

static int mcdDec8x8Klt(int *kltCoeffs, int *recoBlock, int qp, int chroma,
                        int mode, int methNumber)
{
  int
    i,							      /* Loop counter */
    bits = 0,             /* Amount of bits */
    invCoeffs[SIZE_8x8],  /* Inverse quantized coefficients */
    kltBlock[SIZE_8x8],   /* IKLT block */
    kltMet,               /* Number of KLT Method */
    kltNum,               /* Number of KLT basic function set */
    rotation;             /* Rotation flag */

  /*
   * Selecet the correct set of KLT basis functions.
   */
  switch (methNumber) {

    case METH_KLT_8x8_N_1_8:
	    rotation = UNIT_OPER;
      kltMet = METH_KLT_8x8_N_1_8;
	    break;

    case METH_KLT_8x8_N_2_8:
	    rotation = UNIT_OPER;
      kltMet = METH_KLT_8x8_N_2_8;
	    break;

    case METH_KLT_8x8_N_3_8:
	    rotation = MIRROR_DIAG;
      kltMet = METH_KLT_8x8_N_1_8;
	    break;

    case METH_KLT_8x8_N_5_8:
	    rotation = ROTATE_LEFT;
	    kltMet = METH_KLT_8x8_N_1_8;
	    break;

    case METH_KLT_8x8_N_6_8:
	    rotation = MIRROR_VERT;
	    kltMet = METH_KLT_8x8_N_2_8;
	    break;

    case METH_KLT_8x8_N_7_8:
	    rotation = MIRROR_VERT;
	    kltMet = METH_KLT_8x8_N_1_8;
	    break;
  }

  kltNum = kltLookUpTable[kltMet - METH_KLT_8x8_N_0_8];

  /*
   * Perform inverse quantization.
   */
  MCCInvQuantCoef(kltCoeffs, invCoeffs, SIZE_8x8, qp, chroma, mode);

  /*
   * Calculate IKLT.
   */
  KltInverseTransform8x8(invCoeffs, kltBlock, kltNum, mode);

  /*
   * Rotate/mirror the result of IKLT in order to obtain the actual
   * reconstruction. This is needed, since only two sets of basis
   * functions (i.e. -22,5 deg and -45 deg) are implemented.
   */
  switch (rotation) {

    case UNIT_OPER:	
      memcpy(recoBlock, kltBlock, sizeof(int) << 6);
      break;

    case ROTATE_LEFT:	
      for (i = 0; i < SIZE_8x8; i++)
        recoBlock[i] = kltBlock[rotateRight8x8[i]];
      break;

    case MIRROR_VERT:	
      for (i = 0; i < SIZE_8x8; i++)
			  recoBlock[i] = kltBlock[mirrorVert8x8[i]];
		  break;

    case MIRROR_DIAG:	
      for (i = 0; i < SIZE_8x8; i++)
			  recoBlock[i] = kltBlock[mirrorDiag8x8[i]];
		  break;
  }

  return bits;
}


/*
 * mcdDec8x4Dct:
 *
 * Parameters:
 *      qCoeffs         I: Quantized dct coefficients pointer
 *      recoBlock       O: Reconstruction block pointer
 *      dctMeth         I: Number of DCT method
 *      qp              I: Quantization Parameter value
 *      chroma          I: Mode of block (CHROMA or LUMA) 
 *      mode            I: Mode of block (INTRA_BLOCK or INTER_BLOCK) 
 *
 * Function:
 *      Finds 8x4 DCT reconstruction block for given DCT 
 *      coefficients. 
 *
 * Returns:
 *      void
 */

static int mcdDec8x4Dct(int *qCoeffs, int *recoBlock, int dctMeth,
                        int qp, int chroma, int mode)
{
  int
    bits = 0,
    i,
    invCoeffs[SIZE_4x8],        /* Inverse quantized coefficients */
    tmpCoeffs[SIZE_4x8],        /* Inverse zigzag-scanned coefficients */
    tmpReco[SIZE_4x8],          /* IDCT block */
    blockHalf,                  /* Counter for block halves*/
    codedHalves,                /* Flag for coded halves */
    offSet;                     /* Start of DCT coefficients */

  switch (dctMeth) {

    /*
     * 8x4 DCT applied on upper/left or(and) lower/right halves
     * of the 8x8 block.
     */
    case METH_4x8_DCT_1: case METH_4x8_DCT_2: case METH_4x8_DCT_3:
    case METH_8x4_DCT_1: case METH_8x4_DCT_2: case METH_8x4_DCT_3:
      
      /*
       * Check which halves are coded (left/upper and right/lower).
       */ 
      if (dctMeth >= METH_8x4_DCT_1 && dctMeth <= METH_8x4_DCT_3)
        codedHalves = dctMeth - METH_8x4_DCT_1 + 1;
      else
        codedHalves = dctMeth - METH_4x8_DCT_1 + 1;

	    /*
	     * Go through the halves.
	     */
      for (blockHalf = 0; blockHalf < 2; blockHalf++) {

	      if (blockHalf == 0)

          /*
           * Start of DCT coefficients for the left/upper half.
           */
		      offSet = (codedHalves & 1) ? START_VAL_4x8_1 : -1;

	      else

          /*
           * Start of DCT coefficients for the right/lower half.
           */
		      offSet = (codedHalves & 2) ? START_VAL_4x8_2 : -1;

	      if (offSet != -1) {
          /*
           * Scan the coefficients if the half is coded.
           */
		      for (i = 0; i < SIZE_4x8; i++) 
			      tmpCoeffs[i] = qCoeffs[offSet + zig4x8[i]];

          /*
           * Perform inverse quantization.
           */
          MCCInvQuantCoef(tmpCoeffs, invCoeffs, SIZE_4x8, qp, chroma, mode);

          /*
           * Calculate IDCT.
           */
          varSizeInvDctFixedp(invCoeffs, tmpReco, COLS_4x8, ROWS_4x8);

          /*
           * Copy IDCT result to reconstruction block.
           */
		      if (offSet == START_VAL_4x8_1)
			      putIntBlock(tmpReco, recoBlock, COLS_4x8, ROWS_4x8,
                        START_COL_4x8_1, START_ROW_4x8_1, COLS_8x8);

          else
			      putIntBlock(tmpReco, recoBlock, COLS_4x8, ROWS_4x8,
                        START_COL_4x8_2, START_ROW_4x8_2, COLS_8x8);
	      }
      }
      break;

    /* 
     * 8x4 DCT applied on horizontal or vertical part in the
     * middle of the 8x8 block.
     */
    case METH_4x8_DCT_4: case METH_8x4_DCT_4:
      for (i = 0; i < SIZE_4x8; i++)
        tmpCoeffs[i] = qCoeffs[START_VAL_4x8_4 + zig4x8[i]];

        /*
         * Perform inverse quantization.
         */
        MCCInvQuantCoef(tmpCoeffs, invCoeffs, SIZE_4x8, qp, chroma, mode);

        /*
         * Calculate IDCT.
         */
        varSizeInvDctFixedp(invCoeffs, tmpReco, COLS_4x8, ROWS_4x8);

        /*
         * Copy IDCT result to reconstruction block.
         */
	      putIntBlock(tmpReco, recoBlock, COLS_4x8, ROWS_4x8,
                    START_COL_4x8_4, START_ROW_4x8_4, COLS_8x8);
      break;
  }
  return bits;
}

/*
 * mcdDec4x4Klt:
 *
 * Parameters:
 *      kltCoeffs       I: Klt coefficient pointer
 *      recoBlock       O: Reconstruction block pointer
 *      methNumber      I: Number which shows used 4x4 KLT method
 *      qp              I: Quantization Parameter value
 *      chroma          I: Mode of block (CHROMA or LUMA) 
 *      mode            I: Mode of block (INTRA_BLOCK or INTER_BLOCK) 
 *
 * Function:
 *      Finds 4x4 KLT reconstruction block for given KLT method
 *      and coefficients. 
 *
 * Returns:
 *      void
 */

static void mcdDec4x4Klt(int *kltCoeffs, int *recoBlock, int methNumber,
                         int qp, int chroma, int mode)
{
  int
    i,                          /* Loop counter */
    kltNum,                     /* Number of KLT basic function set */
    invCoeffs[SIZE_4x4],        /* Inverse quantized coefficients */
    kltBlock[SIZE_4x4];         /* IKLT block */

  /*
   * Select the correct KLT basis function set.
   */
  kltNum = kltLookUpTable[methNumber - KLT_4x4_N_0_8];

  /*
   * Perform inverse quantization.
   */
  MCCInvQuantCoef(kltCoeffs, invCoeffs, SIZE_4x4, qp, chroma, mode);

  /*
   * Calculate IKLT.
   */
  KltInverseTransform4x4(invCoeffs, kltBlock, kltNum, mode);
  
  /*
   * Rotate/mirror the result of IKLT in order to obtain the actual
   * reconstruction. This is needed, since only two sets of basis
   * functions are implemented.
   */
  switch (methNumber) {

    case KLT_4x4_N_1_8: case KLT_4x4_N_2_8:
		  memcpy(recoBlock, kltBlock, sizeof(int) * SIZE_4x4);
		  break;

    case KLT_4x4_N_3_8:
		  for (i = 0; i < SIZE_4x4; i++)
			  recoBlock[i] = kltBlock[mirrorDiag4x4[i]];
		  break;

    case KLT_4x4_N_5_8:
		  for (i = 0; i < SIZE_4x4; i++)
			  recoBlock[i] = kltBlock[rotateRight4x4[i]];
		  break;

    case KLT_4x4_N_6_8: case KLT_4x4_N_7_8:
		  for (i = 0; i < SIZE_4x4; i++)
			  recoBlock[i] = kltBlock[mirrorVert4x4[i]];
		  break;
  }
}


/*
 * mcdDec4x4Dct:
 *
 * Parameters:
 *      dctCoeffs       I: Quantized dct coefficients in scanning order
 *      recoBlock       O: Reconstruction block pointer
 *      qp              I: Quantization Parameter value
 *      chroma          I: Mode of block (CHROMA or LUMA) 
 *      mode            I: Mode of block (INTRA_BLOCK or INTER_BLOCK) 
 *      methNumber      I: Number of DCT method
 *
 * Function:
 *      Finds 4x4 DCT reconstruction block for given DCT 
 *      coefficients. 
 *
 * Returns:
 *      void
 */

static int mcdDec4x4Dct(int *dctCoeffs, int *recoBlock, int qp, int chroma,
                        int mode, int dctMet)
{
  int
    i,						         	/* Loop counter */
    bits = 0,               /* Amount of bits */
    invCoeffs[SIZE_4x4],    /* Inverse quantized coefficients */
    tmpCoeffs[SIZE_4x4];    /* Inverse zigzag-scanned coefficients */
  
  u_char
    *zigZagTable;           /* Scaning order table */

  /*
   * Scan the DCT coefficients.
   */
  switch (dctMet) {

    case METH_DCT_4x4:
      zigZagTable = zig4x4;
	    break;

    case METH_DCT_4x4_N_0_8:
      zigZagTable = zig4x4Hori;
	    break;

    default:
      zigZagTable = zig4x4Vert;
	    break;
  }
  for (i = 0; i < SIZE_4x4; i++)
    tmpCoeffs[i] = dctCoeffs[zigZagTable[i]];

  /*
   * Perform inverse quantization.
   */
  MCCInvQuantCoef(tmpCoeffs, invCoeffs, SIZE_4x4, qp, chroma, mode);

  /*
   * Calculate IDCT.
   */
  varSizeInvDctFixedp(invCoeffs, recoBlock, COLS_4x4, ROWS_4x4);

  return bits;
}


/*
 * mcdDecExtrapol:
 *
 * Parameters:
 *      predBlock       I: Prediction block pointer
 *      recoBlock       O: Reconstruction block pointer
 *      methNumber      I: Number which shows used 8x8 extrapolation method
 *
 * Function:
 *      Calculates extrapolated reconstruction block for given prediction
 *      block.
 *
 * Returns:
 *      void
 */

static void mcdDecExtrapol(int *predBlock, int *recoBlock,
                           int methNumber)
{
  int 
    i,                /* Increment loop counter */
    pred[SIZE_8x4],   /* Prediction block before extrapolation */
    diff[SIZE_8x4],   /* Difference block between extrapolated value
                         and Prediction block */ 
    extrapSource[9],  /* Pixel values for extrapolation */
    pels;             /* Amount of pixels in the extrapSource buffer */

  /*
   * Initialize reconstruction block with 0.
   */
  for (i = 0; i < SIZE_8x8; i++)
    recoBlock[i] = 0;

  switch (methNumber) {

    /*
     * 8x4 extrapolation applied for upper half of the block.
     */
    case METH_EXTRAPOL_12:

      /*
       * Take 8x4 prediction block.
       */
      getIntBlock(predBlock, pred, COLS_8x4, ROWS_8x4,
                  START_COL_8x4_1, START_ROW_8x4_1, ROWS_8x8);

      /*
       * Take the pels which are used in extrapolation value calculation and
       * produce difference block between prediction and extrapolation value.
       */
      pels = getExtSou8x8(predBlock, extrapSource, HALF_1);
      mcdExtrapolate(pred, diff, extrapSource, pels, SIZE_8x4);

      /*
       * Put difference block into the reconstruction block.
       */
      putIntBlock(diff, recoBlock, COLS_8x4, ROWS_8x4,
                  START_COL_8x4_1, START_ROW_8x4_1, ROWS_8x8);
  	  break;

    /*
     * 8x4 extrapolation applied for left half of the block.
     */
    case METH_EXTRAPOL_14:

      /*
       * Take 4x8 prediction block.
       */
      getIntBlock(predBlock, pred, COLS_4x8, ROWS_4x8,
                  START_COL_4x8_1, START_ROW_4x8_1, ROWS_8x8);

      /*
       * Take the pels which are used in extrapolation value calculation and
       * produce difference block between prediction and extrapolation value.
       */
      pels = getExtSou8x8(predBlock, extrapSource, HALF_3);
      mcdExtrapolate(pred, diff, extrapSource, pels, SIZE_4x8);

      /*
       * Put difference block into the reconstruction block.
       */
      putIntBlock(diff, recoBlock, COLS_4x8, ROWS_4x8,
                  START_COL_4x8_1, START_ROW_4x8_1, ROWS_8x8);
  	  break;

    /*
     * 8x4 extrapolation applied for right half of the block.
     */
    case METH_EXTRAPOL_28:

      /*
       * Take 4x8 prediction block.
       */
      getIntBlock(predBlock, pred, COLS_4x8, ROWS_4x8,
                  START_COL_4x8_2, START_ROW_4x8_2, ROWS_8x8);

      /*
       * Take the pels which are used in extrapolation value calculation and
       * produce difference block between prediction and extrapolation value.
       */
      pels = getExtSou8x8(predBlock, extrapSource, HALF_4);
      mcdExtrapolate(pred, diff, extrapSource, pels, SIZE_4x8);

      /*
       * Put difference block into the reconstruction block.
       */
      putIntBlock(diff, recoBlock, COLS_4x8, ROWS_4x8,
                  START_COL_4x8_2, START_ROW_4x8_2, ROWS_8x8);
	    break;

    /*
     * 8x4 extrapolation applied for lower half of the block.
     */
    case METH_EXTRAPOL_48:

      /*
       * Take 8x4 prediction block.
       */
      getIntBlock(predBlock, pred, COLS_8x4, ROWS_8x4,
                  START_COL_8x4_2, START_ROW_8x4_2, ROWS_8x8);

      /*
       * Take the pels which are used in extrapolation value calculation and
       * produce difference block between prediction and extrapolation value.
       */
      pels = getExtSou8x8(predBlock, extrapSource, HALF_2);
      mcdExtrapolate(pred, diff, extrapSource, pels, SIZE_8x4);

      /*
       * Put difference block into the reconstruction block.
       */
      putIntBlock(diff, recoBlock, COLS_8x4, ROWS_8x4,
                  START_COL_8x4_2, START_ROW_8x4_2, ROWS_8x8);
	    break;
  }
}


/*
 * mcdExtrapolate:
 *
 * Parameters:
 *      predBlock       I: Prediction block pointer
 *      recoBlock       O: Reconstruction block pointer
 *      extrapSource    I: Pels which are used in extrapolation value 
 *                         definition
 *      pels            I: Amount of pels in extrapSource
 *      sizeOfBlk       I: Size of prediction block
 *
 * Function:
 *      Calculates extrapolated reconstruction block for given prediction
 *      block and extrapolation source.
 *
 * Returns:
 *      void
 */

static void mcdExtrapolate(int *predBlock, int *recoBlock,
                           int *extrapSource, int pels, int sizeOfBlk)
{
  int 
    i,                /* Increment loop counter */
    value;            /* Extrapolated value */

  /*
   * Calculate extrapolation value and produce the reconstruction block.
   */
  value = median(extrapSource, pels);
  for (i = 0; i < sizeOfBlk; i++)
    recoBlock[i] = value - predBlock[i];
}


