/*
 * decoder.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Main decoder and auxiliary bitstream manipulation functions
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <time.h>

#include "nccglob.h"
#include "mainFlags.h"
#include "mvc.h"
#include "deblockFilt.h"
#include "IOimg.h"
#include "IntraPredDecoder.h"
#include "decCore.h"

#include "bib.h"
#include "decDemux.h"

/*
 * Global file handle pointing to bitstream file.
 */
FILE
  *inFile;

int
  streamStatus; /* either DCM_OK or DCM_EOS */

/*
 * Size of the long term memory (may be
 * modified by command line argument)
 */
int
  longTermMemSize = 1;

/*
 * Defines related to this module.
 */

#define DCM_OK NCC_OK
#define DCM_EOS (DCM_OK + 1)


/*
 * Exported global variables
 */

extern int decLongTermMemSize;       /* Defined in dec_bits.c and set by
                                      * picture header */

/*
 * Local function prototypes used for bitstream manipulations.
 */
static int dcmGetNextFrame(FILE *inFile, bibBuffer_t **inBuffer);
static int dcmSeekPSC(FILE *inFile, int32 *pos);
static int dcmCheckSliceModeAfterPSC(FILE *inFile, int *sliceMode);
static void dcmUsage(char *progname);

/*
 * Main executable decoder function
 */
int main(int argc, char *argv[])
{

  int
    /*
     * General counter or flag variables.
     */
    i, noMoreSlices,
    numFramesInLongTermMem = 0,


    /*
     * Source related variables
     */
    first_frame = 0,
    pic_sizex   = 0,
    pic_sizey   = 0,
    skip        = 0,

    /*
     * Bit related variables
     */
    bits_dct_y, bits_dct_u, bits_dct_v,
    bits_ovr_y, bits_ovr_uv, bits_mode,
    frame_bits = 0, total_bits, emul_prev_bits;

  int32
    header_bits;
  int16
    bibError = 0;


  FILE
    /*
     * File handle for reporting of bits
     * and possibly PSNR (if the original
     * sequence filename is provided in command line).
     */
    *report,
    *orig_y, *orig_u, *orig_v;

  struct frame
    predfr = {NULL, NULL, NULL}, 
    recfr, outfr, origfr, dif_frame,
    **longTermMem,
    *longTermMemStructs;

  char
    *reconame    = NULL,
    *predname    = NULL,
    *origname    = NULL,
    *report_name = "rep_deco.txt",
    *bitstrname  = NULL;

  double
    psnr_y, psnr_u, psnr_v,
    psnr_pred_y, psnr_pred_u, psnr_pred_v;

  clock_t
    cpu_time, cpu_time_last, cpu_time_start, cpu_time_end;

  vdcInstance_t
    *vdcInstance; /* Video Decoder Core instance */

  int dummy, *BitError = &dummy;

  /*
   * Processing command line options.
   */
  for (i = 1; i < argc; i++)
  {
    if (strcmp(argv[i], "-i") == 0)
    {
      if (++i >= argc) {dcmUsage(argv[0]); exit(1);}
      bitstrname = argv[i];
    }

    if (strcmp(argv[i], "-reco") == 0)
    {
      if (++i >= argc) {dcmUsage(argv[0]); exit(1);}
      reconame = argv[i];
    }

    if (strcmp(argv[i], "-pred") == 0)
    {
      if (++i >= argc) {dcmUsage(argv[0]); exit(1);}
      predname = argv[i];
    }

    if (strcmp(argv[i], "-orig") == 0)
    {
      if (++i >= argc) {dcmUsage(argv[0]); exit(1);}
      origname = argv[i];
    }

    if (strcmp(argv[i], "-report") == 0)
    {
      if (++i >= argc) {dcmUsage(argv[0]); exit(1);}
      report_name = argv[i];
    }

    if (strcmp(argv[i], "-h") == 0)
    {
      dcmUsage(argv[0]); exit(0);
    }
  }

  /*
   * Open the bitstream file.
   */
  if (bitstrname == NULL) {
    fprintf(stderr, "Mandatory -i bitstream_file\n argument");
    dcmUsage(argv[0]);
    exit(1);
  }

  if ((inFile = fopen(bitstrname, "rb")) == NULL) {
    fprintf(stderr,"Cannot open bitstream file %s...\n", bitstrname);
    exit(1);
  }

  /*
   * Initialize the bitstream reading internals.
   */
  streamStatus = dcmRequestSliceRefill(&buff_in, &noMoreSlices);

  /*
   * Upon request, open also the statistics report file.
   */
  if(report_name != NULL)
    if((report = fopen(report_name,"w")) == NULL) {
      fprintf(stderr,"Cannot open file %s for writing...\n",report_name);
      exit(1);
    }

  /*
   * Initialize the Video Decoder Core (VDC) module
   */
  if (vdcLoad() < 0) {
    fprintf(stderr, "vdcLoad failed\n");
    exit(1);
  }

  /*
   * Open a dummy VDC instance in order to query the frame size
   * and the starting frame number from the bitstream.
   */
  vdcInstance = vdcOpenCore(0, 0);

  if (!vdcInstance) {
    fprintf(stderr, "vdcOpenCore failed\n");
    exit(1);
  }
  else
  {
    int 
      qpi, fOK, FrNo, IPB = -1,
      dummySliceMode, dummyUseLoopFilter,
      dummyNumActiveRefFrames,
      dummyOrderedRefFrames[LONG_TERM_MEM_MAX_SIZE];

    /*
     * Read the vital params from bitstream file
     */
    if (vdxCheckSSC(&IPB, /* 0- P, 2- I*/
                    &pic_sizex, &pic_sizey,
                    &FrNo, &skip,
                    &qpi, &header_bits,
                    vdcInstance->vdxInstance, (int32 *)BitError))
      dummySliceMode = 1;

    fOK = vdxGetPicHead(&IPB /* 0 for P, 2 for I */, 
                        &FrNo, &skip,
                        &dummySliceMode,
                        &dummyUseLoopFilter,
                        &qpi, &pic_sizex, &pic_sizey,
                        &dummyNumActiveRefFrames,
                        (int32 *)dummyOrderedRefFrames,
                        &header_bits, 1,
                        vdcInstance->vdxInstance, (int32 *)BitError);

    if(*BitError == VDX_SAW_SYNC) {
      fprintf(stderr,"First frame header bitstream error\n");
      nccExit(1);
    }

    /*
     * If first picture type is not I then decoding is aborted.
     */
    if(IPB != 2) {
      fprintf(stderr, "First frame header indicates non-Intra %d\n", IPB);
      exit(1);
    }

    if(!fOK) {
      fprintf(stderr,"First frame header error.\n");
      exit(1);
    }

    first_frame = FrNo;

    /*
     * Close the dummy VDC instance
     */
    if (vdcCloseCore(vdcInstance) < 0) {
      fprintf(stderr, "vdcCloseCore failed\n");
      exit(1);
  }

    /*
     * Rewind to the beginning of the buffer in order to
     * read the header again inside the vdcDecodeFrame function.
     */

    bibRewindBits(bibNumberOfRewBits(buff_in), buff_in, &bibError);
    if (bibError)
      nccExit(1);
  }

  /*
   * Allocate frame memories
   */
  ioiAllocFrame(&predfr, pic_sizex, pic_sizey);
  ioiAllocFrame(&recfr, pic_sizex, pic_sizey);
  ioiAllocFrame(&outfr, pic_sizex, pic_sizey);

  /*
   * Allocate long term reference frame memory
   */
  longTermMem =
    (struct frame**)calloc(decLongTermMemSize,sizeof(struct frame*));

  if(longTermMem == NULL) {
    fprintf(stderr,"Cannot allocate long term memory!\n");
    exit(1);
  }

  /*
   * Frame structs
   */
  longTermMemStructs =
    (struct frame*)calloc(decLongTermMemSize,sizeof(struct frame));

  if(longTermMemStructs == NULL) {
    fprintf(stderr,"Cannot allocate long term memory!\n");
    exit(1);
  }

  for(i = 0;i < decLongTermMemSize;i++)
    longTermMem[i] = &longTermMemStructs[i]; 

  /* Frames */
  for(i = 0;i < decLongTermMemSize;i++)
    ioiAllocFrame(longTermMem[i], pic_sizex, pic_sizey);


  /*
   * If requested, the original sequence is read only for
   * sole purpose of PSNR computation.
   */
  if (origname != NULL)
  {
    ioiAllocFrame(&origfr, pic_sizex, pic_sizey);
    ioiAllocFrame(&dif_frame, pic_sizex, pic_sizey);
    ioiOpenFile(&orig_y, &orig_u, &orig_v, origname);
    ioiSkipFrames(first_frame, pic_sizex, pic_sizey,
                  orig_y, orig_u, orig_v);
    ioiReadFrameYUV(&origfr, pic_sizex, pic_sizey,
                    orig_y, orig_u, orig_v);
  }

  /*
   * Handle the requested output files.
   */
  if( predname != NULL )
    ioiClearSequence(predname);
  if( reconame != NULL )
    ioiClearSequence(reconame);

  /*
   * Open the proper VDC instance in
   * order to query the frame size
   * and the starting frame number
   */
  vdcInstance = vdcOpenCore(pic_sizex, pic_sizey);
  if (!vdcInstance) {
    fprintf(stderr, "vdcOpenCore failed.\n");
    exit(1);
  }
  else
  {
      
    /*
     * Decode the first I picture by calling vdcDecodeFrame().
     */

    cpu_time_last  = cpu_time_start = clock();

    if (vdcDecodeFrame(vdcInstance, longTermMem, &predfr, &recfr,
                       &outfr, BitError) < 0) {
      fprintf(stderr, "Error in decoding the first INTRA.\n");
      exit(1);
    }
    if(*BitError == VDX_SAW_SYNC) {
      fprintf(stderr, "Bit error in first I picture bitstream\n");
      exit(1);
    }

    /*
     * Set the bit counters to the values dervied from forst I picture.
     */
    total_bits  = vdcInstance->total_bits;
    bits_dct_y  = vdcInstance->bits_dct_y;
    bits_dct_u  = vdcInstance->bits_dct_u;
    bits_dct_v  = vdcInstance->bits_dct_v;
    bits_ovr_y  = vdcInstance->bits_ovr_y;
    bits_ovr_uv = vdcInstance->bits_ovr_uv;
    header_bits = vdcInstance->header_bits;
    emul_prev_bits = 0;

    /*
     * Clock the intra frame CPU time.
     */
    cpu_time = clock();


    /*
     * Update the bitstream internals.
     */
    bibError = 0;
    bibDelete(&buff_in, &bibError);
    if (bibError)
      nccExit(-1);

    if (origname != NULL)
    {
      /*
       * Reconstruction error for first I picture
       */
      ioiCalcErrYUV(&recfr, &origfr, &dif_frame, 1, pic_sizex*pic_sizey,
                    &psnr_y, &psnr_u, &psnr_v, NULL, NULL, NULL);
    }

    /*
     * Report statistics
     */
    if(report_name != NULL)
    {
      fprintf(report,"-----------------\n");
      fprintf(report,"I-Frame %d, qp %d\n", vdcInstance->frameNum, vdcInstance->qp);
      fprintf(report,"-----------------\n");
      if (origname != NULL)
      {
        fprintf(report,"I SNR_Y_rec : %2.2f\n", psnr_y);
        fprintf(report,"I SNR_U_rec : %2.2f\n", psnr_u);
        fprintf(report,"I SNR_V_rec : %2.2f\n", psnr_v);
        fprintf(report,"-----------------\n");
      }
      fprintf(report,"I Coeff_Y   : %d\n", bits_dct_y);
      fprintf(report,"I Coeff_C   : %d\n", bits_dct_u + bits_dct_v);
      fprintf(report,"I CBPY      : %d\n", bits_ovr_y);
      fprintf(report,"I MCBPC     : %d\n", bits_ovr_uv);
      fprintf(report,"I header    : %d\n", header_bits + emul_prev_bits);
      fprintf(report,"=================\n");
      fprintf(report,"I Total     : %d\n", total_bits+header_bits+emul_prev_bits);
      fprintf(report,"I Time      : %.2f\n", 
      (float)(cpu_time-cpu_time_last)/CLOCKS_PER_SEC);
      fflush(report);
    }

    /*
     * Dump reconstructed frame to output file.
     */
    if( reconame != NULL )
      ioiPutFrameYUV(&recfr, reconame, vdcInstance->frameNum, 
                     pic_sizex, pic_sizey);
  }


  /*
   * Read bitstream file and decode P frames one by one.
   */
  while (streamStatus == DCM_OK)
  {

    /*
     * Update internals of the bitstream manipulation.
     */
    streamStatus = dcmRequestSliceRefill(&buff_in, &noMoreSlices);
    
    /*
     * Update update reference frame(s).
     */
    ioiRotateLongTermMem(longTermMem, &numFramesInLongTermMem,
                         decLongTermMemSize);
    ioiExchangeFrames(&recfr, longTermMem[0]);

    /*
     * Initialize timing variables.
     */
    cpu_time_last = clock();

    /*
     * Decode a full P frame.
     */
    if (vdcDecodeFrame(vdcInstance, longTermMem, &predfr, &recfr, &outfr,
                       BitError) < 0) {
      fprintf(stderr, "Error in vdcDecodeFrame for P frames\n");
      exit(1);
    }
    if(*BitError == VDX_SAW_SYNC) {
      fprintf(stderr, "Bit Error in the bitstream.\n");
      exit(1);
    }

    /*
     * Update frame bit counters for current P frame just decoded.
     */
    skip        = vdcInstance->skip;
    total_bits  = vdcInstance->total_bits;
    frame_bits  = vdcInstance->frame_bits;
    bits_dct_y  = vdcInstance->bits_dct_y;
    bits_dct_u  = vdcInstance->bits_dct_u;
    bits_dct_v  = vdcInstance->bits_dct_v;
    bits_ovr_y  = vdcInstance->bits_ovr_y;
    bits_ovr_uv = vdcInstance->bits_ovr_uv;
    bits_mode   = vdcInstance->bits_mode;
    header_bits = vdcInstance->header_bits;
    emul_prev_bits = 0;
    
    /*
     * Clock the P frame CPU time.
     */
    cpu_time = clock();
    
    bibError = 0;
    bibDelete(&buff_in, &bibError);
    if (bibError)
      nccExit(-1);

    if (origname != NULL)
    {
      /*
       * If requested, the original sequence is read only for
       * the sole purpose of PSNR computation.
       */
      ioiSkipFrames(skip, pic_sizex, pic_sizey, orig_y, orig_u, orig_v);
      ioiReadFrameYUV(&origfr, pic_sizex, pic_sizey,
                      orig_y, orig_u, orig_v);
      /*
       * MC prediction PSNR.
       */
      ioiCalcErrYUV(&predfr, &origfr, &dif_frame, 1, pic_sizex*pic_sizey,
                    &psnr_pred_y, &psnr_pred_u, &psnr_pred_v,
                    NULL, NULL, NULL);
      /*
       * Reconstruction PSNR.
       */
      ioiCalcErrYUV(&outfr, &origfr, &dif_frame, 1, pic_sizex*pic_sizey,
                    &psnr_y, &psnr_u, &psnr_v, NULL, NULL, NULL);
    }
    
    /*
     * Report statistics
     */
    if (report_name != NULL)
    {
      fprintf(report,"-----------------\n");
      fprintf(report,"P-Frame %d, qp %d\n", vdcInstance->frameNum, vdcInstance->qp);
      fprintf(report,"-----------------\n");
      if (origname != NULL)
      {
        fprintf(report,"SNR_Y_pred: %2.2f\n", psnr_pred_y);
        fprintf(report,"SNR_U_pred: %2.2f\n", psnr_pred_u);
        fprintf(report,"SNR_V_pred: %2.2f\n", psnr_pred_v);
        fprintf(report,"-----------------\n");
        fprintf(report,"SNR_Y_rec : %2.2f\n", psnr_y);
        fprintf(report,"SNR_U_rec : %2.2f\n", psnr_u);
        fprintf(report,"SNR_V_rec : %2.2f\n", psnr_v);
        fprintf(report,"-----------------\n");
      }
      fprintf(report,"Vectors   : %d\n", frame_bits);
      fprintf(report,"SPL_MRG   : %d\n", 0);
      fprintf(report,"Coeff_Y   : %d\n", bits_dct_y);
      fprintf(report,"Coeff_C   : %d\n", bits_dct_u + bits_dct_v);
      fprintf(report,"CODE_NOT  : %d\n", bits_ovr_y + bits_ovr_uv + bits_mode);
      fprintf(report,"P header  : %d\n", header_bits + emul_prev_bits);
      fprintf(report,"=================\n");
      fprintf(report,"Total     : %d\n",total_bits+header_bits+emul_prev_bits);
      fprintf(report,"Time      : %.2f\n", 
      (float)(cpu_time-cpu_time_last)/CLOCKS_PER_SEC);
      fflush(report);
    }

    /*
     * Dump MC prediction frame to output file.
     */
    if(predname != NULL)
      ioiPutFrameYUV(&predfr, predname, vdcInstance->frameNum,
                     pic_sizex, pic_sizey);
    /*
     * Dump reconstructed frame to output file.
     */
    if(reconame != NULL)
      ioiPutFrameYUV(&outfr, reconame, vdcInstance->frameNum,
                     pic_sizex, pic_sizey);
  }

  /*
   * Terminate decoding.
   */
  fclose(inFile);
  vdcCloseCore(vdcInstance);
  vdcFree();
  /*
   * Close statistics file.
   */
  if (report != NULL)
    fclose(report);


  ioiFreeFrame(&predfr);
  ioiFreeFrame(&recfr);
  ioiFreeFrame(&outfr);

  /*
   * Free frames in long term memory
   */
  for(i = 0;i < decLongTermMemSize;i++)
    ioiFreeFrame(longTermMem[i]);
  /*
   * Free frame structs in long term memory
   */
  free(longTermMemStructs);

  /*
   * Free pointers to long term memory frame structs
   */
  free(longTermMem);

  if (origname != NULL)
  {
    ioiFreeFrame(&origfr);
    ioiFreeFrame(&dif_frame);
    fclose(orig_y);
    fclose(orig_u);
    fclose(orig_v);
  }

  /*
   * Clock the overall execution time.
   */
  cpu_time_end = clock();

  fprintf(stdout, "\nTotal decoding time: %.2f\n",
          (float)(cpu_time_end - cpu_time_start)/CLOCKS_PER_SEC);
  exit(0);
}


/*
 * Local functions
 */

/*
 * dcmGetNextFrame
 *    
 * Function:
 *    This function allocates room for the next frame in the file and
 *    gets the frame to the allocated memory block. Then, it creates
 *    a bit buffer and returns the pointer to it in *inBuffer.
 *
 *    The function assumes that the file pointer is in the beginning of
 *    the wanted frame.
 *    The file pointer position is left to the beginning of the next frame.
 *
 *    When the bit buffer is no longer used, it must be destroyed.
 *
 * Returns:
 *    DCM_OK if successful
 *    DCM_EOS if the end of the file was reached
 */

static int dcmGetNextFrame(FILE *inFile, bibBuffer_t **inBuffer)
{
  int
    seekPSCStatus;

  int16
    bibError = 0; /* error code returned from a call to some bib function */

  int32 
    currPSCPos, 
    nextPSCPos,
    frameSize;

  void
    *frameData; /* memory block storing one encoded frame */

  /* Assume that one is in the beginning of a PSC.
     Get that file pointer position. */
  currPSCPos = ftell(inFile);

  /* Skip the current PSC. */
  if (fseek(inFile, 3, SEEK_CUR) != 0)
    nccExit(1);

  /* Seek the next PSC. */
  seekPSCStatus = dcmSeekPSC(inFile, &nextPSCPos);

  /* Set the file pointer to the beginning of the frame. */
  if (fseek(inFile, currPSCPos, SEEK_SET) != 0)
    nccExit(1);

  /* Allocate memory for the frame. */
  frameSize = nextPSCPos - currPSCPos;
  frameData = nccMalloc(frameSize);

  /* Read the frame data to the allocated buffer. */
  if (fread(frameData, 1, frameSize, inFile) < (size_t) frameSize)
    nccExit(1);

  fprintf(stdout, "%ld frame size\n", frameSize);

  /* Create a bit buffer for the frame. */
  *inBuffer = bibCreate(frameData, frameSize, &bibError);
  if (!inBuffer || bibError)
    nccExit(1);

  /* Return either End of sequence or just ok. */

  streamStatus = seekPSCStatus;

  return seekPSCStatus;
}


/*
 * dcmSeekPSC
 *  
 * Function:
 *    This function finds the next PSC starting from the current position
 *    of the given file. The position is left to the beginning of the found
 *    PSC.
 *
 * Returns:
 *    DCM_OK if successful
 *    DCM_EOS if the end of the file was reached
 */

static int dcmSeekPSC(FILE *inFile, int32 *pos)
{
  const u_int32
    psc = 0x00000040,  /* PSC code */
    eos = 0x0000007c;  /* EOS code */

  int seeking = 1, found = 0, i;
  u_int32 lBuffer = 0; /* a 4-byte window to the file */
  u_char ucBuffer;

  /* Prefill the buffer */
  for (i = 0; i < 3 && seeking; i++) {
    seeking = fread(&ucBuffer, sizeof (u_char), 1, inFile);
    lBuffer = (lBuffer << 8) | ucBuffer;
  }

  /* Seek PSC until it is found or the end of the file is reached */
  while (seeking) {
    u_int32 masked22bits = lBuffer & 0x00fffffc;
    if (masked22bits == psc) {
      found = (int) psc;
      break;
    }
    /*else if (masked22bits == eos) {
      found = (int) eos;
      break;
    }*/
    seeking = fread(&ucBuffer, sizeof (u_char), 1, inFile);
    lBuffer = (lBuffer << 8) | ucBuffer;
  }

  if (found)
    /* Rewing to the beginning of PSC or EOS */
    fseek(inFile, -3, SEEK_CUR);
  *pos = ftell(inFile);

  /* If the end-of-file was reached or an EOS was found */
  if (!found || found == (int) eos)
    /* Return indication of the end of the stream. */
    return DCM_EOS;

  /* Return ok */
  return DCM_OK;
}

/*
 * dcmSeekSC
 *  
 * Parameters:
 *    inFile    I: File pointer to bitstream file.
 *    inBuffer  O: Bit buffer containing one encoded slice.
 *
 * Function:
 *    Finds the next SSC or PSC starting from the current
 *    position of the given file. The position is left to the 
 *    beginning of the found SSC or PSC.
 *
 * Returns:
 *    DCM_OK if successful.
 *    DCM_EOS if the end of the sequence was reached.
 */

static int dcmSeekSC(FILE *inFile, int32 *pos, int *lastSlice)
{
  const u_int32 
    ssc = 0x000000c0,  /* SSC code */
    psc = 0x00000040,  /* PSC code */
    eos = 0x0000007c;  /* EOS code */

  int 
    seeking = 1,       /* Flag for seeking of the start code */ 
    found = 0,         /* Flag for found start code */
    i;

  u_int32 
    maskedBits,
    lBuffer = 0;       /* A 4-byte window to the file */

  u_char 
    ucBuffer;          /* A 1-byte window to the file */

  static u_char
    currTR = 0;
  /* 
   * Prefill the buffer. 
   */
  for (i = 0; i < 3 && seeking; i++) 
  {
    seeking = fread(&ucBuffer, sizeof (u_char), 1, inFile);
    lBuffer = (lBuffer << 8) | ucBuffer;
  }

  /* 
   * Seek for SSC, PSC or EOS until it is found or the end 
   * of the file is reached.
   */
  while (seeking) 
  {
    maskedBits = lBuffer & 0x00ffffc0;
    if (maskedBits == ssc) 
    {
      found = (int) ssc;
      break;
    }
    maskedBits = lBuffer & 0x00fffffc;
    if (maskedBits == psc) 
    {
      found = (int) psc;
      break;
    }
    else if (maskedBits == eos) 
    {
      found = (int) eos;
      break;
    }

    seeking = fread(&ucBuffer, sizeof (u_char), 1, inFile);
    lBuffer = (lBuffer << 8) | ucBuffer;
  }

  /*
   * If start code is found, rewind to the beginning of SSC or PSC. 
   */

  *lastSlice = 1;

  if (((u_int32)found) == ssc)
  {
    /* Get the timestamp of slice, SSC TRI TR */

    if (seeking = fread(&ucBuffer, sizeof (u_char), 1, inFile))
    {
      lBuffer = (lBuffer << 8) | ucBuffer;
      maskedBits = lBuffer & 0x0000003fe0;
      lBuffer = (u_int32) currTR;
      lBuffer = (lBuffer << 5) | 0x0000002000;
      if (maskedBits == lBuffer)
        *lastSlice = 0;
      else
        currTR = (u_char) ((maskedBits >> 5) & 0x00000000ff);

      if (! (maskedBits & 0x0000002000)) {
        deb0p("Found slice without a timestamp");
        nccExit(-1);
      }

      fseek(inFile, -1, SEEK_CUR);
    }
  }
  else if (((u_int32)found) == psc)
  {
    /* Get the timestamp of frame, PSC SM TR */

    if (seeking = fread(&ucBuffer, sizeof (u_char), 1, inFile))
    {

      lBuffer = (lBuffer << 8) | ucBuffer;
      currTR = (u_char) ((lBuffer & 0x00000001fe) >> 1);

      fseek(inFile, -1, SEEK_CUR);
    }
  }

  if (found)
    fseek(inFile, -3, SEEK_CUR);
  *pos = ftell(inFile);

  /* 
   * Return indication of the end of the stream if 
   * the end-of-file was reached or an EOS was found.
   */
  if (!found || found == (int)eos)
    return DCM_EOS;

  /* 
   * Otherwise, return OK 
   */
  return DCM_OK;
}


/*
 * dcmCheckSliceModeAfterPSC
 *
 * Parameters:
 *    inFile    I: File pointer to bitstream file.
 *    sliceMode O: 1 if slice mode is enabled, 0 otherwise
 *
 * Function:
 *    This function check the sliceMode bit which is supposed to
 *    follow immediately the PSC.
 *
 * Returns:
 *    DCM_OK if successful,
 *    exits by nccExit() if no PSC is found at the current position
 *
 */

static int dcmCheckSliceModeAfterPSC(FILE *inFile, int *sliceMode)
{
  u_char
    code[3];

  int32
    currSCPos = ftell(inFile);

  if (fread(code, 1, 3, inFile) < (size_t) 3)
    nccExit(1);

  *sliceMode = (code[2] & 0x02) ? 1 : 0;

  /* Mask to get only the ending of PSC */

  code[2] &= 0xfc;

  if (code[0] == 0 && code[1] == 0 && code[2] == 64)
  {
    /*
     * Set the file pointer to the original position
     */
    if (fseek(inFile, currSCPos, SEEK_SET) != 0)
      nccExit(1);

    return DCM_OK;
  }

  deb0p("dcmCheckSliceModeAfterPSC did not find a PSC");
  nccExit(1); 
}


/*
 * dcmRequestSliceRefill
 *
 * Parameters:
 *    inFile    I: File pointer to bitstream file.
 *    inBuffer  O: Bit buffer containing one encoded slice.
 *    
 * Function:
 *    Allocates room for the next slice in the file and gets 
 *    the slice to the allocated memory block. Then, it creates
 *    a bit buffer and returns the pointer to it in *inBuffer.
 *
 *    The function assumes that the file pointer is in the beginning of
 *    the wanted slice.
 *    The file pointer position is left to the beginning of the next slice.
 *
 *    When the bit buffer is no longer used, it must be destroyed.
 *
 * Returns:
 *    DCM_OK if successful.
 *    DCM_EOS if the end of the sequence was reached.
 * 
 */
int dcmRequestSliceRefill(bibBuffer_t **inBuffer, int *noMoreSlices)
{
  static int
    lastSlice = 0;


  int
    seekSCStatus; /* Status flag for the succes of PSC search */ 

  int16
    bibError = 0; /* Error code */

  int32 
    currSCPos,    /* Position of the current start code */
    nextSCPos,    /* Position of the next start code */
    sliceSize;    /* Sice of the slice data segment */

  void
    *sliceData;   /* Memory block storing one encoded slice */

  /*
   * Free the buffer before requesting new slice
   */

  bibDelete(inBuffer, &bibError);


  if (lastSlice)	  
  {
    *noMoreSlices = 1;
    lastSlice = 0;
    return DCM_OK;
  }
  else
    *noMoreSlices = 0;

  /* 
   * Assume that one is in the beginning of a SSC or PSC
   * and get the corresponding file pointer position. 
   */
  currSCPos = ftell(inFile);

  /* 
   * Skip the current SSC or PSC. 
   */
  if (fseek(inFile, 3, SEEK_CUR) != 0)
    nccExit(1);

  /* 
   * Seek the next SSC or PSC. 
   */
  seekSCStatus = dcmSeekSC(inFile, &nextSCPos, &lastSlice);

  /* 
   * Set the file pointer to the beginning of the slice. 
   */
  if (fseek(inFile, currSCPos, SEEK_SET) != 0)
    nccExit(1);

  /* 
   * Allocate memory for the slice and read the slice data to 
   * the allocated buffer. 
   */
  sliceSize = nextSCPos - currSCPos;
  sliceData = nccMalloc(sliceSize);
  if (fread(sliceData, 1, sliceSize, inFile) < (size_t) sliceSize)
    nccExit(1);

  /*
   *  Create a bit buffer for the slice. 
   */
  *inBuffer = bibCreate(sliceData, sliceSize, &bibError);
  if (!inBuffer || bibError)
    nccExit(1);

  /* 
   * Return either End Of Sequence or just OK. 
   */
  streamStatus = seekSCStatus;

  return seekSCStatus;
}


/*
 * dcmUsage
 *  
 * Function:
 *    Prints the usage information.
 *
 * Returns:
 *    Nothing
 */
static void dcmUsage(char *progname)
{
  fprintf(stderr,"%s\n", progname);
  fprintf(stderr," -i bitstream_filename           (mandatory)\n");
  fprintf(stderr," [-reco reconstruction_filename] defaults to none\n");
  fprintf(stderr," [-pred prediction_filename]     defaults to none\n");
  fprintf(stderr," [-report filename]              defaults to none\n");
  fprintf(stderr," [-orig originals_filename]      defaults to none\n");
}