#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <time.h>

#include "mainFlags.h"
#include "encMux.h"
#include "mvc.h"
#include "mcp.h"
#include "deblockFilt.h"
#include "IOimg.h"

/*
 * Size of the long term memory (may be
 * modified by command line argument)
 */
int
  longTermMemSize = 1; 

/*
 * Cumulative statistics reported
 * at the end of coding the sequence.
 */
static int
  cumulNumOfFrames = 0,
  cumulIntraBits   = 0,
  cumulInterBits   = 0;

static Psnr
  cumulIntraPsnr   = {0.0, 0.0, 0.0},
  cumulInterPsnr   = {0.0, 0.0, 0.0};

/*
 * qpBlockLuma is used to pass luminance qp to
 * vdxGetDctTable() and vdxGetKltTable()
 */
extern int
  qpBlockLuma;


/*
 * Encoder usage can be obtained
 * from command line by using -h or
 * no arguments.
 */

void usage(void)
{
  fprintf(stdout,
  "Encoder arguments:\n"
  "  [-h]       to get this help screen.\n"
  "   -in       input_seq_name               (mandatory)\n"
  "  [-size pic_sizex pic_sizey]             [default 176 144]\n"
  "  [-start]   first frame number           [default 0]\n"
  "  [-last]    last frame number            [default first]\n"
  "  [-skip]    no of skip frames            [default 2]\n"
  "  [-qintra value] quant param for I frame [default 16]\n"
  "  [-q value] quant param for P frames     [default same as I]\n"
  "  [-compl]   Complexity mode              [default 0]\n"
  "                              0 - Anchor   \n"
  "                              1 - Preferred\n"
  "                              2 - Low      \n"
  "  [-report file] per frame statistics     [default rep_enco.txt]\n"
  "  [-cumul file]  per total statistics     [default none]\n"
  "  [-reco reconstruction_filename]         [default none]\n"
  "  [-pred MC_prediction_filename]          [default none]\n"

  "  [-slice]   enables slice mode in bitstream\n"
  "  [-loopf]   use filter in loop\n"
  "  [-ssz]     slice size in bits           [default 1 line of MBs/slice]\n"
  "  [-dupl]    to write repeated reconstructed frames to disk\n"
  "             This is useful if output seqs must be\n"
  "             of same length as the input sequence.\n"
  "  [-rframes] Number of reference frames   [default 1]\n"
  "             in long term memory\n"
  );
}


/*
 * Local auxiliary functions used in the main encoder function
 * for dumping frames to files or for dumping statistics.
 */

static void DumpPictures(char *recoFilename, struct frame *curr_recon,
                         char *predFilename, struct frame *curr_pred,
                         int pic_sizex, int pic_sizey, int l,
                         int pict_coding_type, int write_repeatedly);

static void CalculateStatistics(struct frame *curr_image,
                                struct frame *curr_recon,
                                struct frame *curr_pred,
                                struct frame *dif_frame,
                                int pic_sizex, int pic_sizey,
                                Bits *bits, int l, int qp,
                                clock_t *cpu_time,
                                int pict_coding_type,
                                FILE *report);

/*
 * Encoder main function
 */

int main(int argc, char *argv[])
{

  Bits
    bits;

  int
    /*
     * Default compexity mode is 0 (highest).
     */
    complexityMode = COMPLEXITY_MODE_ANCHOR,

    /* This flag enables loopfilter */

    useLoopFilter = 0,

    /* Source format related variables */

    chosen_frameskip = 2,
    first_frame = 0, last_frame = -1,
    pic_sizex = 176, pic_sizey = 144,
    write_repeatedly = 0,

    numFramesInLongTermMem = 0,

    qpI = 16,
    qp = -1;

  double
    /*
     * Lambda is the lagrangian parameter used
     * in P-frames for rate distortion
     * optimizations. The formula to compute
     * lambda parameter is:
     *
     *                   2
     * lambda = 0.85 * qp
     *
     */
    lambda;

  /*
   * Input and output related filenames.
   */
  char
    str[2000],
    *seqname      = NULL,
    *predname     = NULL,
    *reconame     = NULL,
    *report_name  = "rep_enco.txt",
    *cumulRepName = NULL;

  /*
   * Input and output related file handles.
   */
  FILE
    *fy, *fu, *fv,
    *report;

  struct frame
    curfr, recfr, filtfr,
    predfr, dif_frame, prev_orig,
    **longTermMem, *longTermMemStructs;

  /*
   * Standard ANSI C support for measuring
   * the CPU time for every encoded frame.
   */
  clock_t
    cpu_time = clock();

  int
    /* General variables used during the coding loop */

    i, l,
    sliceMode = 0, sliceBits = DEFAULT_SLICE_BITS,
    forceInext,
    Pframeskip,
    picCodingType,
    prevPicCodingType,
    numCodedFrames;

  /* The lst holding the data about coded blocks & macroblocks */

  object
    *object_list;

  /*
   * The slice buffer holds the map for slice labels. This is
   * needed in the prediction of intra or in the prediction
   * of affine parameters. In the case that sliceMode is 0 (no slices)
   * this map is set to a constant label for the entire frame.
   */
  int16
    *slicePredMap;

  /*
   * Parsing the command line arguments
   * one by one.
   */

  if (argc < 3)
  {
    usage();
    exit(0);
  }

  for (i = 1; i < argc; i++)
  {
    if (strcmp(argv[i], "-in") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      seqname = argv[i];
    }

    if (strcmp(argv[i], "-size") == 0) {
      if (++i >= argc) {usage();exit(0);}
      pic_sizex = atoi(argv[i]);
      if (++i >= argc) {usage();exit(0);}
      pic_sizey = atoi(argv[i]);
      continue;
    }

    if (strcmp(argv[i], "-skip") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      chosen_frameskip = atoi(argv[i]);
    }

    if (strcmp(argv[i], "-loopf") == 0) {
      useLoopFilter = 1;
    }

    if (strcmp(argv[i], "-start") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      first_frame = atoi(argv[i]);
    }

    if (strcmp(argv[i], "-last") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      last_frame = atoi(argv[i]);
    }

    if (strcmp(argv[i], "-qintra") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      qpI = atoi(argv[i]);
    }

    if (strcmp(argv[i], "-q") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      qp = atoi(argv[i]);
    }

    if (strcmp(argv[i], "-report") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      report_name = argv[i];
    }

    if (strcmp(argv[i], "-pred") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      predname = argv[i];
    }

    if (strcmp(argv[i], "-reco") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      reconame = argv[i];
    }

    if (strcmp(argv[i], "-dupl") == 0)
     write_repeatedly = 1;

    if (strcmp(argv[i], "-slice") == 0)
     sliceMode = 1;

    if (strcmp(argv[i], "-ssz") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      sliceBits = atoi(argv[i]);
      sliceMode = 1;
    }

    if (strcmp(argv[i], "-compl") == 0) {
      if (++i >= argc) {usage(); exit(0);}

      switch(atoi(argv[i])){
      case 0: complexityMode = COMPLEXITY_MODE_ANCHOR;    break;
      case 1: complexityMode = COMPLEXITY_MODE_PRACTICAL; break;
      case 2: complexityMode = COMPLEXITY_MODE_LOW;       break;
      default: usage(); exit(0); break;
      }
    }

    if (strcmp(argv[i], "-cumul") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      cumulRepName = argv[i];
    }

    if (strcmp(argv[i], "-rframes") == 0) {
      if (++i >= argc) {usage(); exit(0);}
      longTermMemSize = atoi(argv[i]);

      if (longTermMemSize > LONG_TERM_MEM_MAX_SIZE){
        fprintf(stdout, "Max size for long term memory is %d frames\n",
                LONG_TERM_MEM_MAX_SIZE);
        exit(1);
      }

      if (longTermMemSize > 1)
        fprintf(stdout, "Using long term memory (%d reference frames)\n",
                longTermMemSize);
    }

    if (strcmp(argv[i], "-h") == 0) {
      usage(); exit(0);
    }
    
  }

  /* If multiple reference pictures are enabled, force the
   * complexity mode to 0. (Other modes currently not
   * supported).
   */

   if (longTermMemSize > 1 && complexityMode != COMPLEXITY_MODE_ANCHOR){
     fprintf(stdout, "Using long term memory -> Switching to anchor complexity mode\n");
     complexityMode = COMPLEXITY_MODE_ANCHOR;
   }

  /*
   * If QP for P frames has not been
   * specified then it takes the same
   * value as the QP for I-frames.
   */
  if (qp == -1)
    qp = qpI;

  /*
   * Lambda is the lagrangian parameter used
   * in P-frames for rate distortion
   * optimizations. The formula to compute
   * lambda parameter is:
   *
   *                   2
   * lambda = 0.85 * qp
   *
   */
  lambda = 0.85 * (double) (qp * qp);

  /*
   * If unspecified last frame
   * only encode the first intra frame.
   */
  if(last_frame == -1)
    last_frame = first_frame;

  /*
   * Check validity of params.
   */
  if(first_frame < 0 || first_frame > last_frame)
  {
    fprintf(stdout, "Error: Illegal frame number settings!\n");
    usage();
    exit(1);
  }

  if(seqname == NULL)
  {
    fprintf(stdout,"Error: Input filename is mandatory!\n");
    usage();
    exit(1);
  }

  /*
   * Echo picture size.
   */
  fprintf(stdout,"Image size: %d x %d\n", pic_sizex, pic_sizey);

  /*
   * Initialize the file handles needed
   * to read the Y, u, v input sequence.
   */
  ioiOpenFile(&fy, &fu, &fv, seqname);

  /*
   * Initialize the file handles needed
   * to write per frame statistics report
   * and bitstream file.
   */
  if((report = fopen(report_name,"w")) == NULL) {
    fprintf(stdout,"Cannot open report file %s for writing!\n",
            report_name);
    usage();
    exit(1);
  }

  sprintf(str, "%s.bitstr", report_name);
  if((bitstream_bit = fopen(str,"wb")) == NULL) {
    fprintf(stdout, "Cannot open binary bitstream file %s\n", str);
    exit(1);
  }


  /* Allocate frames memories */

  ioiAllocFrame(&curfr,     pic_sizex, pic_sizey);
  ioiAllocFrame(&predfr,    pic_sizex, pic_sizey);
  ioiAllocFrame(&recfr,     pic_sizex, pic_sizey);
  ioiAllocFrame(&filtfr,    pic_sizex, pic_sizey);
  ioiAllocFrame(&dif_frame, pic_sizex, pic_sizey);
  ioiAllocFrame(&prev_orig, pic_sizex, pic_sizey);


  /*** Long term reference frame memory ***/

  /* Pointers for frame structs */
  longTermMem = (struct frame **) calloc(longTermMemSize,
                                         sizeof(struct frame *));

  if(longTermMem == NULL) {
    fprintf(stdout,"Cannot allocate long term memory!\n");
    exit(1);
  }

  /* Frame structs */
  longTermMemStructs =
    (struct frame*)calloc(longTermMemSize,sizeof(struct frame));

  if(longTermMemStructs == NULL){
    fprintf(stdout,"Cannot allocate long term memory!\n");exit(1);
  }
  
  for(i = 0;i < longTermMemSize;i++)
    longTermMem[i] = &longTermMemStructs[i]; 

  /* Frames */
  for(i = 0; i < longTermMemSize; i++)
    ioiAllocFrame(longTermMem[i], pic_sizex, pic_sizey);

  /*
   * The slice buffer holds the map for slice labels. This is
   * needed in the prediction of intra or in the prediction
   * of affine parameters. In the case that sliceMode is 0 (no slices)
   * this map is set to a constant label for the entire frame.
   */
  if ((slicePredMap = (int16 *) calloc(pic_sizex * pic_sizey >> 6,
     sizeof(int))) == NULL) 
  {
    fprintf(stdout,"Cannot allocate memory for slicePredMap\n");
    exit(1);
  }

  /*
   * Each macroblock has a corresponding
   * set of 5 "block" allocated always.
   * The first structure holds data about
   * the macroblock itself (including information
   * whether it is splitted or not. The
   * following 4 structures hold data about
   * each block within the macroblock.
   */
  if((object_list = (object*) calloc(pic_sizex * pic_sizey / 256 * 5,
                                     sizeof(object))) == NULL) {
    fprintf(stdout,"Cannot allocate region list! \n");
    exit(1);
  }

  /*** Initialize various buffers used in speed optimization  ***/

  initTriplet2Index(); /* for SendDctOrKlt_EventXxX() */
  splSetPicSize(pic_sizex, pic_sizey);
  splInitInterpArray();
  splInitDerivativeArray();
  mvcInitKernel();
  filInitOverallActTable();
  filInitDiffActTable();

  /*
   * Initialize the amount of coded frames.
   */
  numCodedFrames    = 0;

  /*
   * First picture is always I.
   * Previous picture is initialized as undefined.
   */
  forceInext        = 1;
  prevPicCodingType = PCT_NONE;

  /*
   * Initialize the first frame, counter loop variable,
   * and Pframeskip variable.
   */
  ioiSkipFrames(first_frame, pic_sizex, pic_sizey, fy, fu, fv);
  l = first_frame;
  Pframeskip = -1;


  /***************************************************
   *             Main coding loop starts.            *
   ***************************************************/

  for ( ; ; )
  {

    picCodingType = (forceInext) ? PCT_INTRA : PCT_INTER;

    /*
     * Pframeskip is the distance between current and
     * previous frames in terms of timestamps.
     */
    Pframeskip = (Pframeskip == -1) ? 0 : (chosen_frameskip + 1);

    l += Pframeskip;

    if (l > last_frame)
      break;

    /*********************************************
     * Reading of the current frame into &curfr  *
     *********************************************/

    if (Pframeskip > 0)
      ioiSkipFrames(Pframeskip-1, pic_sizex, pic_sizey, fy, fu, fv);
    ioiReadFrameYUV(&curfr, pic_sizex, pic_sizey, fy, fu, fv);

    /*
     * According to picture type select the
     * I or P frame coding units.
     */
    switch (picCodingType)
    {
      case PCT_INTRA:

        Code_I_Frame(&curfr, &recfr, pic_sizex, pic_sizey,
                      l, (qpBlockLuma = qpI), &bits, slicePredMap,
                      sliceMode, sliceBits, complexityMode);

        CalculateStatistics(&curfr, &recfr, NULL, &dif_frame,
                            pic_sizex, pic_sizey, &bits, l,
                            qpI, &cpu_time, PCT_INTRA, report);

        DumpPictures(reconame, &recfr, NULL, NULL,
                     pic_sizex, pic_sizey, l, PCT_INTRA, write_repeatedly);

        ioiRotateLongTermMem(longTermMem, &numFramesInLongTermMem,
                             longTermMemSize);
        ioiExchangeFrames(&recfr, longTermMem[0]);
        ioiExchangeFrames(&curfr, &prev_orig);

        numCodedFrames++;
        /*
         * After an I picture the next picture type is set to P.
         */
        forceInext        = 0;
        prevPicCodingType = PCT_INTRA;

        break;



      case PCT_INTER:

        Code_P_Frame(longTermMem, &curfr, &prev_orig, &predfr, &recfr,
                     &filtfr, pic_sizex, pic_sizey, l,
                     object_list, (qpBlockLuma = qp), lambda, &bits,
                     numFramesInLongTermMem, 
                     slicePredMap, sliceMode, sliceBits,
                     complexityMode, useLoopFilter);

        CalculateStatistics(&curfr, &filtfr, &predfr, &dif_frame,
                            pic_sizex, pic_sizey, &bits, l,
                            qp, &cpu_time, PCT_INTER, report);

        DumpPictures(reconame, &filtfr, predname, &predfr,
                     pic_sizex, pic_sizey, l, PCT_INTER, write_repeatedly);

        ioiRotateLongTermMem(longTermMem, &numFramesInLongTermMem,
                             longTermMemSize);
        ioiExchangeFrames(&recfr, longTermMem[0]);
        ioiExchangeFrames(&curfr, &prev_orig);

        numCodedFrames++;
        prevPicCodingType = PCT_INTER;

    	break;

    }  /* end from switch (picCodingType) */

  } /* end from infinite loop */

  /*
   * Close bitstream file.
   */
  SendEndofSeq();
  fclose(bitstream_bit);

  /*
   * Dump last  frame if
   * repeated writing is enabled.
   */
  if (write_repeatedly)
    DumpPictures(reconame, NULL, predname, NULL, pic_sizex, pic_sizey,
                 last_frame+1, PCT_END, write_repeatedly);

  fprintf(stdout, "Total CPU time: %.2f for %d frames\n",
          (float) clock() / CLOCKS_PER_SEC, numCodedFrames);
  fprintf(stdout, "Frames/s: %.2f \n",
          (float)numCodedFrames / ((float) clock() / CLOCKS_PER_SEC) );

  ioiFreeFrame(&curfr);
  ioiFreeFrame(&predfr);
  ioiFreeFrame(&recfr);
  ioiFreeFrame(&dif_frame);
  ioiFreeFrame(&prev_orig);
  ioiFreeFrame(&filtfr);

  /* Frames in long term memory */
  for(i = 0;i < longTermMemSize;i++)
    ioiFreeFrame(longTermMem[i]);

  /* Frame structs in long term memory */
  free(longTermMemStructs);

  /* Pointers to long term memory frame structs */
  free(longTermMem);

  free((void *) slicePredMap);
  free((void *) object_list);

  /*
   * Close file(s) for statistics.
   * 
   */
  fclose(report);

  if(cumulRepName != NULL)
  {
    FILE *tmpFile;

    tmpFile = fopen(cumulRepName, "a");
    if (tmpFile) {

      if(cumulNumOfFrames > 1)
        fprintf(tmpFile, "%3d %2d %2d %2.2f %2.2f %2.2f %5d "
                                     "%2.2f %2.2f %2.2f %5d "
                                     "%2.2f %2.2f %2.2f %5d \t%.3f\n",
          cumulNumOfFrames, qpI, qp,
          cumulIntraPsnr.psnr_y,
          cumulIntraPsnr.psnr_u,
          cumulIntraPsnr.psnr_v,
          cumulIntraBits,
          cumulInterPsnr.psnr_y / (cumulNumOfFrames-1),
          cumulInterPsnr.psnr_u / (cumulNumOfFrames-1),
          cumulInterPsnr.psnr_v / (cumulNumOfFrames-1),
          cumulInterBits        / (cumulNumOfFrames-1),
          (cumulIntraPsnr.psnr_y+cumulInterPsnr.psnr_y) / cumulNumOfFrames,
          (cumulIntraPsnr.psnr_u+cumulInterPsnr.psnr_u) / cumulNumOfFrames,
          (cumulIntraPsnr.psnr_v+cumulInterPsnr.psnr_v) / cumulNumOfFrames,
          (cumulIntraBits       +cumulInterBits       ) / cumulNumOfFrames,
          ((float) clock() / CLOCKS_PER_SEC) / cumulNumOfFrames
          );
      else
        fprintf(tmpFile, "%3d %2d %2d %2.2f %2.2f %2.2f %5d "
                                     "%2.2f %2.2f %2.2f %5d "
                                     "%2.2f %2.2f %2.2f %5d %.3f\n",
          cumulNumOfFrames, qpI, qp,
          cumulIntraPsnr.psnr_y,
          cumulIntraPsnr.psnr_u,
          cumulIntraPsnr.psnr_v,
          cumulIntraBits,
          0.0, 0.0, 0.0, 0,
          cumulIntraPsnr.psnr_y,
          cumulIntraPsnr.psnr_u,
          cumulIntraPsnr.psnr_v,
          cumulIntraBits,
          ((float) clock() / CLOCKS_PER_SEC) / cumulNumOfFrames);

      fclose(tmpFile);
    }
  }

  exit(0);
}



/*
 * Auxiliary function used by main encoder function
 * to dump calculate and dump statistics about coded sequence.
 */
static void CalculateStatistics(struct frame *curr_image,
                                struct frame *curr_recon,
                                struct frame *curr_pred,
                                struct frame *dif_frame,
                                int pic_sizex, int pic_sizey,
                                Bits *bits, int l, int qp,
                                clock_t *cpu_time,
                               int pict_coding_type, FILE *report)
{

  double
    psnr_pred_y,
    psnr_pred_u,
    psnr_pred_v,
    psnr_y,
    psnr_u,
    psnr_v;

  clock_t
    cpu_time_last;

  float
    run_time;

  static int
    firstIntra = 1;

  cumulNumOfFrames++;

  ioiCalcErrYUV(curr_recon, curr_image, dif_frame, 1,
                pic_sizex*pic_sizey,
                &psnr_y, &psnr_u, &psnr_v,
                NULL, NULL, NULL);

  cpu_time_last = *cpu_time;
  *cpu_time     = clock();
  run_time      = (float)(*cpu_time-cpu_time_last)/CLOCKS_PER_SEC;

  switch (pict_coding_type)
  {
    case PCT_INTRA:

      fprintf(report,"-----------------\n"
                     "I-Frame %d, qp %d\n"
                     "-----------------\n"
                     "I SNR_Y_rec : %2.2f\n"
                     "I SNR_U_rec : %2.2f\n"
                     "I SNR_V_rec : %2.2f\n"
                     "-----------------\n"
                     "I Coeff_Y   : %d\n"
                     "I Coeff_C   : %d\n"
                     "I CBPY      : %d\n"
                     "I MCBPC     : %d\n"
                     "I header    : %d\n"
                     "=================\n"
                     "I Total     : %d\n"
                     "I Time      : %.1f\n",
                     l, qp, psnr_y, psnr_u, psnr_v,
                     bits->coefy, bits->coefu + bits->coefv,
                     bits->cncy + bits->mode, bits->cncuv,
                     bits->header + get_padding_bits() + get_emulprevent_bits(),
                     bits->total  + get_padding_bits() + get_emulprevent_bits(),
                     run_time);

      if(firstIntra)
      {
        firstIntra            = 0;
        cumulIntraBits        = bits->total +
                                get_padding_bits() +
                                get_emulprevent_bits();
        cumulIntraPsnr.psnr_y = psnr_y;
        cumulIntraPsnr.psnr_u = psnr_u;
        cumulIntraPsnr.psnr_v = psnr_v;
      }
      else{
        cumulInterBits        += bits->total +
                                 get_padding_bits() +
                                 get_emulprevent_bits();
        cumulInterPsnr.psnr_y += psnr_y;
        cumulInterPsnr.psnr_u += psnr_u;
        cumulInterPsnr.psnr_v += psnr_v;
      }

      break;

    case PCT_INTER:

      ioiCalcErrYUV(curr_pred, curr_image, dif_frame, 1,
                    pic_sizex*pic_sizey,
                    &psnr_pred_y, &psnr_pred_u, &psnr_pred_v,
                    NULL, NULL, NULL);

      fprintf(report,"-----------------\n"
                     "P-Frame %d, qp %d\n"
                     "-----------------\n"
                     "SNR_Y_pred: %2.2f\n"
                     "SNR_U_pred: %2.2f\n"
                     "SNR_V_pred: %2.2f\n"
                     "-----------------\n"
                     "-----------------\n"
                     "SNR_Y_rec : %2.2f\n"
                     "SNR_U_rec : %2.2f\n"
                     "SNR_V_rec : %2.2f\n"
                     "-----------------\n"
                     "Vectors   : %d\n"
                     "SPL_MRG   : %d\n"
                     "Coeff_Y   : %d\n"
                     "Coeff_C   : %d\n"
                     "CODE_NOT  : %d\n"
                     "P header  : %d\n"
                     "=================\n"
                     "Total     : %d\n"
                     "Time      : %.1f\n"
                     ,l, qp, psnr_pred_y, psnr_pred_u, psnr_pred_v,
                     psnr_y, psnr_u, psnr_v,
                     bits->motcoef, 0,
                     bits->coefy, bits->coefu + bits->coefv,
                     bits->cncy + bits->cncuv + bits->mode,
                     bits->header + get_padding_bits() + get_emulprevent_bits(),
                     bits->total  + get_padding_bits() + get_emulprevent_bits(),
                     run_time
                     );


      cumulInterBits        += bits->total +
                               get_padding_bits() +
                               get_emulprevent_bits();
      cumulInterPsnr.psnr_y += psnr_y;
      cumulInterPsnr.psnr_u += psnr_u;
      cumulInterPsnr.psnr_v += psnr_v;

      break;

  }

  fflush(report);

}


/*
 * Dumps the motion compensated prediction and/or the recostruction
 * frames to the corresponding files.
 *
 */
static void DumpPictures(char *recoFilename, struct frame *curr_recon,
                         char *predFilename, struct frame *curr_pred,
                         int pic_sizex, int pic_sizey, int l,
                         int pict_coding_type, int write_repeatedly)
{

  static struct frame
    recon_buffer,
    pred_buffer;

  static int
    l_P, pic_sizex_P, pic_sizey_P,
    first_invocation = 1;

  static struct frame
    buf_reco_previous,
    buf_pred_previous;

  static int
    l_previous = -1;

  int
    i, duplications;

  if (first_invocation)
  {
    if (recoFilename != NULL) ioiClearSequence(recoFilename);
    if (predFilename != NULL) ioiClearSequence(predFilename);

    if (write_repeatedly)
    {
      ioiAllocFrame(&buf_reco_previous, pic_sizex, pic_sizey);
      ioiAllocFrame(&buf_pred_previous, pic_sizex, pic_sizey);
    }

    first_invocation = 0;
  }


  switch (pict_coding_type)
  {
    case PCT_INTRA:
    case PCT_INTER:

      if (write_repeatedly && l_previous >= 0)
      {
        duplications = l - l_previous - 1;

        if (recoFilename != NULL && curr_recon != NULL)
          for (i=0; i<duplications; i++)
            ioiPutFrameYUV(&buf_reco_previous, recoFilename,
                           l_previous, pic_sizex, pic_sizey);

        if (predFilename != NULL && curr_pred != NULL)
          for (i=0; i<duplications; i++)
            ioiPutFrameYUV(&buf_pred_previous, predFilename,
                           l_previous, pic_sizex, pic_sizey);
      }

      if (recoFilename != NULL && curr_recon != NULL)
        ioiPutFrameYUV(curr_recon, recoFilename, l, pic_sizex, pic_sizey);

      if (predFilename != NULL && curr_pred != NULL)
        ioiPutFrameYUV(curr_pred, predFilename, l, pic_sizex, pic_sizey);

      if (write_repeatedly)
      {
        if (curr_recon != NULL)
          ioiCopyFrame(curr_recon, &buf_reco_previous, pic_sizex, pic_sizey);
        if (curr_pred != NULL)
          ioiCopyFrame(curr_pred,  &buf_pred_previous, pic_sizex, pic_sizey);
        l_previous = l;
      }

      break;

    case PCT_END:

      if (write_repeatedly && l_previous >= 0)
      {
        duplications = l - l_previous - 1;

        if (recoFilename != NULL)
          for (i=0; i<duplications; i++)
            ioiPutFrameYUV(&buf_reco_previous, recoFilename,
                           l_previous, pic_sizex, pic_sizey);

        if (predFilename != NULL)
          for (i=0; i<duplications; i++)
            ioiPutFrameYUV(&buf_pred_previous, predFilename,
                           l_previous, pic_sizex, pic_sizey);

        l_previous = l;
      }

      break;

  }
}



/* Bogus linker on some C compiler require to have this defined */

int dcmRequestSliceRefill(bibBuffer_t **inBuffer, int *noMoreSlices)
{
  return 0;
}