#include <stdlib.h>
#include <stdio.h>
#include <limits.h>

#include "mainFlags.h"
#include "decDemux.h"
#include "debug.h"
#include "IntraPredDecoder.h"
#include "deblockFilt.h"
#include "PredErrorDecoder.h"
#include "nrctypes.h"
#include "toolMCC.h"
#include "decMCC.h"

extern u_char 
  annexTlookup[32];

/*
 * Function implementations
 */


/***************************************************************************
 * Function    : decodeIntraBlock()
 *
 * Purpose     : Decodes the block layer from bitstream for an Intra
 *               block (either luma or chroma). Intra prediction is used
 *               followed by MCC prediction error coding.
 *
 * Arguments   : See the function call
 *
 * Returns     : void
 *
 *               Copyright (c) 1997 Nokia Research Center
 ***************************************************************************/

static void decodeIntraBlock(Byte *recoFr,         /* Out: Reconstruction frame buf  */
                             int16 *dcFrameBuf,    /* Out: Frame buf for DC pred     */
                             u_char *classifBuf,   /* Out: Frame buf for dir classif */
                             int32 *acBits,        /* Out: Bits for intra pred error */
                             int32 *ovhBits,       /* Out: Bits for intra pred info  */
                             int32 mode,           /* In:  Flag to signal I / P frame*/
                             int32 chroma,         /* In:  Flag to signal luma/chroma*/
                             int32 qp,             /* In:  Quant parameter to use    */
                             int32 maxDClevels,    /* In:  Max # of DC quant levels  */
                             int32 blkSizex,       /* In:  Block height              */
                             int32 blkSizey,       /* In:  Block width               */
                             int32 frX,            /* In:  Block x coord inside frame*/
                             int32 frY,            /* In:  Block y coord inside frame*/
                             int32 frStepx,        /* In:  8 for luma, 4 for chroma  */
                             int32 frStepy,        /* In:  8 for luma, 4 for chroma  */
                             int32 frSizex,        /* In:  Frame width # of pels     */
                             int32 frSizey,        /* In:  Frame height # of pels    */
                             int32 xBlocks,        /* In:  Frame width # of blocks   */
                             int32 sliceXmin,      /* In:  Xmin  of current slice    */
                             int32 sliceXmax,      /* In:  Xmax  of current slice    */
                             int32 sliceYmin,      /* In:  Ymin  of current slice    */
                             int32 sliceYmax,      /* In:  Ymax  of current slice    */
                             int16 *slicePredMap,  /* In:  Slice label buffer        */
                             int32 sliceSx,        /* In:  Horizontal slice buf size */
                             int32 sliceSy,        /* In:  Vertical slice buf size   */
                             int32 currSliceID,    /* In:  Label of current slice    */
                             vdxInstance_t
                               *instance,          /* In:  Current instance          */
                             int *BitError)        /* Out: For error handling        */
{
  /*
   * Various constants used inside this functions
   */
  #define IP_METH_DC                  1
  #define IP_METH_BM                 13

  int32
    maxValue = 255, maxLevel = 256, rotateFlag,
    methods = AMOUNT_OF_8x8_METH+1,
    leftEdge, upperEdge, leftDc, upperDc, upleftDc,
    blkSize, frBlkNo, ipContextClassNo, tmpLeft, tmpUpper, classTmp,
    methName, dcdIndex, bestCom = 0, bestMot = 0, cncFlag, methRank,
    class8x8 = 0, dcPredSum, dcDiff, calSta = 0,
    codMeth8x8, pels, bestAc = 0, decBits, dcRecoSum,
    addy, addx, addNo, tmpPredNo, tmpInd, i,
    predBlk[64], diffBlk[64], recoBlk[64],
    methCoeffs[64], dctCoeffs[64], ExtBlk[16], tmpBlk[64],
    dct4x4Meth, ibmvY, ibmvX, save4[4],
    maxNoOfValidVectors, totalNoOfVectors, reducedSearchRange,
    validX[AMOUNT_OF_INTRA_MOTION],
    validY[AMOUNT_OF_INTRA_MOTION],
    validI[AMOUNT_OF_INTRA_MOTION];

  u_char
    predDirIn;

  blkSize  = blkSizex*blkSizey;
  frBlkNo  = (frY / frStepy) * xBlocks + (frX / frStepx);  /* integer division */

  /* 
   * Depending on whether the block at the
   * left and the block at the right are inside
   * the same slice as the current block the
   * variables leftEdge and upperEdge are set
   * to contain the classifier directionality
   * of the corresponding left and right block.
   *
   * The variables leftDc, upperDc and upleftDc
   * are set to be holding information about the
   * DC level of the neighboring blocks.
   *
   * In case the neighboring blocks are
   * outside the curent slice an invalid
   * value of -1 or UCHAR_MAX is assigned
   * for the sake of simplicity.
   */

  if (frX >= frStepx && frY >= frStepy) {
    if (slicePredMap[frBlkNo-1] == currSliceID) {
      leftEdge  = classifBuf[frBlkNo-1];
      leftDc    = dcFrameBuf[frBlkNo-1];
    }
    else {
      leftEdge  = UCHAR_MAX;
      leftDc    = -1;
    }
    if (slicePredMap[frBlkNo-xBlocks] == currSliceID) {
      upperEdge = classifBuf[frBlkNo-xBlocks];
      upperDc   = dcFrameBuf[frBlkNo-xBlocks];
    }
    else {
      upperEdge = UCHAR_MAX;
      upperDc   = -1;
    }
    if (slicePredMap[frBlkNo-xBlocks-1] == currSliceID)
      upleftDc  = dcFrameBuf[frBlkNo-xBlocks-1];
    else
      upleftDc  = -1;
  }
  else if(frX >= frStepx) {
    if (slicePredMap[frBlkNo-1] == currSliceID) {
      leftEdge  = classifBuf[frBlkNo-1];
      leftDc    = dcFrameBuf[frBlkNo-1];
    }
    else {
      leftEdge  = UCHAR_MAX;
      leftDc    = -1;
    }
    upperEdge = UCHAR_MAX;
    upperDc   = -1;
    upleftDc  = -1;
  }
  else if(frY >= frStepy) {
    if (slicePredMap[frBlkNo-xBlocks] == currSliceID) {
      upperEdge = classifBuf[frBlkNo-xBlocks];
      upperDc   = dcFrameBuf[frBlkNo-xBlocks];
    }
    else {
      upperEdge = UCHAR_MAX;
      upperDc   = -1;
    }
    leftEdge  = UCHAR_MAX;
    leftDc    = -1;
    upleftDc  = -1;
  }
  else {
    leftEdge  = UCHAR_MAX;
    upperEdge = UCHAR_MAX;
    leftDc    = -1;
    upperDc   = -1;
    upleftDc  = -1;
  }

  /*
   * Mapping of directional classes (produced
   * by the gradient based cllassifier) into 7
   * directions corresponding to:
   *  0 - horizontal
   *  1 - diagonal /
   *  2 - vertical
   *  3 - diagonal \
   *  4 - flat
   *  5 - texture
   *  6 - block outside current slice or frame
   */
  if(leftEdge == UCHAR_MAX ) tmpLeft = 6;
  else if(leftEdge == N_0_8) tmpLeft = 0;
  else if(leftEdge == N_1_8) tmpLeft = 1;
  else if(leftEdge == N_2_8) tmpLeft = 1;
  else if(leftEdge == N_3_8) tmpLeft = 1;
  else if(leftEdge == N_4_8) tmpLeft = 2;
  else if(leftEdge == N_5_8) tmpLeft = 3;
  else if(leftEdge == N_6_8) tmpLeft = 3;
  else if(leftEdge == N_7_8) tmpLeft = 3;
  else if(leftEdge == FLAT0) tmpLeft = 4;
  else if(leftEdge == TEXT1) tmpLeft = 5;
  else if(leftEdge == TEXT2) tmpLeft = 5;
  else {
    printf("Wrong leftEdge %d in decodeIntraBlock()\n", leftEdge);
    exit(1);
  }
  if(upperEdge == UCHAR_MAX)  tmpUpper = 6;
  else if(upperEdge == N_0_8) tmpUpper = 0;
  else if(upperEdge == N_1_8) tmpUpper = 1;
  else if(upperEdge == N_2_8) tmpUpper = 1;
  else if(upperEdge == N_3_8) tmpUpper = 1;
  else if(upperEdge == N_4_8) tmpUpper = 2;
  else if(upperEdge == N_5_8) tmpUpper = 3;
  else if(upperEdge == N_6_8) tmpUpper = 3;
  else if(upperEdge == N_7_8) tmpUpper = 3;
  else if(upperEdge == FLAT0) tmpUpper = 4;
  else if(upperEdge == TEXT1) tmpUpper = 5;
  else if(upperEdge == TEXT2) tmpUpper = 5;
  else {
    printf("Wrong upperEdge %d in decodeIntraBlock()\n", upperEdge);
    exit(1);
  }

  /*
   * The context provided by Left and Upper blocks
   * is signaled by ipContextClassNo. This variable
   * can take 7 x 7 = 49 posible values.
   */
  ipContextClassNo = tmpLeft*7 + tmpUpper;

  /*
   * The table ipContextLists[] contains
   * the context dependent lists of intra
   * prediction methods. Its size is
   * 49 x (MAX_IP_LIST_SIZE+1). Each of the
   * 49 rows corresponds to a (L, U) context.
   * In ipContextLists[] the first column
   * represents the length of each list which
   * is here copied into variable numOfIntraPredMeths.
   */
  tmpPredNo = ipContextClassNo*(MAX_IP_LIST_SIZE+1);

  /*
   * Since the first column in ipContextLists[]
   * represents the length of each list this auxiliary
   * variable tmpPredNo is incremented to point to the
   * head of intra pred methods list.
   */
  tmpPredNo++;

  /*
   * Read and decode IPMR, ICNCB, and IDCD information from bitstream.
   * The intra prediction method rank is returned in methRank,
   * the index (including sign) is returned in dcdIndex and
   * the indication that there exist a COEFF LAYER is returned in cncFlag.
   */
  bestCom = vdxGetIntraPredDcdCnc(instance, qp, ipContextClassNo, &methRank, &dcdIndex,
                                  &cncFlag, BitError);
  if(*BitError == VDX_SAW_SYNC) return;

  /*
   * Get the current intra prediction method in methName.
   * from the context dependent list.
   */
  methName = ipContextLists[tmpPredNo + methRank];

  /*
   * The BM intra prediction method is not allowed
   * for 4x4 block sizes.
   */
  if(methName == IP_METH_BM && blkSize == SIZE_4x4) {
    printf("Wrong method %d in 4x4 intra chroma decoding\n", methName);
    nccExit(1);
  }
 
  bestMot = 0;

  if(methName == IP_METH_BM)
  {
    /*
     * For chroma in I frames or for any intra blocks
     * in P frames the BM search quits when 136/320 valid
     * vectors are found. A valid vector must point to
     * a block which is fully inside current slice.
     *
     * For luma in I frames up to 320 valid vectors
     * are allowed.
     */
    if(chroma == CHROMA || mode == INTRA_BLOCK)
    {
      totalNoOfVectors = maxNoOfValidVectors = 136;
      reducedSearchRange =  1;
    }
    else
    {
      totalNoOfVectors = maxNoOfValidVectors = AMOUNT_OF_INTRA_MOTION;
      reducedSearchRange =  0;
    }

    /*
     * The search quits when above maxNoOfValidVectors is reached
     * OR when all the 512 vectors have been scanned. The following
     * function finds the set of valid vectors and places their
     * offsets in validX[], validY[].
     */
    findValidVectors(validX, validY, validI,
                     maxNoOfValidVectors, totalNoOfVectors,
                     frX, frY, blkSizex, blkSizey, frSizex,
                     sliceXmin, sliceXmax, sliceYmin, sliceYmax,
                     slicePredMap, sliceSx, sliceSy, currSliceID,
                     mode);

    /*
     * Read and decode the block matching vector into ibmvY, ibmvX.
     * Depending on reducedSearchRange flag different FLC bits are
     * expected for reading intra BM vector components.
     */
    bestMot = vdxGetIntraBMvector(&ibmvY, &ibmvX, reducedSearchRange,
                                  instance, BitError);
    if(*BitError == VDX_SAW_SYNC) return;

    /*
     * Compensate by block matching the current intra predicted block.
     */
    dcPredSum = predIntraBlockMatch(recoFr, predBlk, frX, frY,
                                    blkSizex, blkSizey, frSizex,
                                    ibmvY, ibmvX);
  }
  else
  {
    /* Any intra prediction method which is not Block Matching */

    /*
     * Perform DC prediction (if methName is IP_METH_DC) or
     * one of the 11 directional prediction (if methName is
     * otherwise). The IP_METH_BM was dealt with just above.
     * If the given prediction method cannot be performed
     * such when at the border of the slice the return value
     * is negative. Otherwise, it's the sum of pred values.
     */
    dcPredSum = predIntra(recoFr, predBlk, frX, frY, methName, blkSizex,
                          blkSizey, frSizex, frSizey, sliceXmin, sliceXmax,
                          sliceYmin, sliceYmax, leftDc, upperDc, upleftDc,
                          slicePredMap, sliceSx, sliceSy, currSliceID);
    if(dcPredSum < -1) {
      printf("dcPredSum = %d < -1\n", dcPredSum);
      nccExit(1);
    }
  }

  /*
   * The inverse quantisation value of DC difference is
   * returned into the variable dcDiff.
   */
  dcDiff = ipdDCinvQuant(maxValue, maxDClevels, dcdIndex);

  /*
   * Intra prediction block values + DC are saturated to
   * 0, 255 range and placed in predBlk[]. The intra prediction
   * error block is prepared in diffBlk[] block buffer.
   */
  for(i=0; i<blkSize; i++)
  {
    predBlk[i] += dcDiff;

    if (predBlk[i] > 255)
      predBlk[i] = 255;
    else if (predBlk[i] < 0)
      predBlk[i] =   0;
  }

  /*
   * This stage is entered only if
   * COEFF LAYER is present. This is
   * signaled by the decoded cncFlag.
   */
  if (cncFlag)
  {

    /*
     * Initialize buffers.
     */
    for(i = 0; i < 64; i++)
    {
      methCoeffs[i] = 0;
      dctCoeffs[i]  = 0;
      diffBlk[i]    = 0;
    }

    /*
     * First read and decode METH8 / METH4 then
     * decode TCOEFF (or METH4 then TCOEFF if quadrants
     * of an 8x8 block are present).
     */

    if(blkSize == SIZE_8x8)
    {
      /*
       * If BM was uses as intra prediction then
       * the predicted texture is directionality classified
       * in order to select the set of coding methods.
       */
      if(methName == IP_METH_BM)
        blockEdge((int *)predBlk, blkSizex, blkSizey, &predDirIn);
      else
        predDirIn = TEXT2;

      /*
       * Derive class8x8.
       */

      class8x8 = methName + (NON_CLASSIFIED+1);

      /*
       * Determine how to translate a block having a certain
       * directionality into a tag specific to a set of
       * coding methods.
       */
      classTmp = changeClassNew(class8x8, predDirIn, qp, chroma,
                                mode,(int *)&rotateFlag);

      /*
       * Read and decode METH8.
       */
      bestAc = pedLoadMethod(classTmp, &codMeth8x8, chroma, blkSize,
                             instance, (int *)BitError);
      if(*BitError == VDX_SAW_SYNC) return;

      /*
       * If we deal with a coding method having active
       * quadrants then read and decode the METH4 - TCOEFF layer
       * for each of the active quadrant.
       */
      if(codMeth8x8 == METH_4x4_15)
      {
        /*
         * For each active quadrant classification of prediction is done.
         * Storing is done in methCoeffs[] about the block class
         * and what rotation is needed for 4x4 blocks.
         */
        findPredNums8x8((int *)predBlk, rotateFlag, qp, (int *)methCoeffs, 
                        chroma, codMeth8x8, mode);

        /*
         * Read and decode METH4 for each active quadrant.
         */
        bestAc += pedLoad4x4Info(methCoeffs, chroma, mode,
                                 qp, blkSize, codMeth8x8,
                                 instance, (int *)BitError);
        if(*BitError == VDX_SAW_SYNC) return;

        /*
         * Detect (if any) which are quandrants that are
         * coded with a 4x4 transform coding.
         */
        dct4x4Meth = findDctMeth((int *)methCoeffs, codMeth8x8, (int *)save4);

        if(dct4x4Meth >= METH_4x4_DCT_1 && dct4x4Meth <= METH_4x4_DCT_15)
        {
          /*
           * Read data for TCOEFF in case we have a 4x4
           * transform for some active quadrants.
           */
           bestAc += pedLoadDctCoeffs(dctCoeffs, dct4x4Meth, chroma,
                                      qp, mode, (int *)save4,
                                      instance, (int *)BitError);
          if(*BitError == VDX_SAW_SYNC) return;
        }
      }
      else if((codMeth8x8 >= METH_8x8_DCT && codMeth8x8 <= METH_8x3_DIAG_DCT) ||
        (codMeth8x8 >= METH_KLT_8x8_N_0_8 && codMeth8x8 <= METH_KLT_8x8_TEXT) ||
        (codMeth8x8 >= METH_DCT_8x8_N_0_8 && codMeth8x8 <= METH_DCT_8x8_N_4_8))
      {
        /*
         * If METH8 indicates an 8x8 or 4x4 transform coding
         * then read the TCOEFFs into methCoeffs[].
         */
        bestAc += pedLoadDctCoeffs(methCoeffs, codMeth8x8, chroma,
                                   qp, mode, NULL,
                                   instance, (int *)BitError);
        if(*BitError == VDX_SAW_SYNC) return;
      }

      /*
       * All the coefficients had been read and and decoded
       * so the recosntruction for 8x8 blocks can be performed.
       */
      decBits = dec8x8(predBlk, diffBlk, methCoeffs, dctCoeffs, qp,
                       codMeth8x8, class8x8, chroma, mode, predDirIn);
    }
    else if(blkSize == SIZE_4x4)
    {
      codMeth8x8 = METH_4x4_1;

      /*
       * The methCoeffs[13] holds the QP dependent list to be used
       * when accessing the coding method.
       */
      predDirIn = (u_char) methName;
      methCoeffs[13] = findPredNum4x4((int *)predBlk, qp, chroma, mode,
                                      blkSize, predDirIn, (int *)&rotateFlag);
      methCoeffs[14] = rotateFlag;

      /*
       * Read and decode METH4.
       */
      bestAc = pedLoad4x4Info(methCoeffs, chroma, mode,
                              qp, blkSize, codMeth8x8,
                              instance, (int *)BitError);
      if(*BitError == VDX_SAW_SYNC) return;

      /*
       * Detect whether the 4x4 block is coded by
       * using a 4x4 transform coding.
       */
      dct4x4Meth = findDctMeth((int *)methCoeffs, codMeth8x8, (int *)save4);

      /*
       * Read data for TCOEFF in case we have a 4x4
       * transform for the current 4x4 block.
       */
      if(dct4x4Meth >= METH_4x4_DCT_1 && dct4x4Meth <= METH_4x4_DCT_15)
      {
        bestAc += pedLoadDctCoeffs(dctCoeffs, dct4x4Meth, chroma,
                                   qp, mode, (int *)save4,
                                   instance, (int *)BitError);
        if(*BitError == VDX_SAW_SYNC) return;
      }

      /*
       * The coding method was put to methCoeffs[10].
       * In case it indicates 4x4 extrapolation then
       * call getExtSouJunct() to collect those pixels
       * lying at the border of the block. They will be
       * the pixels used for the extrapolation.
       */
      if (methCoeffs[10] == EXTRAPOL_METH)
        pels = getExtSouJunct((int *)predBlk, (int *)ExtBlk, blkSizex, blkSizey);


      /*
       * All the coefficients had been read and and decoded
       * so the reconstruction for 4x4 blocks can be performed.
       */
      decBits = dec4x4(predBlk, diffBlk, methCoeffs, dctCoeffs, qp,
                       ExtBlk, pels, chroma, mode, blkSize);
    }

    /*
     * Saturate to 0..255 range the reconstructed block.
     */
    for(i=0; i<blkSize; i++)
    {
      recoBlk[i] = predBlk[i] + diffBlk[i];
      if(recoBlk[i] > 255)    recoBlk[i] = 255;
      else if(recoBlk[i] < 0) recoBlk[i] = 0;
    }
  }
  else
    /*
     * If cncFlag is 0 then no prediction error is present.
     * In that case the recosntruction is same as intra prediction.
     */
    for(i=0; i<blkSize; i++)
      recoBlk[i] = predBlk[i];

  /*
   * Put the reconstructed block into the current
   * reconstruction frame buffer.
   */
  putIntBlktoByteImage((int *)recoBlk, recoFr, blkSizex, blkSizey,
                       frX, frY, frSizex);

  /*
   * Next, the buffers classifBuf[] and dcFrameBuf[] are
   * updated to reflect the directionality and DC level
   * of the current reconstructed block. This information
   * is used when intra predicting the following intracoded
   * blocks inside the frame.
   * The "granularity" for luma is 8x8 whereas for chroma
   * is 4x4 irrespective if the current chroma block is 8x8.
   * Therefore, for chroma 8x8 the 4x4 left and 4x4 upper
   * neighbors of first quadrant are actually used for
   * intra DC prediction and context class building.
   */
  for(addy=0; addy<blkSizey; addy+=frStepy)
    for(addx=0; addx<blkSizex; addx+=frStepx)
    {
      addNo = addy / frStepy * xBlocks + addx / frStepx;       /* integer division */

      /*
       * Extract the current block.
       */
      getIntBlkFromByteImage(recoFr, (int *)tmpBlk, frStepx, frStepy,
                             frX+addx, frY+addy, frSizex);

      /*
       * Directionality classify the block.
       */
      blockEdge((int *)tmpBlk, frStepx, frStepy, 
                &classifBuf[frBlkNo+addNo]);

      /*
       * Map the Texture coarse to texture smooth since
       * they are treated as same during context class building.
       */
      if(classifBuf[frBlkNo+addNo] == TEXT2)
        classifBuf[frBlkNo+addNo] = TEXT1;

      /*
       * Store also the quantised DC used later for intra DC prediction.
       */
      for(i=0, dcRecoSum = 0; i < frStepx * frStepy; i++)
        dcRecoSum += tmpBlk[i];

      dcFrameBuf[frBlkNo+addNo] = ipdDCquantAndInvQuant(dcRecoSum, frStepx*frStepy,
                                                        maxValue, maxLevel,
                                                        (int *)&tmpInd);
    }

  /*
   * Inform to upper level the bits and error spent
   * to intra encode the current block.
   */
  *acBits  = bestAc;
  *ovhBits = bestCom + bestMot;

  /*
   * The embedded deblocking filtering is done within the intra block
   * coding loop just after the current block has been reconstructed.
   * The filtering is done in-place and for all picture components
   * (luminance, chorminance u and chrominance v) but only in I-frames.
   * For intra blocks in P-frames the filtering is done at the end
   * of frame coding in same manner as for inter blocks.
   */
  if(mode == FIRST_BLOCK)
    filCompFiltIntra(recoFr, frSizex, frSizey, qp, recoFr, frX, frY);
}

/***************************************************************************
 * Function    : SendIntraMccRegion()
 *
 * Purpose     : Encodes and dumps the MB/Block layer bitstream for Intra
 *               mode blocks or MBs. Intra prediction is used followed by
 *               MCC prediction error coding.
 *
 * Arguments   : See the function call
 *
 * Returns     : void
 *
 *               Copyright (c) 1997 Nokia Research Center
 ***************************************************************************/

/* qpBlockLuma used to pass qpY to vdxGetDctTable() and vdxGetKltTable() */
extern int
  qpBlockLuma;

void decDecodeIntraMccRegion(struct frame *recoFr, /* Out: Reconstruction frame      */
                             int16 *slicePredMap,  /* In:  Slice label buffer        */
                             int currSliceID,      /* In:  Label of current slice    */
                             int sliceXmin,        /* In:  Xmin  of current slice    */
                             int sliceXmax,        /* In:  Xmax  of current slice    */
                             int sliceYmin,        /* In:  Ymin  of current slice    */
                             int sliceYmax,        /* In:  Ymax  of current slice    */
                             object *currBlockOrMBstruct, /* In: Ptr to MB/B object  */
                             int regionSize,       /* In:  Region size (16 or 8)     */
                             int picSizeX,         /* In:  Width of the image        */
                             int picSizeY,         /* In:  Height of the image       */
                             int mode,             /* In:  Intra in I or P frames    */
                             int qp,               /* In:  Quant. param. for luma    */
                             int *bitsPErrLuma,    /* Out: Bits for pred err in luma */
                             int *bitsPErrChroma,  /* Out: Bits for pred err chroma  */
                             int *bitsIPredLuma,   /* Out: Bits for luma intra pred  */
                             int *bitsIPredChroma, /* Out: Bits for chroma intra pred*/
                             vdxInstance_t
                               *instance,          /* In:  Current instance          */
                             int *BitError)        /* Out: For error handling        */
{
  Byte
    /*
     * Each input and output buffer has three components:
     * 0 (luma), 1 (chroma U), 2 (chroma V) which are
     * initialized in the followings.
     */
    *reco[3];

  static int16
    /*
     * Memory allocation for certain buffers are made
     * during the first invocation only.
     */
    first_invo = 1,
    /*
     * The buffer dcFrameBuf[0,1,2] holds the DC values for
     * all the luma and chroma blocks which have been so far
     * encoded. These values are used in DC intra prediction.
     * The allocation is made for every 8x8 luma and every
     * 4x4 chroma u and v blocks.
     */
    *dcFrameBuf[3];
  
  static u_char
    /*
     * The buffer classifBuf[0,1,2] holds the results of
     * classification for all the luma and chroma blocks
     * which have been so far encoded. This information is
     * used in context classification of Left and Upper
     * blocks. The allocation is made for every 8x8 luma
     * and every 4x4 chroma u and v blocks.
     */
    *classifBuf[3];

  int
    x, y, comp,
    sliceSx, sliceSy,
    xmin, xmax, ymin, ymax,
    chroma, xBlocks, yBlocks, qpY, qpUv,
    frQp, frSizex, frSizey, blkSizex, blkSizey,
    sliceMinx, sliceMaxx, sliceMiny, sliceMaxy,
    frStepx, frStepy, acBits, ovhBits,
    /*
     * Counters for number of bits used for
     * intra prediction for luma, chroma u,
     * and chroma v.
     */
    ipMethBits[3] = {0, 0, 0},
    /*
     * Counters for number of bits used for
     * intra prediction error coding for
     * luma, chroma u, and chroma v.
     */
    ipPErrBits[3]  = {0, 0, 0};

  reco[0] = recoFr->yptr, reco[1] = recoFr->uptr, reco[2] = recoFr->vptr;

  /*
   * Number of 8x8 blocks for horizontal
   * and vertical dimensions.
   */
  xBlocks = picSizeX/8;
  yBlocks = picSizeY/8;

  /*
   * Number of 8x8 blocks on horizontal
   * and vertical dimensions of the current
   * intra slice.
   */
  sliceSx = (sliceXmax-sliceXmin+1)/8;
  sliceSy = (sliceYmax-sliceYmin+1)/8;

  /*
   * QP and lambda used for luma and chroma.
   */

  /* qpBlockLuma used to pass qpY to vdxGetDctTable() and vdxGetKltTable() */
  qpBlockLuma = qpY = (int)qp;
  qpUv     = annexTlookup[qp];

  if (first_invo) 
  {
    first_invo = 0;
    /*
     * The buffer dcFrameBuf[0,1,2] holds the DC values for
     * all the luma and chroma blocks which have been so far
     * encoded. These values are used in DC intra prediction.
     * The allocation is made for every 8x8 luma and every
     * 4x4 chroma u and v blocks.
     */
    if ((dcFrameBuf[0] =
        (int16 *)malloc(3 * xBlocks * yBlocks * sizeof(int16))) 
        == NULL)
    {
      deb0f(stderr,"Cannot allocate static memory for dcFrameBuf\n");
      nccExit(1);
    }
    dcFrameBuf[1] = dcFrameBuf[0] + xBlocks * yBlocks;
    dcFrameBuf[2] = dcFrameBuf[0] + (xBlocks * yBlocks << 1);

    /*
     * The buffer classifBuf[0,1,2] holds the results of
     * classification for all the luma and chroma blocks
     * which have been so far encoded. This information is
     * used in context classification of Left and Upper
     * blocks. The allocation is made for every 8x8 luma
     * and every 4x4 chroma u and v blocks.
     */
    if((classifBuf[0] =
        (u_char *)malloc(3 * xBlocks * yBlocks * sizeof(u_char))) 
       == NULL) 
    {
      deb0f(stderr,"Cannot allocate static memory for classifBuf\n");
      nccExit(1);
    }
    classifBuf[1] = classifBuf[0] + xBlocks * yBlocks;
    classifBuf[2] = classifBuf[0] + (xBlocks * yBlocks << 1);
  }

  /*
   * For each frame component (0 = luma, 1 = chroma u, 2 = chroma v)
   * do the intra encoding and sending of bits for the 8x8 or 4x4
   * blocks contained within the MB or block. This order is important
   * because sending of bits takes place inside codeIntraBlock()
   * function.
   */
  for (comp = 0; comp < 3; comp++)
  {
    switch (comp)
    {
      /* Luminance parameters for comp 0 (luma) */
      case 0:

        chroma    = LUMA;
        frQp      = qpY;
        blkSizex  = blkSizey = 8;
        frStepx   = frStepy  = 8;
        frSizex   = picSizeX;
        frSizey   = picSizeY;
        sliceMinx = (sliceXmin+0);
        sliceMaxx = (sliceXmax+1);
        sliceMiny = (sliceYmin+0);
        sliceMaxy = (sliceYmax+1);
        xmin      = currBlockOrMBstruct->xmin;
        xmax      = currBlockOrMBstruct->xmax;
        ymin      = currBlockOrMBstruct->ymin;
        ymax      = currBlockOrMBstruct->ymax;

        break;

      /*
       * Chrominance parameters for comp 1 (chroma U)
       * are chronologically set after the luminance
       * component is completed.
       */
      case 1:

        chroma    = CHROMA;
        frQp      = qpUv;
        blkSizex  = blkSizey = ((regionSize == 16) ? 8 : 4);
        frStepx   = frStepy  = 4;
        frSizex   /= 2;
        frSizey   /= 2;
        sliceMinx /= 2;
        sliceMaxx /= 2;
        sliceMiny /= 2;
        sliceMaxy /= 2;
        xmin      /= 2;
        xmax      /= 2;
        ymin      /= 2;
        ymax      /= 2;

        break;

      /*
       * Chrominance parameters for comp 2 (chroma V)
       * are the same as for chroma U component.
       */
      case 2:

        break;

    }

    for(y = ymin; y < ymax; y += 8)
    {
      for(x = xmin; x < xmax; x += 8) 
      {
        /*
         * Intra encode and send the current 8x8 or 4x4 block
         * inside the given luma or chroma component.
         */
        decodeIntraBlock(reco[comp], dcFrameBuf[comp],
                         classifBuf[comp],
                         &acBits, &ovhBits,
                         mode, chroma, frQp,
                         maxDcLevels[frQp],
                         blkSizex, blkSizey, x, y,
                         frStepx, frStepy,
                         frSizex, frSizey, xBlocks,
                         sliceMinx, sliceMaxx,
                         sliceMiny, sliceMaxy,
                         slicePredMap,
                         sliceSx, sliceSy, currSliceID,
                         instance, BitError);
        if (*BitError == VDX_SAW_SYNC) return;

        /*
         * Accumulate the amount of bits for
         * IPMR, ICNCB, IDCD and IBMV.
         */
        ipMethBits[comp] += ovhBits;

        /*
         * Accumulate the amount of bits for
         * COEFF LAYER.
         */
        ipPErrBits[comp] += acBits;

      }
    }
  }

  /*
   * Total amount of bits for the current
   * region. The current region can be either
   * an 16x16 macroblock (four 8x8 luma + two 8x8 chroma)
   * or 8x8 block (one 8x8 luma + two 4x4 chroma).
   */
  *bitsIPredLuma   = ipMethBits[0];
  *bitsIPredChroma = ipMethBits[1] + ipMethBits[2];
  *bitsPErrLuma    = ipPErrBits[0];
  *bitsPErrChroma  = ipPErrBits[1] + ipPErrBits[2];
}


/***************************************************************************
 * Function    : Decode_I_Frame()
 *
 * Purpose     : Decodes the frame & slice layer bitstream for Intra
 *               coded pictures. Intra prediction is used followed by
 *               MCC prediction error coding.
 *
 * Arguments   : See the function call
 *
 * Returns     : void
 *
 *               Copyright (c) 1997 Nokia Research Center
 ***************************************************************************/

int32 Decode_I_Frame(struct frame *recoFr, /* Out: Reconstruction frame          */
                     int32 picSizeX,         /* In:  Width of the image            */
                     int32 picSizeY,         /* In:  Height of the image           */
                     int32 frameNumber,      /* In:  Picture frame number          */
                     int32 qp,               /* In:  Quant. param. for luma        */
                     int32 *bitsDctY,        /* Out: Picture amount of coeff bits Y*/
                     int32 *bitsDctU,        /* Out: Picture amount of coeff bits U*/
                     int32 *bitsDctV,        /* Out: Picture amount of coeff bits V*/
                     int32 *bitsOvrY,        /* Out: Picture amount of ovh bits Y  */
                     int32 *bitsOvrUV,       /* Out: Picture amount of ovh bits UV */
                     int32 *cumBitsSliceHeaders, /* Out: Accumulates bits for header */
                     int16 *slicePredMap,    /* In:  Slice label buffer            */
                     int sliceMode,          /* In:  Flag to signal slice mode     */
                     vdxInstance_t
                       *instance,            /* In:  Current instance          */
                     int *BitError)          /* Out: For error handling        */
{
  /*
   * The current Macroblock has a corresponding
   * data structure holding geometry and other
   * type of information.
   */
  object
    currMBstruct;

  int32
    x, y, mbNum, mbCols, tmp, totalBits = 0,
    currSliceID = 0, sliceCurrBits = *cumBitsSliceHeaders,
    sliceBits = DEFAULT_SLICE_BITS,
    intraSliceFlag, requestNewSlice, noMoreSlices = 0,
    sliceYmin, sliceXmin, sliceYmax, sliceXmax,
    bitsCoefY, bitsCoefC, bitsOvh, bitsOvhY, bitsOvhC;

  int
    IPB = 2, numMBsInFrame;

  *bitsDctY = *bitsDctU = *bitsDctV = *bitsOvrY = *bitsOvrUV = 0;

  /*
   * Number of MBs per row and inside frame.
   */
  mbCols = picSizeX >> 4;
  numMBsInFrame = (picSizeX >> 4)*(picSizeY >> 4);

  /*
   * The slicePredMap tells encoder the causal zone from where it
   * can use intra prediction to predict the current intra macroblock.
   * Slice parameters and slice map are initialized
   * before the coding loop starts.
   */
  for (tmp = (picSizeX * picSizeY >> 6) - 1; tmp >= 0; tmp--)
    slicePredMap[tmp] = SHRT_MAX;

  sliceYmin = picSizeY;
  sliceXmin = picSizeX;
  sliceYmax = 0;
  sliceXmax = 0;

  /*
   * If slice mode enabled, get the header of very first Slice's without an SSC
   */
  if (sliceMode)
  {
    bitsOvh = vdxGetSliceHead((int *)&currSliceID, (int *)&intraSliceFlag,
                              NULL, NULL, (int *)&picSizeX, (int *)&picSizeY, 
                              0, 1, &IPB,
                              instance, (int32 *)BitError);
    if (*BitError == VDX_SAW_SYNC) return (int32)0;

    *cumBitsSliceHeaders += bitsOvh;

    sliceCurrBits = bitsOvh;
  }


  /*
   * Macroblock intra decoding loop starts.
   */

  for (y = 0, mbNum = 0; y < picSizeY; y += 16)
    for (x = 0; x < picSizeX; x += 16, mbNum++)
    {
      if (sliceMode)
      {
        u_int32
          currBitPos = bibNumberOfFlushedBits(buff_in);

        /*
         * Check if the bits left in the buffer are sufficient
         * to decode one more macroblock inside the current slice
         */

        if (buff_in->bitsLeft < 8)
        {
          requestNewSlice = 1;

          bitsOvh = buff_in->bitsLeft + buff_in->numEmulPreventBits;

          if (buff_in->bitsLeft > 0)
          {
            /* Detect if all bits left are 0s */

            if (! vdxOnlyZeroBits(instance, (int32 *)BitError))
              requestNewSlice = 0;
            if(*BitError == VDX_SAW_SYNC) return (int32)0;
          }
          else if (buff_in->bitsLeft < 0) {
            *BitError = VDX_SAW_SYNC;
            return (int32)0;
          }

          if (requestNewSlice)
          {

            *cumBitsSliceHeaders += bitsOvh;

            /* Request new slice buffer refill */

            dcmRequestSliceRefill(&buff_in, (int *)&noMoreSlices);

            if (noMoreSlices)
            {
              y = picSizeY;
              break;
            }
            else
            {
              int
                localFrNum    = frameNumber,
                localIPB      = IPB,
                localPicSizeX = picSizeX,
                localPicSizeY = picSizeY;

              bitsOvh = vdxGetSliceHead((int *)&currSliceID, 
                                        (int *)&intraSliceFlag,
                                        (int *)&qp, &localFrNum,
                                        &localPicSizeX, &localPicSizeY,
                                        1, 1, &localIPB,
                                        instance, (int32 *)BitError);
              if(*BitError == VDX_SAW_SYNC) return (int32)0;

              if (localFrNum != frameNumber || localIPB != IPB ||
                  localPicSizeX != picSizeX || localPicSizeY != picSizeY) {
                deb0p("Locals are different than frame globals");
                nccExit(-1);
              }

              *cumBitsSliceHeaders += bitsOvh;

              sliceCurrBits = bitsOvh;

              fprintf(stdout, "Slice start at I %3d %2d\n", frameNumber, currSliceID);
              fprintf(stderr, "Slice start at I %3d %2d\n", frameNumber, currSliceID);

              if (currSliceID < mbNum)
              {
                /* Slice of a future frame, perhaps */
                *BitError = VDX_SAW_SYNC;
                return (int32)0;
              }
            }

            sliceYmin = picSizeY;
            sliceXmin = picSizeX;
            sliceYmax = 0;
            sliceXmax = 0;
          }
        }

        /* Jump to the begening of this slice */

        if (currSliceID > mbNum)
        {
          mbNum = currSliceID;
          y = currSliceID / mbCols;           /* integer division */
          x = currSliceID - y * mbCols;
          y <<= 4;
          x <<= 4;
        }
      }

      /*
       * The slicePredMap tells decoder about causal zone from where it
       * can use intra prediction to predict the current intra macroblock.
       * Here, all the four blocks of the current MB take the same
       * label as the current slice (which is currSliceID).
       */
      tmp = (y >> 3) * (picSizeX >> 3) + (x >> 3);
      slicePredMap[tmp] =
      slicePredMap[tmp + 1] =
      slicePredMap[tmp + (picSizeX >> 3)] =
      slicePredMap[tmp + (picSizeX >> 3) + 1] = (int16) currSliceID;

      /*
       * Update the geometry of the slice which has developped so far.
       */
      if (y < sliceYmin) sliceYmin = y;
      if (x < sliceXmin) sliceXmin = x;
      if (y + 15 > sliceYmax) sliceYmax = y + 15;
      if (x + 15 > sliceXmax) sliceXmax = x + 15;

      /*
       * Current region is always a 16x16 pel macroblock.
       */
      currMBstruct.xmin  = (short)x;
      currMBstruct.xmax  = (short)x + 15;
      currMBstruct.ymin  = (short)y;
      currMBstruct.ymax  = (short)y + 15;

      /*
       * Actual intra decoding of current MB (given its position
       * and the geometry of the slice it belongs to).
       */
      decDecodeIntraMccRegion(recoFr, slicePredMap, currSliceID,
                              sliceXmin, sliceXmax, sliceYmin, sliceYmax,
                              &currMBstruct, 16,
                              picSizeX, picSizeY, FIRST_BLOCK,
                              qp, (int *)&bitsCoefY, (int *)&bitsCoefC, 
                              (int *)&bitsOvhY, (int *)&bitsOvhC,
                              instance, (int *)BitError);
      if(*BitError == VDX_SAW_SYNC) return (int32)0;

      /*
       * Frame bit counters are updated.
       */
      bitsOvh    = bitsOvhY + bitsOvhC;

      *bitsDctY += bitsCoefY;
      *bitsDctU += bitsCoefC;
      *bitsOvrY += bitsOvh;

      sliceCurrBits += bitsCoefY + bitsCoefC + bitsOvh;
    }

  /*
   * This condition initializes the bitstream for the next frame/slice.
   */
  if (! noMoreSlices)
  {
    *cumBitsSliceHeaders += buff_in->bitsLeft + buff_in->numEmulPreventBits;
    dcmRequestSliceRefill(&buff_in, (int *)&noMoreSlices);
  }

  /*
   * I-frame total amount of bits is computed and returned.
   */
  totalBits = *bitsDctY + *bitsDctU + *bitsDctV +
                          *bitsOvrY + *bitsOvrUV;

  return totalBits;
}

