#ifndef _MAIN_H
#define _MAIN_H

#include <stdio.h>
#include <time.h>
#include <limits.h>

#include "globals.h"
#include "mainFlags.h"
#include "tools.h"
#include "nrctypes.h"

#define ENABLE_FULL_MB_MODE      1
#define DEFAULT_SLICE_BITS       -1

/* Complexity modes.
 * 
 * COMPLEXITY_MODE_ANCHOR    High Complexity (best quality)
 *
 * COMPLEXITY_MODE_PRACTICAL Preferred mode of operation. Able to encode
 *                           most of the QCIF sequences real-time on 600MHz
 *                           Pentium systems. Minor degradation compared
 *                           to Anchor mode.
 *
 * COMPLEXITY_MODE_LOW       Further reduced complexity with some degradation
 *                             - No 32x32 motion estimation.
 *                             - No motion search around the predictor (motion
 *                               params of the initial optimization are used).
 *                             - Prediction of motion parameters is not checked
 *                               if initial SE is too large
 */

#define COMPLEXITY_MODE_ANCHOR    0
#define COMPLEXITY_MODE_PRACTICAL 1
#define COMPLEXITY_MODE_LOW       2

/* Region modes */
#define REGMODE_INTER_NOPRED     0
#define REGMODE_INTER_PRED       1
#define REGMODE_INTER_PRED_DELTA 2
#define REGMODE_INTER_COPY       3
#define REGMODE_INTRA            4
#define REGMODE_SPLITTED         5

#define REGMODE_INTER_NOPRED_8x8     6
#define REGMODE_INTER_PRED_8x8       7
#define REGMODE_INTER_PRED_DELTA_8x8 8
#define REGMODE_INTER_COPY_8x8       9
#define REGMODE_INTRA_8x8            10

/* Prediction error coding modes */
#define PECMODE_SKIPPED          0
#define PECMODE_CODED            1

/* Prediction directions */
#define PREDDIR_NOTHING          0
#define PREDDIR_UP               1
#define PREDDIR_LEFT             2
#define PREDDIR_BOTH             3

/* Prediction selection (when two possibilities from the same direction)
 * PREDSEL_1 refers to upper/leftmost option, PREDSEL_2 to the lower/
 * rightmost option.
 */
#define PREDSEL_1                0
#define PREDSEL_2                1
#define PREDSEL_NOTNEEDED        0
#define PREDSEL_NEEDED           1

/* Maximum affine index that can be transmitted */
#define AFFINE_MAX_INDEX         1108

/* If PSC_BitStuff_Enable is undefined, the encoder
 * will not put bit stuffing against PSC emulation
 */
#define PSC_BitStuff_Enable

#define INTRA_PREDICTION          1
#define INTER_BLOCK               0
#define INTRA_BLOCK               1
#define FIRST_BLOCK               2

#define EDGE_THR                  1300  /* 4x4 classifiers variance
                                         * threshold */

/* Constants relating to entropy coding of motion coefficients */
   
#define MATRIX_SIZE               6

/* Definitions relating to region anf frame modes */

#define MODE_INTER                1
#define MODE_INTRA                2
#define MODE_COPY                 3
#define MODE_ZEROM                4
#define MODE_INTER_NO_MOTION      4
#define MODE_INTRA_PRED           5
#define MODE_BFRAME_INTER         7
#define MODE_BFRAME_COPY          8
#define MODE_ZM_NOINTER           10
#define MODE_MOT_NOINTER          11

#define AMOUNT_OF_INTRA_MOTION    320

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <limits.h>

#include "PredErrorCoder.h"
#include "decDct.h"
#include "encDct.h"
#include "toolDct.h"
#include "toolMCC.h"

#ifndef   HUGE
#define   HUGE  DBL_MAX
#endif


/*
 *
 * Initial_Motion_Estimation:
 *
 * Function:
 *       Estimates the affine motion parameters associated with the input
 *       region. Records the associated prediction mse, and sets zero_motion 
 *       flag to 1 if the estimated motion coefficients are zero
 * 
 *       Uses block matching + Gauss-Newton minimizations on current and
 *       previous images on subsampled and regular images
 *
 */

void Initial_Motion_Estimation(object *currRegion,
                               Byte *currY, Byte *prevY,
                               Byte *currU, Byte *prevU,
                               Byte *currV, Byte *prevV,
                               Byte *subsCurrY, Byte *subsPrevY,
                               int picSizeX, int picSizeY, 
                               int stage, int qp, double lambda,
                               Byte *zeroMotionFlag, double *motCoefLoc,
                               double *mseOut, int complexityMode, int *BMvectors);

/*
 *  Segment_Lagrangian_Cost() computes the lagrangian cost of the current
 *  segment (region).
 */

double lagrSegmentCost(struct frame *prevfr, struct frame *curfr,
                       object *current_region, object *object_list,int regionIndex,
                       int *motCoefQuant, int32 *affineDataPred,
                       int pic_sizex, int pic_sizey,
                       double lambda, int qp, int aQP, int regionMode,
                       int predDir, int tryPec, u_char *pecModeOut,
                       int *bitsOut);

void SendIntraMccRegion(struct frame *currfr, struct frame *recofr,
                        int16 *slicePredMap, int currSliceID,
                        int sliceXmin, int sliceXmax, int sliceYmin,
                        int sliceYmax, object *current_region,
                        int regionSize, int predPossible,
                        int sizex, int sizey, int mode, int qp, float lambda,
                        int *bits_reg_mcc_y, int *bits_reg_mcc_uv,
                        int *bits_reg_ovh_y, int *bits_reg_ovh_uv,
                        int *bits_reg_mode, int gen_bitstr_flag, int complexityMode);

/************************************************************************
 *  Code_P_Frame() performs the complete prediction & reconstruction
 *  of one frame (recfr) from the previous reconstructed frame (prevfr).
 *  First, it performs motion estimation, motion compensation and then
 *  prediction error coding. It returns the total number of bits encoded.
 *
 *  Input args:
 *    prevfr (a pointer to Yuv frame buffer holding the reference frame)
 *    currfr (a pointer to Yuv frame buffer holding the current original frame)
 *    pic_sizex, pic_sizey (luminance frame size)
 *    fr_no (the frame number of the currently decoded frame)
 *    merge_flag (flag allowing for enabling/disabling of merging)
 *    remove_flag (flag allowing for enabling/disabling of coeff. removal)
 *    qp (quantizer parameter for DCT in prediction error coding)
 *    lambda (the R-D value used in merging for the Lagrangian cost)
 * 
 *  Output args:
 *    predfr (a pointer to Yuv frame buffer holding the mc predicted frame)
 *    recfr (a pointer to Yuv frame buffer holding the reconstructed frame)
 *    filteredfr (a pointer to Yuv frame buffer holding the filtered frame)
 *    object_list (array of objects of the decoded regions)
 *    no_of_regions (total number of regions in the decoded frame)
 *    bits (number of bits spent in various submodules of the encoder)
 *    psnr_y, psnr_u, psnr_v (PSNR for the reconstructed Y, u, and v frames)
 *    
 *
 *  Warnings:
 *    object_list needs to be pre-allocated large enough.
 *    prevfr, predfr, recfr need to be pre-allocated for Yuv frames.
 *************************************************************************/

int Code_P_Frame(struct frame **longTermMem, struct frame *curfr,
                 struct frame *prev_orig, struct frame *predfr,
                 struct frame *recfr, struct frame *filteredfr,
                 int pic_sizex, int pic_sizey, int l,
                 object *object_list, int qp, double lmb,
                 Bits *bits, int numRefFrames,
                 int16 *slicePredMap, int sliceMode, int sliceBits,
                 int complexityMode, int useLoopFilter);

int Code_I_Frame(struct frame *origfr, struct frame *resufr,
                 int picSizexY, int picSizeyY, int fr_no, int qpY,
                 Bits *bits,
                 int16 *slicePredMap, int sliceMode, int sliceBits, int complexityMode);


void Dump_Huff_PDFs(char *filename, char *header);

/*** modulCoeffRemove.c ***/

double coeffRem(object *currReg, object *object_list,int regionIndex, double *motCoefGlob, double *motCoefGlobPred,
                int *motCoefOrthQuant, int32 *affineDataPred, struct frame *curfr,
                struct frame *prevfr, int sizex, int sizey, int qp, int aQP,
                double lambda, unsigned char *regionMode, unsigned char *pecMode,
                int predDir, int testWithPrediction);

/*** toolCoeffRemove.c ***/

void removeCoeff(double R[MATRIX_SIZE][MATRIX_SIZE],double *z,
                 double *motCoeffs, int *tab_reduce, int *numCoeffsLeft);
int clearZeros(double R[MATRIX_SIZE][MATRIX_SIZE], double *z, double *motCoeffs,
               int *tab_reduce, int *numCoeffsLeft, int aQP);

/*** split_label.c ***/

void split_one_label(object *currRegion, object *object_list,
                     int sizex, int sizey, int pic_sizex, int pic_sizey,
                     int refFrame);

/*** toolGaussNewton.c ***/

double gnsGaussNewtonStep(object *region, int picSizeX, int picSizeY,
                          Byte *currY, Byte *prevY, double newMotCoef[],
                          double oldMotCoef[]);  

/*** intra_pred.c ***/

#endif
