/*
 * mcp.c
 *
 *      Copyright (c) 1999 Nokia Research Center, Tampere, FINLAND
 *
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "nrctypes.h"
#include "debug.h"
#include "mainFlags.h"
#include "mcp.h"
#include "mvc.h"
#include "tools.h"
#include "decDemux.h"

int affineQPs[32] =
  /* 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 */
  {2,2,2,2,2,2,3,3,3,3,3,3,3,4,4,4,4,4,4,4,5,5,5,5,5,6,6,6,7,7,8,8};

/* Fixed point numbers with 16 bits after the point */
#define AFFSTEP_MV_16F     4096
#define AFFSTEP_MVXY_16F   6664
#define AFFSTEP_DELTA_16F  -889
#define AFFSTEP_MV_8F      8192
#define AFFSTEP_MVXY_8F   12513
#define AFFSTEP_DELTA_8F  -3575

/*
 * mcpOrthToMvDelta
 * 
 *   Transforms motion parameters in orthonormalized system to
 *   'Motion Vector + deltas' representation
 */

void mcpOrthToMvDelta(int *motCoefOrth, int32 *affineData, int regionSize)
{
  int stepMv, stepMvXy, stepDelta;

  if(regionSize == 8){
    stepMv    = AFFSTEP_MV_8F;
    stepMvXy  = AFFSTEP_MVXY_8F;
    stepDelta = AFFSTEP_DELTA_8F;
  }
  else{
    stepMv    = AFFSTEP_MV_16F;
    stepMvXy  = AFFSTEP_MVXY_16F;
    stepDelta = AFFSTEP_DELTA_16F;
  }

  affineData[0] = (stepMv   * motCoefOrth[0] +
                   stepMvXy * motCoefOrth[1] +
                   stepMvXy * motCoefOrth[2]);
  affineData[1] = stepDelta * motCoefOrth[1];
  affineData[2] = stepDelta * motCoefOrth[2];
  affineData[3] = (stepMv   * motCoefOrth[3] +
                   stepMvXy * motCoefOrth[4] +
                   stepMvXy * motCoefOrth[5]);
  affineData[4] = stepDelta * motCoefOrth[4];
  affineData[5] = stepDelta * motCoefOrth[5];
}

void mcpCalcFinalParams(object *currRegion)
{
  int32
    j,
    regionSize = currRegion->no_of_pels == 64 ? 8 : 16;

  /* (Refinement) Motion parameters */

  if(currRegion->regionMode == REGMODE_INTER_NOPRED     ||
     currRegion->regionMode == REGMODE_INTER_NOPRED_8x8 ||
     currRegion->regionMode == REGMODE_INTER_PRED_DELTA ||
     currRegion->regionMode == REGMODE_INTER_PRED_DELTA_8x8)
    mcpOrthToMvDelta(currRegion->motCoefOrth, currRegion->affineData,
		                 regionSize);
  else
    for(j = 0; j < 6; j++)
      currRegion->affineData[j] = 0;

  /* Check if this segment is motion compensated */

  if(currRegion->regionMode == REGMODE_INTER_NOPRED     ||
     currRegion->regionMode == REGMODE_INTER_NOPRED_8x8 ||
     currRegion->regionMode == REGMODE_INTER_PRED       ||
     currRegion->regionMode == REGMODE_INTER_PRED_8x8   ||
     currRegion->regionMode == REGMODE_INTER_PRED_DELTA ||
     currRegion->regionMode == REGMODE_INTER_PRED_DELTA_8x8){

    /* Add predicted motion parameters (if any) */

    if(currRegion->regionMode == REGMODE_INTER_PRED       ||
       currRegion->regionMode == REGMODE_INTER_PRED_8x8   ||
       currRegion->regionMode == REGMODE_INTER_PRED_DELTA ||
       currRegion->regionMode == REGMODE_INTER_PRED_DELTA_8x8)
      for(j = 0; j < 6; j++)
	      currRegion->affineData[j] += currRegion->affineDataPred[j];
  }
}

/*
 * mcpBuildSegmPred
 *
 *   Builds motion compansated prediction for 8x8 or 16x16 block.
 *   The motion parameters to be used are stored in the attribute
 *   list of the block. The prediction block is recosntructed to
 *   a 16x16 size buffer, starting from the upper left corner of
 *   the block.
 */

void mcpBuildSegmPred(int32 picSizeX, int32 picSizeY,
                      u_char *prevY, u_char *prevU, u_char *prevV,
                      object *currRegion, u_char *predY, u_char *predU,
                      u_char *predV)
{
  int32 
    j, jj, k,
    frameInd, predBufferInd,
    xMin = currRegion->xmin, 
    xMax = currRegion->xmax, 
    yMin = currRegion->ymin, 
    yMax = currRegion->ymax,
    regionSize = currRegion->no_of_pels == 64 ? 8 : 16;

  /* Check if this segment is motion compensated */

  if(currRegion->regionMode == REGMODE_INTER_NOPRED     ||
     currRegion->regionMode == REGMODE_INTER_NOPRED_8x8 ||
     currRegion->regionMode == REGMODE_INTER_PRED       ||
     currRegion->regionMode == REGMODE_INTER_PRED_8x8   ||
     currRegion->regionMode == REGMODE_INTER_PRED_DELTA ||
     currRegion->regionMode == REGMODE_INTER_PRED_DELTA_8x8){

    /* Build luminance and chrominance prediction */

    mvcBuildPredRegionYUV(currRegion, currRegion->affineData, prevY, predY,
			                    prevU, predU, prevV, predV, picSizeX, picSizeY);

  }
  else{

    /* Copy previous region from the reference frame */

    for (j = yMin, jj = 0; j <= yMax; j++, jj++)
    {
      frameInd = j * picSizeX + xMin;
      predBufferInd = jj << 4;
      for (k = xMin; k <= xMax;  k++, frameInd++, predBufferInd++)
      {
        predY[predBufferInd] = prevY[frameInd];
      }
    }

    xMin >>= 1;
    xMax >>= 1;
    yMin >>= 1;
    yMax >>= 1;
    picSizeX >>= 1;

    for (j = yMin, jj = 0; j <= yMax; j++, jj++)
    {
      frameInd = j * picSizeX + xMin;
      predBufferInd = jj << 3;
      for (k = xMin; k <= xMax;  k++, frameInd++, predBufferInd++)
      {
        predU[predBufferInd] = prevU[frameInd];
        predV[predBufferInd] = prevV[frameInd];
      }
    }
  }
}


/*
 * mcpTransformGlob2LocCoeffs
 *
 *   Transfers the coefficients from the global coordinate
 *   system of the frame to the local coordinate system of
 *   the bounding box
 */

void mcpTransformGlob2LocCoeffs(double A[6], object *current_region,
                                int pic_sizex, int pic_sizey, double B[6])
{

  int
    sizex, sizey,
    xmin, xmax, ymin, ymax,i;

  double
    a1, b1, a2, b2, ax, ay, bx, by, 
    centx1, centy1, centx2, centy2;
    
  xmin  = current_region->xmin; xmax = current_region->xmax;
  ymin  = current_region->ymin; ymax = current_region->ymax;
  sizex = xmax - xmin;
  sizey = ymax - ymin;
  centx1 = (double) (xmin + xmax) / 2.0;
  centy1 = (double) (ymin + ymax) / 2.0;
  a1 = (double) sizex / 2.0;
  b1 = (double) sizey / 2.0;
  centx2 = ((double) (pic_sizex-1)) / 2.0;
  centy2 = ((double) (pic_sizey-1)) / 2.0;
  a2 = ((double)(pic_sizex - 1)) / 2.0;
  b2 = ((double)(pic_sizey - 1)) / 2.0;

  ax = a1/a2;
  bx = (centx1 - centx2)/a2;

  ay = b1/b2;
  by = (centy1 - centy2)/b2;

  /**** HORIZONTAL VECTOR COMPONENT IN NEW COORD. SYSTEM ****/
  
  
  B[0] = A[0] + A[1]*by + A[2]*bx;
  
  B[1] = A[1]*ay;
  B[2] = A[2]*ax;
  
  /**** VERTICAL VECTOR COMPONENT IN NEW COORD. SYSTEM ****/
  
  B[3] = A[3] + A[4]*by + A[5]*bx;
  
  B[4] = A[4]*ay;
  B[5] = A[5]*ax;

  for (i = 0; i < 6;i++)             /* not needed in fixed point code !!*/
   if (fabs(B[i]) < 1e-10) B[i] = 0; 
  
}

/*
 * mcpTransformLoc2GlobCoeffs
 *
 *   Transfers the coefficients from the local coordinate
 *   system of the bounding box to the global coordinate
 *   system of the frame.
 */

void mcpTransformLoc2GlobCoeffs(double A[6], object *current_region,
                                int pic_sizex, int pic_sizey, double B[6])
{
  int
    sizex, sizey,i,
    xmin, xmax, ymin, ymax;

  double
    a1, b1, a2, b2, ax, ay, bx, by, 
    centx1, centy1, centx2, centy2;

  xmin  = current_region->xmin; xmax = current_region->xmax;
  ymin  = current_region->ymin; ymax = current_region->ymax;
  sizex = xmax - xmin;
  sizey = ymax - ymin;
  centx1 = (double) (xmin + xmax)/2.0;
  centy1 = (double) (ymin + ymax)/2.0;
  a1 = (double) sizex/2.0;
  b1 = (double) sizey/2.0;
  centx2 = ((double) (pic_sizex-1))/2.0;
  centy2 = ((double) (pic_sizey-1))/2.0;
  a2 = ((double)(pic_sizex-1))/2.0;
  b2 = ((double)(pic_sizey-1))/2.0;

  ax=a2/a1;
  bx=(centx2-centx1)/a1;

  ay=b2/b1;
  by=(centy2-centy1)/b1;

  /**** HORIZONTAL VECTOR COMPONENT IN NEW COORD. SYSTEM ****/
  
  
  B[0] = A[0] + A[1]*by + A[2]*bx;
  
  B[1] = A[1]*ay;
  B[2] = A[2]*ax;
  
  /**** VERTICAL VECTOR COMPONENT IN NEW COORD. SYSTEM ****/
  
  B[3] = A[3] + A[4]*by + A[5]*bx;
  
  B[4] = A[4]*ay;
  B[5] = A[5]*ax;

  for (i = 0; i < 6;i++)           /*not necessary in fixed point code!!*/
   if (fabs(B[i]) < 1e-10) B[i] = 0;
}

/*
 * mcpFindNeighs
 *
 *   Finds neighboring segments in one direction (up or left) which
 *   can be used for motion vector field prediction. 
 */

void mcpFindNeighs(object *object_list, int blkIndex, int regionSize, int dir,
                   int mbPerLine, int posX, int posY, int *numOfNeighs,
                   int neighIndexes[2], int refFrame, int *motionNeighs)
{

  int
    mbIndex = 5 * ((posX >> 4) + (posY >> 4) * mbPerLine),
    neighIndex;

  *numOfNeighs  = 0;
  *motionNeighs = 0;

  if(dir == PREDDIR_UP){

    /* Predictions across the slice boundaries are not allowed */
    if(posY >= 16)
      neighIndex = mbIndex - 5*mbPerLine;
    else
      neighIndex = mbIndex;

    if(posY > 0 &&
       object_list[neighIndex].sliceID == object_list[mbIndex].sliceID){

      /* 16x16 macroblock */
      if(regionSize == 16){

        neighIndex = mbIndex - 5*mbPerLine;

        /* Upstairs neighbour splitted */
        if (object_list[neighIndex].regionMode == REGMODE_SPLITTED){
          
          /* Two 8x8 neighbours */
          if(object_list[neighIndex+3].regionMode==REGMODE_INTER_NOPRED || 
            object_list[neighIndex+3].regionMode==REGMODE_INTER_PRED   || 
            object_list[neighIndex+3].regionMode==REGMODE_INTER_PRED_DELTA){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex].refFrameNumber == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex+3;
              *numOfNeighs += 1;
            }
          }
          if(object_list[neighIndex+4].regionMode==REGMODE_INTER_NOPRED || 
            object_list[neighIndex+4].regionMode==REGMODE_INTER_PRED   || 
            object_list[neighIndex+4].regionMode==REGMODE_INTER_PRED_DELTA){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex].refFrameNumber == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex+4;
              *numOfNeighs += 1;
            }
          }
        }
        /* Upstairs neighbour not splitted */
        else{
          
          /* One 16x16 neighbour */
          if(object_list[neighIndex].regionMode == REGMODE_INTER_NOPRED || 
            object_list[neighIndex].regionMode == REGMODE_INTER_PRED   || 
            object_list[neighIndex].regionMode == REGMODE_INTER_PRED_DELTA){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex].refFrameNumber == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex;
              *numOfNeighs += 1;
            }
          }
        }
      }

      /* 8x8 region */
      else{

        /* Prediction from outside the macroblock */
        if(!(posY & 15)){

          neighIndex = mbIndex - 5*mbPerLine;

          /* Upstairs neighbour splitted */
          if (object_list[neighIndex].regionMode == REGMODE_SPLITTED)
            neighIndex = blkIndex - 5*mbPerLine + 2;
        }
        /* Prediction from inside the macroblock  */
        else
          neighIndex = blkIndex - 2;

        if((object_list[neighIndex].regionMode == REGMODE_INTER_NOPRED || 
            object_list[neighIndex].regionMode == REGMODE_INTER_PRED   || 
            object_list[neighIndex].regionMode == REGMODE_INTER_PRED_DELTA)){

          *motionNeighs = 1;

          if(object_list[neighIndex].refFrameNumber == refFrame){
            neighIndexes[*numOfNeighs] = neighIndex;
            *numOfNeighs += 1;
          }
        }
      }
    }
  }
  else{

    /* Predictions across the slice boundaries are not allowed */
    if(posX >= 16)
      neighIndex = mbIndex - 5;
    else
      neighIndex = mbIndex;

    if(posX > 0 &&
       object_list[neighIndex].sliceID == object_list[mbIndex].sliceID){

      /* 16x16 macroblock */
      if(regionSize == 16){

        neighIndex = mbIndex - 5;

        /* Left neighbour splitted */
        if (object_list[neighIndex].regionMode == REGMODE_SPLITTED){
          
          /* Two 8x8 neighbours */
          if(object_list[neighIndex+2].regionMode==REGMODE_INTER_NOPRED || 
            object_list[neighIndex+2].regionMode==REGMODE_INTER_PRED   ||
            object_list[neighIndex+2].regionMode==REGMODE_INTER_PRED_DELTA){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex].refFrameNumber == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex+2;
              *numOfNeighs += 1;
            }
          }
          if(object_list[neighIndex+4].regionMode==REGMODE_INTER_NOPRED || 
            object_list[neighIndex+4].regionMode==REGMODE_INTER_PRED   || 
            object_list[neighIndex+4].regionMode==REGMODE_INTER_PRED_DELTA){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex].refFrameNumber == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex+4;
              *numOfNeighs += 1;
            }
          }
        }
        /* Left neighbour not splitted */
        else{
          
          /* One 16x16 neighbour */
          if(object_list[neighIndex].regionMode == REGMODE_INTER_NOPRED || 
            object_list[neighIndex].regionMode == REGMODE_INTER_PRED   || 
            object_list[neighIndex].regionMode == REGMODE_INTER_PRED_DELTA){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex].refFrameNumber == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex;
              *numOfNeighs += 1;
            }
          }
        }
      }

      /* 8x8 region */
      else{

        /* Prediction from outside the macroblock */
        if(!(posX & 15)){

          neighIndex = mbIndex - 5;

          /* Left neighbour splitted */
          if (object_list[neighIndex].regionMode == REGMODE_SPLITTED)
            neighIndex = blkIndex - 4;
        }
        /* Prediction from inside the macroblock  */
        else
          neighIndex = blkIndex - 1;

        if((object_list[neighIndex].regionMode == REGMODE_INTER_NOPRED || 
            object_list[neighIndex].regionMode == REGMODE_INTER_PRED   || 
            object_list[neighIndex].regionMode == REGMODE_INTER_PRED_DELTA)){

          *motionNeighs = 1;

          if(object_list[neighIndex].refFrameNumber == refFrame){
            neighIndexes[*numOfNeighs] = neighIndex;
            *numOfNeighs += 1;
          }
        }
      }
    }
  }
}

/*
 * mcpFindPredAlternatives
 *
 *   Finds neighboring segments which can be used for motion
 *   vector field prediction. 
 */

void mcpFindPredAlternatives(object *object_list, int blkIndex, int regionSize,
                             int posX, int posY, int mbPerLine,
                             u_char *predPossible, u_char *predDirsAvail,
                             u_char *predDir, u_char *predSelNeededUp,
                             u_char *predSelNeededLeft, int refFrame,
                             int *motionNeighs)
{

  int
    neighRegsUpper = 0,
    neighRegsLeft  = 0,
    neighIndexes[2],
    motionNeighsUp,
    motionNeighsLeft;

  /* Check the number of prediction possibilities from up */
  mcpFindNeighs(object_list, blkIndex, regionSize, PREDDIR_UP, mbPerLine,
                posX, posY, &neighRegsUpper, neighIndexes, refFrame,
                &motionNeighsUp);

  /* Check the number of prediction possibilities from left */
  mcpFindNeighs(object_list, blkIndex, regionSize, PREDDIR_LEFT, mbPerLine,
                posX, posY, &neighRegsLeft, neighIndexes, refFrame,
                &motionNeighsLeft);

  *motionNeighs = (motionNeighsUp || motionNeighsLeft) ? 1 : 0;

  if (neighRegsUpper > 0 && neighRegsLeft > 0){
    *predPossible  = 1;
    *predDirsAvail = PREDDIR_BOTH;
    *predDir       = PREDDIR_NOTHING; /* Decoded/chosen later if needed */
  }
  else if(neighRegsUpper > 0){
    *predPossible  = 1;
    *predDirsAvail = PREDDIR_UP;
    *predDir       = PREDDIR_UP;
  }
  else if(neighRegsLeft > 0){
    *predPossible  = 1;
    *predDirsAvail = PREDDIR_LEFT;
    *predDir       = PREDDIR_LEFT;
  }
  else{
    *predPossible  = 0;
    *predDirsAvail = PREDDIR_NOTHING;
    *predDir       = PREDDIR_NOTHING;
  }

  *predSelNeededUp   = (neighRegsUpper > 1) ? PREDSEL_NEEDED:PREDSEL_NOTNEEDED;
  *predSelNeededLeft = (neighRegsLeft > 1)  ? PREDSEL_NEEDED:PREDSEL_NOTNEEDED;
}

/*
 * mcpFindNeighs
 *
 *   Finds neighboring segments in one direction (up or left) which
 *   can be used for motion vector field prediction. 
 */

void mcpFindNeighsDecoder(objectBuffer_t *object_list, int blkIndex, 
                          int regionSize, int dir, int mbPerLine, int posX,
                          int posY, int *numOfNeighs, int neighIndexes[2],
                          int refFrame, int *motionNeighs)
{

  int
    mbIndex = ((posX >> 4) + (posY >> 4) * mbPerLine) << 2,
    neighIndex;

  *numOfNeighs  = 0;
  *motionNeighs = 0;

  if(dir == PREDDIR_UP){

    /* Predictions across the slice boundaries are not allowed */
    if(posY >= 16)
      neighIndex = mbIndex - (mbPerLine << 2);
    else
      neighIndex = mbIndex;

    if(posY > 0 &&
       object_list[neighIndex].sliceID == object_list[mbIndex].sliceID){

      /* 16x16 macroblock */
      if(regionSize == 16){

        neighIndex = mbIndex - (mbPerLine << 2);

        /* Upstairs neighbour splitted */
        if (object_list[neighIndex].regionMode >= REGMODE_SPLITTED){

          /* Two 8x8 neighbours */
          if(object_list[neighIndex + 2].regionMode==REGMODE_INTER_NOPRED_8x8 || 
             object_list[neighIndex + 2].regionMode==REGMODE_INTER_PRED_8x8   || 
             object_list[neighIndex + 2].regionMode==REGMODE_INTER_PRED_DELTA_8x8){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex + 2].refFrame == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex + 2;
              *numOfNeighs += 1;
            }
          }
          if(object_list[neighIndex + 3].regionMode==REGMODE_INTER_NOPRED_8x8 || 
             object_list[neighIndex + 3].regionMode==REGMODE_INTER_PRED_8x8   || 
             object_list[neighIndex + 3].regionMode==REGMODE_INTER_PRED_DELTA_8x8){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex + 3].refFrame == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex + 3;
              *numOfNeighs += 1;
            }
          }
        }
        /* Upstairs neighbour not splitted */
        else{

          /* One 16x16 neighbour */
          if(object_list[neighIndex].regionMode == REGMODE_INTER_NOPRED || 
             object_list[neighIndex].regionMode == REGMODE_INTER_PRED   || 
             object_list[neighIndex].regionMode == REGMODE_INTER_PRED_DELTA){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex].refFrame == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex;
              *numOfNeighs += 1;
            }
          }
        }
      }

      /* 8x8 region */
      else{

        /* Prediction from outside the macroblock */
        if(!(posY & 15)){

          neighIndex = mbIndex - (mbPerLine << 2);

          /* Upstairs neighbour splitted */
          if (object_list[neighIndex].regionMode >= REGMODE_SPLITTED)
            neighIndex = blkIndex - (mbPerLine << 2) + 2;
        }
        /* Prediction from inside the macroblock  */
        else
          neighIndex = blkIndex - 2;

        if(object_list[neighIndex].regionMode == REGMODE_INTER_NOPRED     || 
           object_list[neighIndex].regionMode == REGMODE_INTER_NOPRED_8x8 || 
           object_list[neighIndex].regionMode == REGMODE_INTER_PRED       || 
           object_list[neighIndex].regionMode == REGMODE_INTER_PRED_8x8   || 
           object_list[neighIndex].regionMode == REGMODE_INTER_PRED_DELTA ||
           object_list[neighIndex].regionMode == REGMODE_INTER_PRED_DELTA_8x8){
            
          *motionNeighs = 1;
            
          if(object_list[neighIndex].refFrame == refFrame){
            neighIndexes[*numOfNeighs] = neighIndex;
            *numOfNeighs += 1;
          }
        }
      }
    }
  }
  else{

    /* Predictions across the slice boundaries are not allowed */
    if(posX >= 16)
      neighIndex = mbIndex - 4;
    else
      neighIndex = mbIndex;

    if(posX > 0 &&
       object_list[neighIndex].sliceID == object_list[mbIndex].sliceID){

      /* 16x16 macroblock */
      if(regionSize == 16){

        neighIndex = mbIndex - 4;

        /* Left neighbour splitted */
        if (object_list[neighIndex].regionMode >= REGMODE_SPLITTED){

          /* Two 8x8 neighbours */
          if(object_list[neighIndex + 1].regionMode==REGMODE_INTER_NOPRED_8x8 || 
             object_list[neighIndex + 1].regionMode==REGMODE_INTER_PRED_8x8   ||
             object_list[neighIndex + 1].regionMode==REGMODE_INTER_PRED_DELTA_8x8){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex + 1].refFrame == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex + 1;
              *numOfNeighs += 1;
            }
          }
          if(object_list[neighIndex + 3].regionMode==REGMODE_INTER_NOPRED_8x8 || 
             object_list[neighIndex + 3].regionMode==REGMODE_INTER_PRED_8x8   || 
             object_list[neighIndex + 3].regionMode==REGMODE_INTER_PRED_DELTA_8x8){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex + 3].refFrame == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex + 3;
              *numOfNeighs += 1;
            }
          }
        }
        /* Left neighbour not splitted */
        else{

          /* One 16x16 neighbour */
          if(object_list[neighIndex].regionMode == REGMODE_INTER_NOPRED || 
             object_list[neighIndex].regionMode == REGMODE_INTER_PRED   || 
             object_list[neighIndex].regionMode == REGMODE_INTER_PRED_DELTA){
            
            *motionNeighs = 1;
            
            if(object_list[neighIndex].refFrame == refFrame){
              neighIndexes[*numOfNeighs] = neighIndex;
              *numOfNeighs += 1;
            }
          }
        }
      }

      /* 8x8 region */
      else{

        /* Prediction from outside the macroblock */
        if(!(posX & 15)){

          neighIndex = mbIndex - 4;

          /* Left neighbour splitted */
          if (object_list[neighIndex].regionMode >= REGMODE_SPLITTED)
            neighIndex = blkIndex - 3;
        }
        /* Prediction from inside the macroblock  */
        else
          neighIndex = blkIndex - 1;

        if(object_list[neighIndex].regionMode == REGMODE_INTER_NOPRED     || 
           object_list[neighIndex].regionMode == REGMODE_INTER_NOPRED_8x8 || 
           object_list[neighIndex].regionMode == REGMODE_INTER_PRED       || 
           object_list[neighIndex].regionMode == REGMODE_INTER_PRED_8x8   || 
           object_list[neighIndex].regionMode == REGMODE_INTER_PRED_DELTA ||
           object_list[neighIndex].regionMode == REGMODE_INTER_PRED_DELTA_8x8){
            
          *motionNeighs = 1;
            
          if(object_list[neighIndex].refFrame == refFrame){
            neighIndexes[*numOfNeighs] = neighIndex;
            *numOfNeighs += 1;
          }
        }
      }
    }
  }
}

/*
 * mcpFindPredAlternatives
 *
 *   Finds neighboring segments which can be used for motion
 *   vector field prediction. 
 */

 void mcpFindPredAlternativesDecoder(objectBuffer_t *object_list, int blkIndex,
                                    int regionSize, int posX, int posY, 
                                    int mbPerLine, u_char *predPossible, 
                                    u_char *predDirsAvail, u_char *predDir, 
                                    u_char *predSelNeededUp,
                                    u_char *predSelNeededLeft,
                                    int refFrame,
                                    int *motionNeighs)
{

  int
    neighRegsUpper = 0,
    neighRegsLeft  = 0,
    neighIndexes[2],
    motionNeighsUp,
    motionNeighsLeft;

  /* Check the number of prediction possibilities from up */
  mcpFindNeighsDecoder(object_list, blkIndex, regionSize, PREDDIR_UP, 
                       mbPerLine, posX, posY, &neighRegsUpper, neighIndexes,
                       refFrame, &motionNeighsUp);

  /* Check the number of prediction possibilities from left */
  mcpFindNeighsDecoder(object_list, blkIndex, regionSize, PREDDIR_LEFT, 
                       mbPerLine, posX, posY, &neighRegsLeft, neighIndexes,
                       refFrame, &motionNeighsLeft);

  *motionNeighs = (motionNeighsUp || motionNeighsLeft) ? 1 : 0;

  if (neighRegsUpper > 0 && neighRegsLeft > 0){
    *predPossible  = 1;
    *predDirsAvail = PREDDIR_BOTH;
    *predDir       = PREDDIR_NOTHING; /* Decoded/chosen later if needed */
  }
  else if(neighRegsUpper > 0){
    *predPossible  = 1;
    *predDirsAvail = PREDDIR_UP;
    *predDir       = PREDDIR_UP;
  }
  else if(neighRegsLeft > 0){
    *predPossible  = 1;
    *predDirsAvail = PREDDIR_LEFT;
    *predDir       = PREDDIR_LEFT;
  }
  else{
    *predPossible  = 0;
    *predDirsAvail = PREDDIR_NOTHING;
    *predDir       = PREDDIR_NOTHING;
  }

  *predSelNeededUp   = (neighRegsUpper > 1) ? PREDSEL_NEEDED:PREDSEL_NOTNEEDED;
  *predSelNeededLeft = (neighRegsLeft > 1)  ? PREDSEL_NEEDED:PREDSEL_NOTNEEDED;
}


