#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "mainFlags.h"
#include "mcp.h"

/*
 * coeffRem
 * 
 * Generates a set of motion models by removing motion parameters one by one.
 * Evaluates Lagrangian cost for all the models and chooses the one yielding
 * minumum cost.
 */

double coeffRem(object *currReg, object *object_list,int regionIndex, double *motCoefGlob,
                double *motCoefGlobPred, int *motCoefOrthQuant, int32 *affineDataPred,
                struct frame *currFr, struct frame *prevFr, int picSizeX, int picSizeY,
                int qp, int aQP, double lambda, unsigned char *regionMode,
                unsigned char *pecMode, int predDir, int testWithPrediction)
{

  double
    R[6][6], z[6], G_dummy[6][6], x[6], y_dummy[6], dummy, 
    rate, rateMin,
    motCoefGlobDelta[6],
    motCoefAll[6][6];

  int    
    j,
    bits, bitsOut,
    allCoefsRemoved = 0,
    coeffsTotal,
    coeffsLeft,
    modelsToTry,
    activeCoeffs[6],
    motCoefOrthQuantTest[6];

  u_char
    pecModeOut,
    regionModeTest;

  rateMin = HUGE;

  /* Select the mode: with or without prediction motion vector field */

  if(testWithPrediction){
    regionModeTest  = REGMODE_INTER_PRED_DELTA;
    allCoefsRemoved = currReg->zeroRefinement;

    for(j=0; j<6; j++)
      motCoefGlobDelta[j] = motCoefGlob[j] - motCoefGlobPred[j];
  }
  else{
    regionModeTest  = REGMODE_INTER_NOPRED;
    allCoefsRemoved = currReg->zero_motion;

    for(j=0; j<6; j++)
      motCoefGlobDelta[j] = motCoefGlob[j];
  }

  if(!allCoefsRemoved){
    
    for(j=0; j<6; j++)
      activeCoeffs[j] = 1;

    /* The matrices R and z are calculated using Cholesky
     * decomposition                                          
     */

    gnsBuildTriangLsEqnCRem(prevFr->yptr, currFr->yptr, currReg,
                            picSizeX, picSizeY, R, z, 
                            &dummy, G_dummy, y_dummy, motCoefGlob,
                            motCoefGlobDelta);

    /* Coefficients x corresponding to the orthonormal polynomials 
     * are obtained by solving R * x = z                           
     */

    gnsBacksubstitution(R, z, 6, x);

    for(j = 0; j < 6; j++)  
      motCoefAll[0][j] = x[j];

    /* Remove coefficients quantized to zero from the model */

    coeffsTotal = 6;

    clearZeros(R, z, motCoefAll[0], activeCoeffs, &coeffsTotal, aQP);

    coeffsLeft = coeffsTotal;

    if(coeffsTotal){

      modelsToTry = 1;

      /* Remove coefficients one by one (starting from the one estimated
       * to have least effect on prediction accuracy)
       */
      
      while(coeffsLeft > 1) {

        removeCoeff(R, z, motCoefAll[modelsToTry++],
                    activeCoeffs, &coeffsLeft);
      }

      /* Calculate Lagrangian costs for the models created */

      while(modelsToTry > 0){

        modelsToTry--;

        /* Start with full model */
        orthQuantParams(motCoefAll[coeffsTotal-modelsToTry-1],
                        motCoefOrthQuantTest, aQP);
        rate =
          lagrSegmentCost(prevFr, currFr, currReg,  object_list, regionIndex,
                          motCoefOrthQuantTest, affineDataPred,
                          picSizeX, picSizeY, lambda, qp, aQP,
                          regionModeTest, predDir, 1, &pecModeOut, &bitsOut);
        
        if(rate < rateMin || modelsToTry == coeffsTotal - 1){
          
          rateMin     = rate;
          *regionMode = regionModeTest;
          *pecMode    = pecModeOut;
          bits        = bitsOut;
      
          for(j = 0;j < 6;j++)
            motCoefOrthQuant[j]= motCoefOrthQuantTest[j];
        }
      }
    }
    else{
      allCoefsRemoved = 1;
      rateMin         = HUGE;
    }
  }

  if(allCoefsRemoved){
    rateMin = HUGE;
  }
  else{
    if (motCoefOrthQuant[0] == 0 &&
        motCoefOrthQuant[1] == 0 &&
        motCoefOrthQuant[2] == 0 &&
        motCoefOrthQuant[3] == 0 &&
        motCoefOrthQuant[4] == 0 &&
        motCoefOrthQuant[5] == 0){
      allCoefsRemoved = 1;
      rateMin         = HUGE;
    }
  }

  /* Lagrangian without motion (refinement) */

  if(testWithPrediction)
    regionModeTest = REGMODE_INTER_PRED;
  else
    regionModeTest = REGMODE_INTER_COPY;

  rate =
    lagrSegmentCost(prevFr, currFr, currReg,  object_list,regionIndex, motCoefOrthQuant,
                    affineDataPred, picSizeX, picSizeY, lambda,
                    qp, aQP, regionModeTest, predDir,
                    1, &pecModeOut, &bitsOut);

  if(rate < rateMin){
    rateMin     = rate;
    *regionMode = regionModeTest;
    *pecMode    = pecModeOut;
    bits        = bitsOut;
  }

  return rateMin;
}

