#include <stdio.h>

#include "tools.h"
#include "mainFlags.h"
#include "mcp.h"
#include "motionPred.h"
#include "mvc.h"


/*
 * findPredRegion
 *
 * Warps the region using motion parameters of the possible prediction
 * candidates (neighboring regions). Selects the one yielding minimum
 * SE and stores the information to the currRegion structure.
 */

void findPredRegion(object *currRegion, object *objectList, int regionIndex,
                    struct frame *currfr, struct frame *prevfr,
                    struct frame *predfr, double lambda,
                    int picSizeX, int picSizeY, int bitsOverheadUp,
                    int bitsOverheadLeft, int complexityMode,
                    int refFrame)
{
  int 
    i, x, y, xMin, xMax, yMin, yMax, frInd, seTmp, yPred, predInd, regInd,
    regionSize,
    neighInd, predDir, predSel,
    numOfNeighs, neighIndexes[4],
    numOfNeighsUp, neighIndexesUp[2],
    numOfNeighsLeft, neighIndexesLeft[2],
    mbPerLine = picSizeX/16,
    motionNeighs;

  double
    lagr, lagrMin;

  Byte
    *predY, *currY,
    *predU, *currU,
    *predV, *currV;

  int32
    affineDataPred[6],
    *affineDataPtr;

  regionSize = currRegion->no_of_pels == 64 ? 8 : 16;

  /* Find the indexes of the prediction candidates */

  if(currRegion->predDirsAvail == PREDDIR_BOTH ||
     currRegion->predDirsAvail == PREDDIR_UP)
    mcpFindNeighs(objectList, regionIndex, regionSize, PREDDIR_UP, mbPerLine,
                  currRegion->xmin, currRegion->ymin, &numOfNeighsUp,
                  neighIndexesUp, refFrame, &motionNeighs);
  else
    numOfNeighsUp = 0;

  if(currRegion->predDirsAvail == PREDDIR_BOTH ||
     currRegion->predDirsAvail == PREDDIR_LEFT)
    mcpFindNeighs(objectList, regionIndex, regionSize, PREDDIR_LEFT,mbPerLine,
                  currRegion->xmin, currRegion->ymin, &numOfNeighsLeft,
                  neighIndexesLeft, refFrame, &motionNeighs);
  else
    numOfNeighsLeft = 0;

  /* Put all indexes to the same array */

  numOfNeighs = numOfNeighsUp + numOfNeighsLeft;

  for(i = 0;i < numOfNeighs;i++)
    if(i < numOfNeighsUp)
      neighIndexes[i] = neighIndexesUp[i];
    else
      neighIndexes[i] = neighIndexesLeft[i-numOfNeighsUp];

  /* Evaluate prediction error with the predicted motion vector fields
   * and choose the one yielding best SE for further analysis
   */

  if(numOfNeighs > 1)
  {
    lagrMin = HUGE;

    predY = predfr->yptr;
    predU = predfr->uptr;
    predV = predfr->vptr;
    currY = currfr->yptr;
    currU = currfr->uptr;
    currV = currfr->vptr;

    xMin  = currRegion->xmin;
    xMax  = currRegion->xmax;
    yMin  = currRegion->ymin;
    yMax  = currRegion->ymax;

    for(i = 0;i < numOfNeighs;i++){

      neighInd          = neighIndexes[i];
      affineDataPtr     = &(objectList[neighInd].affineData[0]);

      affineDataPred[0] =
	      (affineDataPtr[0] +
         affineDataPtr[1] * (currRegion->ymin - objectList[neighInd].ymin) +
         affineDataPtr[2] * (currRegion->xmin - objectList[neighInd].xmin));
      affineDataPred[1] = affineDataPtr[1];
      affineDataPred[2] = affineDataPtr[2];

      affineDataPred[3] =
	      (affineDataPtr[3] +
	       affineDataPtr[4] * (currRegion->ymin - objectList[neighInd].ymin) +
	       affineDataPtr[5] * (currRegion->xmin - objectList[neighInd].xmin));
      affineDataPred[4] = affineDataPtr[4];
      affineDataPred[5] = affineDataPtr[5];

      mvcBuildPredRegionYUV(currRegion,
                            affineDataPred,
                            prevfr->yptr, predfr->yptr,
                            prevfr->uptr, predfr->uptr,
                            prevfr->vptr, predfr->vptr,
                            picSizeX, picSizeY);

      lagr = 0;

      for(y = yMin, yPred = 0; y <= yMax; y++, yPred++){
        frInd = y * picSizeX + xMin;
        predInd = yPred << 4;
        for(x = xMin; x <= xMax; x++, frInd++, predInd++){
          seTmp = predY[predInd] - currY[frInd];
          lagr += seTmp * seTmp;
        }
      }
      for(y = yMin >> 1, yPred = 0; y <= yMax >> 1; y++, yPred++){
        frInd = (y * picSizeX + xMin) >> 1;
        predInd = yPred << 3;
        for(x = xMin >> 1; x <= xMax >> 1; x++, frInd++, predInd++){
          seTmp = predU[predInd] - currU[frInd];
          lagr += seTmp * seTmp;
          seTmp = predV[predInd] - currV[frInd];
          lagr += seTmp * seTmp;
        }
      }

      if(i < numOfNeighsUp)
        lagr += lambda*bitsOverheadUp;
      else
        lagr += lambda*bitsOverheadLeft;

      /* Best choise so far */
      if(lagr < lagrMin){

        lagrMin       = lagr;
        regInd        = neighIndexes[i];

        /* Prediction from up */
        if(i < numOfNeighsUp){

          predDir = PREDDIR_UP;

          if(numOfNeighsUp == 2 && i == 1)
            predSel = PREDSEL_2;
          else
            predSel = PREDSEL_1;
        }
        /* Prediction from left */
        else{

          predDir =  PREDDIR_LEFT;

          if(numOfNeighsLeft == 2 && i == numOfNeighsUp+1)
            predSel = PREDSEL_2;
          else
            predSel = PREDSEL_1;
        }
      }
    }
  }
  else{

    /* Just one neighbour */
    regInd  = neighIndexes[0];
    predDir = numOfNeighsUp > 0 ? PREDDIR_UP : PREDDIR_LEFT;
    predSel = PREDSEL_1;
    lagrMin = 0;
  }

  currRegion->predDir        = predDir;
  currRegion->predSel        = predSel;
  currRegion->predNeighIndex = regInd;
  currRegion->predMinLagr    = lagrMin;

  /* Predicted field in MV+deltas format */
  affineDataPtr = &(objectList[regInd].affineData[0]);

  currRegion->affineDataPred[0] =
    (affineDataPtr[0] +
     affineDataPtr[1] * (currRegion->ymin - objectList[regInd].ymin) +
     affineDataPtr[2] * (currRegion->xmin - objectList[regInd].xmin));
  currRegion->affineDataPred[1] = affineDataPtr[1];
  currRegion->affineDataPred[2] = affineDataPtr[2];

  currRegion->affineDataPred[3] =
    (affineDataPtr[3] +
     affineDataPtr[4] * (currRegion->ymin - objectList[regInd].ymin) +
     affineDataPtr[5] * (currRegion->xmin - objectList[regInd].xmin));
  currRegion->affineDataPred[4] = affineDataPtr[4];
  currRegion->affineDataPred[5] = affineDataPtr[5];

  transMvDeltasToGlobPoly(currRegion->affineDataPred,
                          currRegion->motCoefPredGlob,
                          currRegion->xmin, currRegion->ymin,
                          picSizeX, picSizeY);

}

/*
 * motionPred
 *
 * Calculates the lagrangian cost for the region if coded with
 * motion vector field prediction. Changes the region mode if
 * this is the best alternative (so far).
 */

void motionPred(object *currRegion, object *objectList, int regionIndex,
                struct frame *currFr, struct frame *prevFr,
                struct frame *predFr, int qp, int aQP, double lambda,
                int picSizeX, int picSizeY, int complexityMode)
{
  int 
    i,
    regionSize,
    mbPerLine = picSizeX/16,
    motCoefOrthDelta[6];

  double
    lagr;

  unsigned char
    regionMode, pecMode;

  regionSize = currRegion->no_of_pels == 64 ? 8 : 16;

  lagr = coeffRem(currRegion, objectList,regionIndex, currRegion->motCoefPredRefinedGlob,
                  currRegion->motCoefPredGlob, motCoefOrthDelta,
                  currRegion->affineDataPred, currFr, prevFr, picSizeX, picSizeY,
                  qp, aQP, lambda, &regionMode, &pecMode, currRegion->predDir, 1);

  if(lagr < currRegion->lagr){
    
    currRegion->lagr       = lagr;
    currRegion->regionMode = regionMode;
    currRegion->pecMode    = pecMode;
    
    memcpy(currRegion->motCoefOrth, motCoefOrthDelta, 6 * sizeof(int));
    
    /* Parameters yielding the selected motion vector field
     * is stored to currRegion->affineData
     */

    if(regionMode == REGMODE_INTER_PRED_DELTA)
      mcpOrthToMvDelta(motCoefOrthDelta, currRegion->affineData, regionSize);
    else
      for(i = 0;i < 6;i++)
        currRegion->affineData[i] = 0;
      
    for(i = 0;i < 6;i++)
      currRegion->affineData[i] += currRegion->affineDataPred[i];    
  }
}


