#include "mainFlags.h"
#include "mcp.h"
#include "tools.h"
#include "encMux.h"

#define flor(X) (((X) < 0) ? ((int)(X) - 1) : ((int)(X)))

/*
 * transMvDeltasToGlobPoly
 *
 * Transforms 'Motion vector + deltas' representation to
 * affine motion paremeters in 'global system' (origin
 * in the middle of the image and picture corners in (-1,-1),
 * (1,-1), (-1,1) and (1,1))
 */

void transMvDeltasToGlobPoly(int32 *affineData, double *motCoefGlob,
			                       int xPos, int yPos,int picSizeX, int picSizeY)
{
  float64
    x, y, invPicSizeX, invPicSizeY, centX, centY;

  invPicSizeX = 1.0 / ((float64)(picSizeX - 1));
  centX       = (float64)(picSizeX >> 1) - 0.5;

  invPicSizeY = 1.0 / ((float64)(picSizeY - 1));
  centY       = (float64)(picSizeY >> 1) - 0.5;

  x = (2.0 * ((float64)xPos - centX)) * invPicSizeX;
  y = (2.0 * ((float64)yPos - centY)) * invPicSizeY;

  motCoefGlob[1] = centY*(affineData[1]/65536.0);
  motCoefGlob[2] = centX*(affineData[2]/65536.0);
  motCoefGlob[4] = centY*(affineData[4]/65536.0);
  motCoefGlob[5] = centX*(affineData[5]/65536.0);

  motCoefGlob[0] = 
    affineData[0]/65536.0 - y*motCoefGlob[1] - x*motCoefGlob[2];

  motCoefGlob[3] =
    affineData[3]/65536.0 - y*motCoefGlob[4] - x*motCoefGlob[5];
}

/*
 * transGlobPolyToMvDeltas
 *
 * Transforms affine motion paremeters in 'global system' (origin
 * in the middle of the image and picture corners in (-1,-1),
 * (1,-1), (-1,1) and (1,1)) to 'Motion vector + deltas' representation.
 */

void transGlobPolyToMvDeltas(double *motCoefGlob,int32 *affineData, 
			                       int xPos, int yPos,int picSizeX, int picSizeY)
{
  float64
    x, y, invPicSizeX, invPicSizeY, centX, centY;

  invPicSizeX = 1.0 / ((float64)(picSizeX - 1));
  centX       = (float64)(picSizeX >> 1) - 0.5;

  invPicSizeY = 1.0 / ((float64)(picSizeY - 1));
  centY       = (float64)(picSizeY >> 1) - 0.5;

  x = (2.0 * ((float64)xPos - centX)) * invPicSizeX;
  y = (2.0 * ((float64)yPos - centY)) * invPicSizeY;
 
  affineData[3] = (int32)(ceil(65536.0 * (motCoefGlob[3] + y*motCoefGlob[4] + x*motCoefGlob[5])));
  affineData[0] = (int32)(ceil(65536.0 * (motCoefGlob[0] + y*motCoefGlob[1] + x*motCoefGlob[2])));
  affineData[5] = (int32)(ceil(65536.0 * (motCoefGlob[5]/centX)));
  affineData[4] = (int32)(ceil(65536.0 * (motCoefGlob[4]/centY)));
  affineData[2] = (int32)(ceil(65536.0 * (motCoefGlob[2]/centX)));
  affineData[1] = (int32)(ceil(65536.0 * (motCoefGlob[1]/centY)));
}


/*
 * orthQuantParams
 *
 * Quantizes motion parameters.
 */

void orthQuantParams(double *motCoef, int *motCoefQuant, int aQP)
{
  int
    i, tmpInt;
  double
    tmpDbl;

  for(i = 0; i < 6; i++) {
    tmpDbl = motCoef[i] / aQP + 0.5;
    tmpInt = flor(tmpDbl);
    if(tmpInt >  AFFINE_MAX_INDEX) tmpInt =  AFFINE_MAX_INDEX;
    if(tmpInt < -AFFINE_MAX_INDEX) tmpInt = -AFFINE_MAX_INDEX;
    motCoefQuant[i] = tmpInt * aQP;
  }
}

/*
 * MatchBlock
 *
 * Performs block matching in SAE sense.
 */

void MatchBlock(Byte *prev, Byte *curr, int x, int y,
                int blkSize, int sizex, int sizey,
                int *vecX, int *vecY)
{

#define PENALTY         100

  int
    mvStep  = (blkSize >= 16) ? 2 : 1,    /* Step between motion vectors */
    mvRange = (blkSize >= 16) ? 14 : 15,  /* Max. range of the mv coordinates */
    i, j, k, l,                           /* Looping variables */
    maxX = sizex-blkSize-x,               /* Max. mv x-coordinate (in frame buffer) */
    maxY = sizey-blkSize-y,               /* Max. mv y-coordinate (in frame buffer) */
    bestSad, currSad,                     /* Best sad so far and current sad */
    yStart, yStop,                        /* Min. & max. mv y-coordinate (after range checks) */
    xStart, xStop,                        /* Min. & max. mv y-coordinate (after range checks) */
    blk2[32][32],                         /* Storage for current block */
    tmp;

  Byte
    *prevPtr;                             /* Temp. pointer to previous buffer */


  curr += sizex*y + x;
  prev += sizex*y + x;

  for(i = 0; i < blkSize; i++) {
    for(j = 0; j < blkSize; j++)
      blk2[i][j] = *(curr+j);
    curr += sizex;
  }
  curr -= blkSize*sizex;

  /*
   * Evaluate zero motion vector.
   */
  *vecX = 0;
  *vecY = 0;
  prevPtr = prev;
  for(bestSad = -PENALTY, i = 0; i < blkSize; i++) {
    for(j = 0; j < blkSize; j++) {
      tmp = blk2[i][j] - prevPtr[j];
      bestSad += tmp < 0 ? -tmp : tmp;
    }
    prevPtr += sizex;
  }

  if(bestSad>0)
  {

    /* Boundary & range checks */
    yStart = -y   < -mvRange ? -mvRange : -y;
    yStop  = maxY < mvRange  ? maxY : mvRange;
    xStart = -x   < -mvRange ? -mvRange : -x;
    xStop  = maxX < mvRange  ? maxX : mvRange;

    /*
     * Go throught all motion vectors beetween search area and evaluate
     * MAD for each of them.
     */
    for(k = yStart; k <= yStop; k += mvStep) {
      for(l = xStart; l <= xStop; l += mvStep) {

        if ((k|l) != 0) {
          prevPtr = prev + sizex*k + l;
          for(i = 0, currSad = 0; i < blkSize; i++) {
            for(j = 0; j < blkSize; j++) {
              tmp = blk2[i][j] - prevPtr[j];
              currSad += tmp < 0 ? -tmp : tmp;
            }
            if (currSad >= bestSad) break;
            prevPtr += sizex;
          }
          /*
           * Test whether current sad is better than the best sad so far
           */
          if(currSad < bestSad) {
            bestSad = currSad;
            *vecX = l;
            *vecY = k;
          }
        }

      } /*end of loop l  */
    } /*end of loop k */
  } /* of bestSad>0 */
}

/*
 * subsBlkMatching
 *
 * Initializes the parameters for the block matching procedure
 * and invokes it. Input images are expected to be subsampled
 * by a factor of two before entering this function.
 */

void subsBlkMatching(object *region, Byte *prev, Byte *curr,
                     int n, int m, int *vecX, int *vecY)
{
  int
    x, y,
    blkSizex = (region->xmax - region->xmin + 1) >> 1,
    blkSizey = (region->ymax - region->ymin + 1) >> 1,
    blkSize;

  /* In case region is not square we simply clip the sides
   * to make it square
   */
  blkSize = (blkSizex < blkSizey ? blkSizex : blkSizey);
  x        = (region->xmin + blkSizex - blkSize) >> 1;
  y        = (region->ymin + blkSizey - blkSize) >> 1;

  MatchBlock(prev, curr, x, y, blkSize, n, m, vecX, vecY);

  *vecX *= 2;
  *vecY *= 2;
}

/*
 * split_one_label
 *
 * Creates subregions by splitting the region into four (if
 * possible). New regions inherit the motion vector field
 * of the old one.
 */

void split_one_label(object *currRegion, object *object_list,
                     int sizex, int sizey, int picSizeX, int picSizeY,
                     int refFrame)
{

  int
    mbPerLine, mbIndex;

  object
    *outRegion;

  currRegion->splitted[refFrame] = 1;

  mbPerLine = picSizeX/16;
  mbIndex   = currRegion->xmin/16 + (currRegion->ymin/16)*mbPerLine;

  /*** REGION 1 - UPPER LEFT REGION ***/

  if(sizex == 16)
    outRegion = &object_list[5*mbIndex];
  else
    outRegion = &object_list[5*mbIndex+1];

  outRegion->splitted[refFrame]   = 0;
  outRegion->xmin       = currRegion->xmin;
  outRegion->ymin       = currRegion->ymin;
  outRegion->xmax       = currRegion->xmin + sizex - 1;
  outRegion->ymax       = currRegion->ymin + sizey - 1;
  outRegion->no_of_pels = sizex*sizey;
  outRegion->zero_motion= 0;

  /* New motion coefficients */
  mcpTransformGlob2LocCoeffs(currRegion->motCoefEstim[refFrame], outRegion,
                             picSizeX, picSizeY, outRegion->motCoefEstim[refFrame]);

  /*** REGION 2 - UPPER RIGHT REGION ***/

  if ((currRegion->xmin + 2*sizex - 1) < picSizeX){

    if(sizex == 16)
      outRegion = &object_list[5*(mbIndex+1)];
    else
      outRegion = &object_list[5*mbIndex+2];

    outRegion->splitted[refFrame]   = 0;
    outRegion->xmin       = currRegion->xmin + sizex;
    outRegion->xmax       = currRegion->xmax;
    outRegion->ymin       = currRegion->ymin;
    outRegion->ymax       = currRegion->ymin + sizey - 1;
    outRegion->no_of_pels = sizex*sizey;
    outRegion->zero_motion= 0;

    /* New motion coefficients */
    mcpTransformGlob2LocCoeffs(currRegion->motCoefEstim[refFrame], outRegion,
                               picSizeX, picSizeY, outRegion->motCoefEstim[refFrame]);
    
  }

  /*** REGION 3 - LOWER LEFT REGION ***/

  if ((currRegion->ymin + 2*sizey-1) < picSizeY){

    if(sizex == 16)
      outRegion = &object_list[5*(mbIndex+mbPerLine)];
    else
      outRegion = &object_list[5*mbIndex+3];

    outRegion->splitted[refFrame]   = 0;
    outRegion->xmin       = currRegion->xmin;
    outRegion->xmax       = currRegion->xmin + sizex - 1;
    outRegion->ymin       = currRegion->ymin + sizey;
    outRegion->ymax       = currRegion->ymax;
    outRegion->no_of_pels = sizex*sizey;
    outRegion->zero_motion= 0;

    /* New motion coefficients */
    mcpTransformGlob2LocCoeffs(currRegion->motCoefEstim[refFrame], outRegion,
                               picSizeX, picSizeY, outRegion->motCoefEstim[refFrame]);
 
  }

  /*** REGION 4 - LOWER RIGHT REGION ***/

  if ((currRegion->xmin+2*sizex-1) < picSizeX && 
      (currRegion->ymin+2*sizey-1) < picSizeY){

    if(sizex == 16)
      outRegion = &object_list[5*(mbIndex+mbPerLine+1)];
    else
      outRegion = &object_list[5*mbIndex+4];

    outRegion->splitted[refFrame] = 0;
    outRegion->xmin       = currRegion->xmin + sizex;
    outRegion->xmax       = currRegion->xmax;
    outRegion->ymin       = currRegion->ymin + sizey;
    outRegion->ymax       = currRegion->ymax;
    outRegion->no_of_pels = sizex*sizey;
    outRegion->zero_motion= 0;

    /* New motion coefficients */
    mcpTransformGlob2LocCoeffs(currRegion->motCoefEstim[refFrame], outRegion,
                               picSizeX, picSizeY, outRegion->motCoefEstim[refFrame]);

  }
}
