/*
 * toolDct.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Auxiliary functions for prediction error coding and 
 *    reconstructing the prediction error from the coding information.
 *
 */


#include "toolMCC.h"
#include "toolDct.h"

extern void varSizeDctFixedp( int *source, int *destination, int cols, int rows );


/*
 * Define global variables and tables
 */

/*
 * These coefs and cosis are used when DCT transform is calculated.
 * dct_init = 0 means that these values have not calculated.
 * dct_init = 1 means that these values have calculated.
 */
static double coef3[3];
static double coef4[4];
static double coef8[8];
static double cosi3[3][3];
static double cosi4[4][4];
static double cosi8[8][8];
static int dct_init = 0;

/*
 * Zigzag scanning orders for different sizes of DCT.
 */
extern short DctQuantTable[31][4096];

/*
 * These tables consists of lengths of variable length codes.
 * vlc_table_init = 0 means that tables have not initialized.
 * vlc_table_init = 1 means that tables have initialized.
 */
static struct vlc_table vlc_8x8x0Y[LAST][SIZE_8x8][LEVEL_07];
static struct vlc_table vlc_8x8i0Y[LAST][SIZE_8x8][LEVEL_07];
static struct vlc_table vlc_4x4x0Y[LAST][SIZE_4x4][LEVEL_07];
static struct vlc_table vlc_4x4i0C[LAST][SIZE_4x4][LEVEL_07];

static int vlc_table_init = 0;

/*
 * Define local function prototypes.
 */

static int countVlcBits(int last, int run, int level, int cols, int rows,
                        int qp, int chroma, int mode);

static void initCosines();
static void initCoefs();
static void initDctVlcTable();


void quantCoefINTER(int *source, int *destination, int size, int qp, int start)
{
  int i;		/* Loop increment counter	*/				

  short int *qtableptr;

  qtableptr = &(DctQuantTable[qp-1][2048]);

  /*
   * Go through all coefficients.
   */
  for(i=start; i<size; i++) {
    destination[i] = (int)(*(qtableptr+source[i]));
  }
}


int newQuantCoefint(int *dctCoef, int *quantDctCoef, int size, int qp,
		    int chroma, int mode, int start)
{
  int
    i,                  /* Loop increment counter			*/
    nonZeros = 0,	/* Number of non-zero coefficients		*/
    qp2;		/* 2*qp*/

  short int
    *qtableptr;		/* Table of quantized inter coefficient values  */


  for(i=0; i<start; i++) {
    quantDctCoef[i] = 0;
  }

  if(qp != 0){
    if(mode != INTRA_BLOCK) {
      qtableptr = &(DctQuantTable[qp-1][2048]);
      for(i=start; i<size; i++) {
        quantDctCoef[i] = (int)qtableptr[dctCoef[i]];
        if(quantDctCoef[i] != 0) 
          nonZeros++;
      }
    }else {
      qp2 = 2*qp;
      for(i=start; i<size; i++) {
        quantDctCoef[i] = dctCoef[i]/qp2;
        if(quantDctCoef[i] != 0) 
          nonZeros++;
        if(quantDctCoef[i] >= LEVEL_07)  
          quantDctCoef[i] = LEVEL_07-1;
        else if(quantDctCoef[i] <= -LEVEL_07)
          quantDctCoef[i] = 1-LEVEL_07;
      } /* for-loop */
    }
  }else{ /* if qp == 0 */
    for(i=start; i<size; i++) {
      quantDctCoef[i] = dctCoef[i];
      if(quantDctCoef[i] != 0) 
        nonZeros++;
      if(quantDctCoef[i] >= LEVEL_07)  
        quantDctCoef[i] = LEVEL_07-1;
      else if(quantDctCoef[i] <= -LEVEL_07)
        quantDctCoef[i] = 1-LEVEL_07;
    }
  }
  return nonZeros;
}

int varSizeDctAndQuant(int* diffBlk,int* quantDctCoeffs,int sizeOfBlk,int qp,
                       int chroma, int mode, int start,int sizedx, int sizedy,
                       double* diffBlkFloat,int* dctCoefInt)
{
  int
    ReturnVal;

  varSizeDctFixedp( diffBlk, dctCoefInt, sizedx,sizedy);

  ReturnVal = newQuantCoefint(dctCoefInt, quantDctCoeffs, sizeOfBlk, qp,
                              chroma, mode, start);

  return ReturnVal;
}


void newInvQuantCoefint(int *quantDctCoef, int *dctCoef, int size, int qp,
                        int chroma, int mode)
{
  int
    i,                  /* Loop increment counter                       */
    coefValue,		/* Coefficient value				*/
    qpEvenOdd,
    qp2;                /* 2*qp                                         */

  i = 0;
  if (qp == 0) {
    for (i = 0; i < size; i++)
      dctCoef[i] = quantDctCoef[i];
  }
  else {
    /*
     * Check if quantization parameter is odd or even and
     * make the inverse quantization depending of it.
     */
    qpEvenOdd = qp - 1 + (qp&1);
    qp2 = 2*qp;

    for (i = 0; i < size; i++) {
      coefValue = quantDctCoef[i];
      if (coefValue != 0) {
        if(coefValue < 0) {
          coefValue = qp2*coefValue - qpEvenOdd;
          if(coefValue <= -LEVEL_11) {		/* Clipping		*/
            coefValue = (-LEVEL_11 + 1);
          }
        }
        else {
          coefValue = qp2*coefValue + qpEvenOdd;
          if(coefValue >= LEVEL_11) {		/* Clipping		*/
            coefValue = (LEVEL_11 - 1);
          }
        }
      }
      dctCoef[i] = coefValue;
    }
  }
}

/*
 * initCosines:
 *
 * Parameters:
 *
 * Function:
 *	Initializes cosine values for variable size DCT calculation.
 *
 * Returns:
 *	void
 */

static void initCosines()
{
  int i, j;			/* Loop increment counters		*/
  double pi = 3.141592653589793238462643383279; /* 30 decimals pi-value	*/

  for(j = 0; j < COLS_3x8; j++) {
    for(i = 0; i < COLS_3x8; i++) {
      cosi3[i][j] = cos(((2*i + 1) * j * pi)/(2*COLS_3x8));
    }
  }
  for(j = 0; j < COLS_4x8; j++) {
    for(i = 0; i < COLS_4x8; i++) {
      cosi4[i][j] = cos(((2*i + 1) * j * pi)/(2*COLS_4x8));
    }
  }
  for(j = 0; j < COLS_8x8; j++) {
    for(i = 0; i < COLS_8x8; i++) {
      cosi8[i][j] = cos(((2*i + 1) * j * pi)/(2*COLS_8x8));
    }
  }
}


/*
 * initCoefs:
 *
 * Parameters:
 *
 * Function:
 *	Initializes coef values for variable size DCT calculation.
 *
 * Returns:
 *	void
 */

static void initCoefs()
{
  int i;		/* Loop increment counter			*/ 
  
  for(i = 0; i < COLS_3x8; i++) {
    if(i == 0) coef3[i] = sqrt(1.0/COLS_3x8);
    else       coef3[i] = sqrt(2.0/COLS_3x8);
  }
  for(i = 0; i < COLS_4x8; i++) {
    if(i == 0) coef4[i] = sqrt(1.0/COLS_4x8);
    else       coef4[i] = sqrt(2.0/COLS_4x8);
  }
  for(i = 0; i < COLS_8x8; i++) {
    if(i == 0) coef8[i] = sqrt(1.0/COLS_8x8);
    else       coef8[i] = sqrt(2.0/COLS_8x8);
  }
}


/*
 * initDctVlcTable:
 *
 * Parameters:
 *
 * Function:
 *	Initializes Variable Length Code (VLC) tables for variable
 *	size DCTs.
 *
 * Returns:
 *	void
 */

static void initDctVlcTable()
{
  int last, run, level;

  for(last = 0; last < LAST; last++) {
    for(run = 0; run < RUN; run++) {

      for(level = 0; level < LEVEL_07; level++) {
	if(run < SIZE_8x8) {
          vlc_8x8x0Y[last][run][level].length = ESCAPE;
          vlc_8x8i0Y[last][run][level].length = ESCAPE;
        }
	if(run < SIZE_4x4) {
	  vlc_4x4x0Y[last][run][level].length = ESCAPE;
	  vlc_4x4i0C[last][run][level].length = ESCAPE;
        }
      }

    }
  }

  vlc_8x8x0Y[0][ 0][ 1].length =  3;
  vlc_8x8x0Y[0][ 0][ 2].length =  5;
  vlc_8x8x0Y[0][ 0][ 3].length =  7;
  vlc_8x8x0Y[0][ 0][ 4].length =  8;
  vlc_8x8x0Y[0][ 0][ 5].length =  9;
  vlc_8x8x0Y[0][ 0][ 6].length = 10;
  vlc_8x8x0Y[0][ 0][ 7].length = 10;
  vlc_8x8x0Y[0][ 0][ 8].length = 11;
  vlc_8x8x0Y[0][ 0][ 9].length = 11;
  vlc_8x8x0Y[0][ 0][10].length = 12;
  vlc_8x8x0Y[0][ 0][11].length = 12;
  vlc_8x8x0Y[0][ 0][12].length = 12;
  vlc_8x8x0Y[0][ 1][ 1].length =  4;
  vlc_8x8x0Y[0][ 1][ 2].length =  7;
  vlc_8x8x0Y[0][ 1][ 3].length =  9;
  vlc_8x8x0Y[0][ 1][ 4].length = 11;
  vlc_8x8x0Y[0][ 1][ 5].length = 12;
  vlc_8x8x0Y[0][ 1][ 6].length = 13;
  vlc_8x8x0Y[0][ 2][ 1].length =  5;
  vlc_8x8x0Y[0][ 2][ 2].length =  9;
  vlc_8x8x0Y[0][ 2][ 3].length = 11;
  vlc_8x8x0Y[0][ 2][ 4].length = 13;
  vlc_8x8x0Y[0][ 3][ 1].length =  6;
  vlc_8x8x0Y[0][ 3][ 2].length = 10;
  vlc_8x8x0Y[0][ 3][ 3].length = 11;
  vlc_8x8x0Y[0][ 4][ 1].length =  6;
  vlc_8x8x0Y[0][ 4][ 2].length = 10;
  vlc_8x8x0Y[0][ 4][ 3].length = 13;
  vlc_8x8x0Y[0][ 5][ 1].length =  6;
  vlc_8x8x0Y[0][ 5][ 2].length = 11;
  vlc_8x8x0Y[0][ 5][ 3].length = 13;
  vlc_8x8x0Y[0][ 6][ 1].length =  7;
  vlc_8x8x0Y[0][ 6][ 2].length = 11;
  vlc_8x8x0Y[0][ 6][ 3].length = 13;
  vlc_8x8x0Y[0][ 7][ 1].length =  7;
  vlc_8x8x0Y[0][ 7][ 2].length = 11;
  vlc_8x8x0Y[0][ 8][ 1].length =  7;
  vlc_8x8x0Y[0][ 8][ 2].length = 11;
  vlc_8x8x0Y[0][ 9][ 1].length =  7;
  vlc_8x8x0Y[0][ 9][ 2].length = 11;
  vlc_8x8x0Y[0][10][ 1].length =  8;
  vlc_8x8x0Y[0][10][ 2].length = 13;
  vlc_8x8x0Y[0][11][ 1].length =  8;
  vlc_8x8x0Y[0][12][ 1].length =  8;
  vlc_8x8x0Y[0][13][ 1].length =  9;
  vlc_8x8x0Y[0][14][ 1].length =  9;
  vlc_8x8x0Y[0][15][ 1].length = 10;
  vlc_8x8x0Y[0][16][ 1].length = 10;
  vlc_8x8x0Y[0][17][ 1].length = 10;
  vlc_8x8x0Y[0][18][ 1].length = 10;
  vlc_8x8x0Y[0][19][ 1].length = 10;
  vlc_8x8x0Y[0][20][ 1].length = 10;
  vlc_8x8x0Y[0][21][ 1].length = 10;
  vlc_8x8x0Y[0][22][ 1].length = 10;
  vlc_8x8x0Y[0][23][ 1].length = 12;
  vlc_8x8x0Y[0][24][ 1].length = 12;
  vlc_8x8x0Y[0][25][ 1].length = 13;
  vlc_8x8x0Y[0][26][ 1].length = 13;
  vlc_8x8x0Y[1][ 0][ 1].length =  5;
  vlc_8x8x0Y[1][ 0][ 2].length = 10;
  vlc_8x8x0Y[1][ 0][ 3].length = 12;
  vlc_8x8x0Y[1][ 1][ 1].length =  7;
  vlc_8x8x0Y[1][ 1][ 2].length = 12;
  vlc_8x8x0Y[1][ 2][ 1].length =  7;
  vlc_8x8x0Y[1][ 3][ 1].length =  7;
  vlc_8x8x0Y[1][ 4][ 1].length =  7;
  vlc_8x8x0Y[1][ 5][ 1].length =  8;
  vlc_8x8x0Y[1][ 6][ 1].length =  8;
  vlc_8x8x0Y[1][ 7][ 1].length =  8;
  vlc_8x8x0Y[1][ 8][ 1].length =  8;
  vlc_8x8x0Y[1][ 9][ 1].length =  9;
  vlc_8x8x0Y[1][10][ 1].length =  9;
  vlc_8x8x0Y[1][11][ 1].length =  9;
  vlc_8x8x0Y[1][12][ 1].length =  9;
  vlc_8x8x0Y[1][13][ 1].length =  9;
  vlc_8x8x0Y[1][14][ 1].length =  9;
  vlc_8x8x0Y[1][15][ 1].length =  9;
  vlc_8x8x0Y[1][16][ 1].length =  9;
  vlc_8x8x0Y[1][17][ 1].length = 10;
  vlc_8x8x0Y[1][18][ 1].length = 10;
  vlc_8x8x0Y[1][19][ 1].length = 10;
  vlc_8x8x0Y[1][20][ 1].length = 10;
  vlc_8x8x0Y[1][21][ 1].length = 10;
  vlc_8x8x0Y[1][22][ 1].length = 10;
  vlc_8x8x0Y[1][23][ 1].length = 10;
  vlc_8x8x0Y[1][24][ 1].length = 10;
  vlc_8x8x0Y[1][25][ 1].length = 11;
  vlc_8x8x0Y[1][26][ 1].length = 11;
  vlc_8x8x0Y[1][27][ 1].length = 11;
  vlc_8x8x0Y[1][28][ 1].length = 11;
  vlc_8x8x0Y[1][29][ 1].length = 12;
  vlc_8x8x0Y[1][30][ 1].length = 12;
  vlc_8x8x0Y[1][31][ 1].length = 12;
  vlc_8x8x0Y[1][32][ 1].length = 12;
  vlc_8x8x0Y[1][33][ 1].length = 13;
  vlc_8x8x0Y[1][34][ 1].length = 13;
  vlc_8x8x0Y[1][35][ 1].length = 13;
  vlc_8x8x0Y[1][36][ 1].length = 13;
  vlc_8x8x0Y[1][37][ 1].length = 13;
  vlc_8x8x0Y[1][38][ 1].length = 13;
  vlc_8x8x0Y[1][39][ 1].length = 13;
  vlc_8x8x0Y[1][40][ 1].length = 13;
  vlc_8x8x0Y[0][ 0][ 0].length =  7;              /* escape-code */

  /* VLC table for INTRA TCOEF */

  vlc_8x8i0Y[0][ 0][ 1].length =  3;	/* 10s      */
  vlc_8x8i0Y[0][ 0][ 2].length =  4;	/* 110s      */
  vlc_8x8i0Y[0][ 0][ 3].length =  5;	/* 1110 s     */
  vlc_8x8i0Y[0][ 0][ 4].length =  6;	/* 0110 0s    */
  vlc_8x8i0Y[0][ 0][ 5].length =  6;	/* 0110 1s    */
  vlc_8x8i0Y[0][ 0][ 6].length =  7;	/* 0100 00s    */
  vlc_8x8i0Y[0][ 0][ 7].length =  7;	/* 0100 01s    */
  vlc_8x8i0Y[0][ 0][ 8].length =  7;	/* 0100 10s    */
  vlc_8x8i0Y[0][ 0][ 9].length =  8;	/* 0010 110s    */
  vlc_8x8i0Y[0][ 0][10].length =  9;	/* 0001 1011 s   */
  vlc_8x8i0Y[0][ 0][11].length = 10;	/* 0001 0000 0s  */
  vlc_8x8i0Y[0][ 0][12].length = 10;	/* 0001 0000 1s  */
  vlc_8x8i0Y[0][ 0][13].length = 10;	/* 0000 1101 0s  */
  vlc_8x8i0Y[0][ 0][14].length = 10;	/* 0000 1101 1s  */
  vlc_8x8i0Y[0][ 0][15].length = 10;	/* 0000 1110 0s  */
  vlc_8x8i0Y[0][ 0][16].length = 10;	/* 0000 1110 1s  */
  vlc_8x8i0Y[0][ 0][17].length = 10;	/* 0000 1111 0s  */
  vlc_8x8i0Y[0][ 0][18].length = 10;	/* 0000 1111 1s  */
  vlc_8x8i0Y[0][ 0][19].length = 12;	/* 0000 0100 011s  */
  vlc_8x8i0Y[0][ 0][20].length = 12;	/* 0000 0100 010s  */
  vlc_8x8i0Y[0][ 0][21].length = 13;	/* 0000 0101 0111s */
  vlc_8x8i0Y[0][ 0][22].length = 13;	/* 0000 0101 0110s */
  vlc_8x8i0Y[0][ 0][23].length = 13;	/* 0000 0101 0101s */
  vlc_8x8i0Y[0][ 0][24].length = 13;	/* 0000 0101 0100s */
  vlc_8x8i0Y[0][ 0][25].length = 13;	/* 0000 0101 0011s */
  vlc_8x8i0Y[0][ 1][ 1].length =  5;	/* 1111 s     */
  vlc_8x8i0Y[0][ 1][ 2].length =  7;	/* 0101 00s    */
  vlc_8x8i0Y[0][ 1][ 3].length =  8;	/* 0010 100s    */
  vlc_8x8i0Y[0][ 1][ 4].length =  9;	/* 0001 1110 s   */
  vlc_8x8i0Y[0][ 1][ 5].length = 11;	/* 0000 0011 11s  */
  vlc_8x8i0Y[0][ 1][ 6].length = 12;	/* 0000 0100 001s  */
  vlc_8x8i0Y[0][ 1][ 7].length = 13;	/* 0000 0101 0000s */
  vlc_8x8i0Y[0][ 2][ 1].length =  6;	/* 0101 1s    */
  vlc_8x8i0Y[0][ 2][ 2].length =  8;	/* 0010 101s    */
  vlc_8x8i0Y[0][ 2][ 3].length = 11;	/* 0000 0011 10s  */
  vlc_8x8i0Y[0][ 2][ 4].length = 11;	/* 0000 0010 01s  */
  vlc_8x8i0Y[0][ 3][ 1].length =  7;	/* 0101 01s    */
  vlc_8x8i0Y[0][ 3][ 2].length =  9;	/* 0001 1101 s   */
  vlc_8x8i0Y[0][ 3][ 3].length = 11;	/* 0000 0011 01s  */
  vlc_8x8i0Y[0][ 3][ 4].length = 13;	/* 0000 0101 0001s */
  vlc_8x8i0Y[0][ 4][ 1].length =  7;	/* 0100 11s    */
  vlc_8x8i0Y[0][ 4][ 2].length = 10;	/* 0001 0001 1s  */
  vlc_8x8i0Y[0][ 4][ 3].length = 12;	/* 0000 0000 111s  */
  vlc_8x8i0Y[0][ 5][ 1].length =  8;	/* 0010 111s    */
  vlc_8x8i0Y[0][ 5][ 2].length = 10;	/* 0001 0001 0s  */
  vlc_8x8i0Y[0][ 5][ 3].length = 13;	/* 0000 0101 0010s */
  vlc_8x8i0Y[0][ 6][ 1].length =  9;	/* 0001 1100 s   */
  vlc_8x8i0Y[0][ 6][ 2].length = 11;	/* 0000 0011 00s  */
  vlc_8x8i0Y[0][ 7][ 1].length =  9;	/* 0001 1111 s   */
  vlc_8x8i0Y[0][ 7][ 2].length = 11;	/* 0000 0010 11s  */
  vlc_8x8i0Y[0][ 8][ 1].length = 10;	/* 0001 0010 1s  */
  vlc_8x8i0Y[0][ 8][ 2].length = 11;	/* 0000 0010 10s  */
  vlc_8x8i0Y[0][ 9][ 1].length = 10;	/* 0001 0010 0s  */
  vlc_8x8i0Y[0][ 9][ 2].length = 12;	/* 0000 0000 110s  */
  vlc_8x8i0Y[0][10][ 1].length = 11;	/* 0000 1000 01s  */
  vlc_8x8i0Y[0][11][ 1].length = 11;	/* 0000 1000 00s  */
  vlc_8x8i0Y[0][12][ 1].length = 11;	/* 0000 0010 00s  */
  vlc_8x8i0Y[0][13][ 1].length = 12;	/* 0000 0100 000s  */
  vlc_8x8i0Y[1][ 0][ 1].length =  5;	/* 0111 s     */
  vlc_8x8i0Y[1][ 0][ 2].length =  7;	/* 0011 00s    */
  vlc_8x8i0Y[1][ 0][ 3].length =  8;	/* 0010 000s    */
  vlc_8x8i0Y[1][ 0][ 4].length =  9;	/* 0001 0011 s   */
  vlc_8x8i0Y[1][ 0][ 5].length = 10;	/* 0000 1000 1s  */
  vlc_8x8i0Y[1][ 0][ 6].length = 10;	/* 0000 1001 0s  */
  vlc_8x8i0Y[1][ 0][ 7].length = 11;	/* 0000 0001 00s  */
  vlc_8x8i0Y[1][ 0][ 8].length = 12;	/* 0000 0100 111s  */
  vlc_8x8i0Y[1][ 0][ 9].length = 12;	/* 0000 0100 110s  */
  vlc_8x8i0Y[1][ 0][10].length = 13;	/* 0000 0101 1111s */
  vlc_8x8i0Y[1][ 1][ 1].length =  7;	/* 0011 11s    */
  vlc_8x8i0Y[1][ 1][ 2].length = 10;	/* 0000 1001 1s  */
  vlc_8x8i0Y[1][ 1][ 3].length = 11;	/* 0000 0001 01s  */
  vlc_8x8i0Y[1][ 1][ 4].length = 12;	/* 0000 0100 101s  */
  vlc_8x8i0Y[1][ 2][ 1].length =  7;	/* 0011 10s    */
  vlc_8x8i0Y[1][ 2][ 2].length = 10;	/* 0000 1010 0s  */
  vlc_8x8i0Y[1][ 2][ 3].length = 12;	/* 0000 0100 100s  */
  vlc_8x8i0Y[1][ 3][ 1].length =  7;	/* 0011 01s    */
  vlc_8x8i0Y[1][ 3][ 2].length = 11;	/* 0000 0001 10s  */
  vlc_8x8i0Y[1][ 3][ 3].length = 13;	/* 0000 0101 1110s */
  vlc_8x8i0Y[1][ 4][ 1].length =  8;	/* 0010 001s    */
  vlc_8x8i0Y[1][ 4][ 2].length = 11;	/* 0000 0001 11s  */
  vlc_8x8i0Y[1][ 5][ 1].length =  8;	/* 0010 011s    */
  vlc_8x8i0Y[1][ 5][ 2].length = 13;	/* 0000 0101 1101s */
  vlc_8x8i0Y[1][ 6][ 1].length =  8;	/* 0010 010s    */
  vlc_8x8i0Y[1][ 6][ 2].length = 13;	/* 0000 0101 1100s */
  vlc_8x8i0Y[1][ 7][ 1].length =  9;	/* 0001 0100 s   */
  vlc_8x8i0Y[1][ 7][ 2].length = 13;	/* 0000 0101 1011s */
  vlc_8x8i0Y[1][ 8][ 1].length =  9;	/* 0001 0101 s   */
  vlc_8x8i0Y[1][ 9][ 1].length =  9;	/* 0001 1010 s   */
  vlc_8x8i0Y[1][10][ 1].length =  9;	/* 0001 1001 s   */
  vlc_8x8i0Y[1][11][ 1].length =  9;	/* 0001 1000 s   */
  vlc_8x8i0Y[1][12][ 1].length =  9;	/* 0001 0111 s   */
  vlc_8x8i0Y[1][13][ 1].length =  9;	/* 0001 0110 s   */
  vlc_8x8i0Y[1][14][ 1].length = 10;	/* 0000 1100 1s  */
  vlc_8x8i0Y[1][15][ 1].length = 10;	/* 0000 1010 1s  */
  vlc_8x8i0Y[1][16][ 1].length = 10;	/* 0000 1011 0s  */
  vlc_8x8i0Y[1][17][ 1].length = 10;	/* 0000 1100 0s  */
  vlc_8x8i0Y[1][18][ 1].length = 10;	/* 0000 1011 1s  */
  vlc_8x8i0Y[1][19][ 1].length = 12;	/* 0000 0000 100s  */
  vlc_8x8i0Y[1][20][ 1].length = 12;	/* 0000 0000 101s  */
  vlc_8x8i0Y[1][21][ 1].length = 13;	/* 0000 0101 1000s */
  vlc_8x8i0Y[1][22][ 1].length = 13;  	/* 0000 0101 1001s */
  vlc_8x8i0Y[1][23][ 1].length = 13;	/* 0000 0101 1010s */
  vlc_8x8i0Y[0][ 0][ 0].length =  7;    /* 0000 011 (?) escape-code*/

  vlc_4x4x0Y[0][ 0][ 1].length =  5;
  vlc_4x4x0Y[0][ 0][ 2].length =  7;
  vlc_4x4x0Y[0][ 0][ 3].length =  9;
  vlc_4x4x0Y[0][ 0][ 4].length = 11;
  vlc_4x4x0Y[0][ 1][ 1].length =  5;
  vlc_4x4x0Y[0][ 1][ 2].length =  9;
  vlc_4x4x0Y[0][ 2][ 1].length =  6;
  vlc_4x4x0Y[0][ 2][ 2].length = 10;
  vlc_4x4x0Y[0][ 3][ 1].length =  7;
  vlc_4x4x0Y[0][ 3][ 2].length = 12;
  vlc_4x4x0Y[0][ 4][ 1].length =  7;
  vlc_4x4x0Y[0][ 5][ 1].length =  8;
  vlc_4x4x0Y[0][ 6][ 1].length =  9;
  vlc_4x4x0Y[0][ 7][ 1].length =  9;
  vlc_4x4x0Y[0][ 8][ 1].length = 11;
  vlc_4x4x0Y[0][ 9][ 1].length = 10;
  vlc_4x4x0Y[0][10][ 1].length = 11;
  vlc_4x4x0Y[1][ 0][ 1].length =  4;
  vlc_4x4x0Y[1][ 0][ 2].length =  9;
  vlc_4x4x0Y[1][ 1][ 1].length =  4;
  vlc_4x4x0Y[1][ 1][ 2].length =  9;
  vlc_4x4x0Y[1][ 2][ 1].length =  4;
  vlc_4x4x0Y[1][ 2][ 2].length = 11;
  vlc_4x4x0Y[1][ 3][ 1].length =  4;
  vlc_4x4x0Y[1][ 3][ 2].length = 11;
  vlc_4x4x0Y[1][ 4][ 1].length =  5;
  vlc_4x4x0Y[1][ 4][ 2].length = 11;
  vlc_4x4x0Y[1][ 5][ 1].length =  5;
  vlc_4x4x0Y[1][ 6][ 1].length =  6;
  vlc_4x4x0Y[1][ 7][ 1].length =  6;
  vlc_4x4x0Y[1][ 8][ 1].length =  6;
  vlc_4x4x0Y[1][ 9][ 1].length =  7;
  vlc_4x4x0Y[1][10][ 1].length =  8;
  vlc_4x4x0Y[1][11][ 1].length =  9;
  vlc_4x4x0Y[1][12][ 1].length =  9;
  vlc_4x4x0Y[1][13][ 1].length = 11;
  vlc_4x4x0Y[1][14][ 1].length = 12;
  vlc_4x4x0Y[0][ 0][ 0].length =  8;              /* escape-code */

  /* VLC table for INTRA TCOEF */

  vlc_4x4i0C[0][ 0][ 1].length =  4;
  vlc_4x4i0C[0][ 0][ 2].length =  6;
  vlc_4x4i0C[0][ 0][ 3].length =  7;
  vlc_4x4i0C[0][ 0][ 4].length =  8;
  vlc_4x4i0C[0][ 0][ 5].length =  8;
  vlc_4x4i0C[0][ 0][ 6].length =  9;
  vlc_4x4i0C[0][ 0][ 7].length =  9;
  vlc_4x4i0C[0][ 0][ 8].length = 10;
  vlc_4x4i0C[0][ 0][ 9].length = 10;
  vlc_4x4i0C[0][ 0][10].length = 10;
  vlc_4x4i0C[0][ 0][11].length = 10;
  vlc_4x4i0C[0][ 0][12].length = 11;
  vlc_4x4i0C[0][ 0][13].length = 11;
  vlc_4x4i0C[0][ 0][14].length = 11;
  vlc_4x4i0C[0][ 0][15].length = 10;
  vlc_4x4i0C[0][ 0][16].length = 10;
  vlc_4x4i0C[0][ 0][17].length = 11;
  vlc_4x4i0C[0][ 0][18].length = 11;
  vlc_4x4i0C[0][ 0][19].length = 11;
  vlc_4x4i0C[0][ 0][20].length = 12;
  vlc_4x4i0C[0][ 0][21].length = 12;
  vlc_4x4i0C[0][ 1][ 1].length =  7;
  vlc_4x4i0C[0][ 1][ 2].length =  9;
  vlc_4x4i0C[0][ 1][ 3].length = 10;
  vlc_4x4i0C[0][ 1][ 4].length = 10;
  vlc_4x4i0C[0][ 1][ 5].length = 11;
  vlc_4x4i0C[0][ 1][ 6].length = 13;
  vlc_4x4i0C[0][ 1][ 7].length = 12;
  vlc_4x4i0C[0][ 1][ 8].length = 12;
  vlc_4x4i0C[0][ 1][ 9].length = 11;
  vlc_4x4i0C[0][ 1][10].length = 12;
  vlc_4x4i0C[0][ 1][11].length = 12;
  vlc_4x4i0C[0][ 2][ 1].length =  9;
  vlc_4x4i0C[0][ 2][ 2].length = 11;
  vlc_4x4i0C[0][ 2][ 3].length = 11;
  vlc_4x4i0C[0][ 3][ 1].length = 10;
  vlc_4x4i0C[0][ 4][ 1].length = 11;
  vlc_4x4i0C[0][ 5][ 1].length = 12;
  vlc_4x4i0C[1][ 0][ 1].length =  3;
  vlc_4x4i0C[1][ 0][ 2].length =  4;
  vlc_4x4i0C[1][ 0][ 3].length =  6;
  vlc_4x4i0C[1][ 0][ 4].length =  7;
  vlc_4x4i0C[1][ 0][ 5].length =  8;
  vlc_4x4i0C[1][ 0][ 6].length = 10;
  vlc_4x4i0C[1][ 0][ 7].length =  9;
  vlc_4x4i0C[1][ 0][ 8].length =  9;
  vlc_4x4i0C[1][ 0][ 9].length = 12;
  vlc_4x4i0C[1][ 0][10].length = 10;
  vlc_4x4i0C[1][ 0][11].length =  8;
  vlc_4x4i0C[1][ 0][12].length =  8;
  vlc_4x4i0C[1][ 0][13].length =  7;
  vlc_4x4i0C[1][ 0][14].length =  6;
  vlc_4x4i0C[1][ 0][15].length =  7;
  vlc_4x4i0C[1][ 0][16].length =  7;
  vlc_4x4i0C[1][ 0][17].length =  7;
  vlc_4x4i0C[1][ 0][18].length =  9;
  vlc_4x4i0C[1][ 0][19].length =  9;
  vlc_4x4i0C[1][ 0][20].length =  9;
  vlc_4x4i0C[1][ 0][21].length = 15;
  vlc_4x4i0C[1][ 0][22].length = 14;
  vlc_4x4i0C[1][ 0][23].length = 12;
  vlc_4x4i0C[1][ 1][ 1].length =  5;
  vlc_4x4i0C[1][ 1][ 2].length =  8;
  vlc_4x4i0C[1][ 1][ 3].length = 11;
  vlc_4x4i0C[1][ 1][ 4].length = 14;
  vlc_4x4i0C[1][ 1][ 5].length = 16;
  vlc_4x4i0C[1][ 1][ 6].length = 16;
  vlc_4x4i0C[1][ 1][ 7].length = 14;
  vlc_4x4i0C[1][ 1][ 8].length = 13;
  vlc_4x4i0C[1][ 1][ 9].length = 12;
  vlc_4x4i0C[1][ 2][ 1].length =  5;
  vlc_4x4i0C[1][ 2][ 2].length =  9;
  vlc_4x4i0C[1][ 2][ 3].length = 11;
  vlc_4x4i0C[1][ 3][ 1].length =  7;
  vlc_4x4i0C[1][ 3][ 2].length = 11;
  vlc_4x4i0C[1][ 4][ 1].length =  7;
  vlc_4x4i0C[1][ 5][ 1].length =  9;
  vlc_4x4i0C[1][ 6][ 1].length =  9;
  vlc_4x4i0C[1][ 7][ 1].length = 10;
  vlc_4x4i0C[1][ 8][ 1].length = 11;
  vlc_4x4i0C[1][ 9][ 1].length = 12;
  vlc_4x4i0C[1][10][ 1].length = 12;
  vlc_4x4i0C[0][ 0][ 0].length =  9;              /* escape-code */

  vlc_table_init = 1;
}


